/*
**
**	 #########	   #	   # ######### #
**		#	    #	  #  #	       #
**	      ##   #	 #   #	 #   #	       #
**	     #	   #	 #    ###    #######   #
**	   ##	   #	 #   #	 #   #	       #
**	  #	   #	 #  #	  #  #	       #
**	 #########  ###### #	   # ######### #########
**			 #
**			 #
**		    #####
**
**
**	A control program for the ZyXEL U-1496 series MODEM for use with the
**	Linux operating system (and other UNIX-like systems)
**	Now also supports the ZyXEL 2864 series, the Omni MODEM and TA
**	series, and the Linux ISDN modem emulation (ttyIx).
**
**	This program manages the use of the ZyXEL MODEM for:
**
**	- dial-in, for use by "uucp" and other programs started from
**	  a login
**
**	- dial-in security (checking username and password)
**
**	- dial-back security
**
**	- dial-out, e.g. using "cu", "uucp" or "seyon"
**
**	- receiving FAXes
**
**	- sending FAXes
**
**	- polling for FAXes
**
**	- handling incoming FAX polls
**
**	- answering machine (voice mailbox)
**
**	Everything is done in a single program, to keep things simple
**	and fast.  Multi-program solutions are available from others.
**	(e.g. modgetty, mgetty+sendfax, vboxgetty)
**
**	Copyright 1993-2000 by Rob Janssen, PE1CHL
**
**	This program uses parts of "getty_ps", written by Paul Sutcliffe:
**	Copyright 1989,1990 by Paul Sutcliffe Jr.
**
**	Permission is hereby granted to copy, reproduce, redistribute,
**	or otherwise use this software as long as: there is no monetary
**	profit gained specifically from the use or reproduction or this
**	software, it is not sold, rented, traded or otherwise marketed,
**	and this copyright notice is included prominently in any copy
**	made.
**
**	The author make no claims as to the fitness or correctness of
**	this software for any use whatsoever, and it is provided as is.
**	Any use of this software is at the user's own risk.
*/

/*************************************************************************
**	User configuration section					**
*************************************************************************/


/*  Feature selection
 */

#undef	ZFAX				/* use AT# FAX commands (vs CLASS 2) */
#define ASCIIPID			/* PID stored in ASCII in lock file */
#define BOTHPID				/* ... or perhaps not */
#define DOUNAME				/* use uname() to get hostname */
#define FIDO				/* allow fido logins */
#define LOGFILE				/* log to file */
#define LOGUTMP				/* need to update utmp/wtmp files */
#define SETTERM				/* need to set TERM in environment */
#undef	TRYMAIL				/* mail logmesgs if CONSOLE unavailable */
#ifdef SVR4
#define SVR4LOCK			/* SysV R4 style lockfiles */
#endif

/* Pathnames that you might want to change
 */

#define CONSOLE		"/dev/console"	/* place to print fatal problems */

#ifdef LOGFILE
#undef	LOGFILE
#define LOGFILE		"/var/log/%s.log"	/* place to log events and errors */
						/* %s gets invocation name */
#else
#define LOGFILE		CONSOLE
#endif

#ifdef SVR4LOCK
#define LOCK		"/usr/spool/locks/LK.%03u.%03u.%03u" /* SysV R4 */
#else
#define LOCK		"/var/lock/LCK..%s"	/* others */
#endif

#ifdef ETC_DEFAULT
#define DEFAULTS	"/etc/default/%s"	/* name of defaults file */
#else
#define DEFAULTS	"/etc/%s.conf"		/* FSSTND style name */
#endif
#define ISSUE		"/etc/issue"		/* name of the issue file */
#define LOGIN		"/bin/login"		/* name of login program */
#define PHOSTNAME	"/bin/hostname"		/* prints our hostname */
#define TTYTYPE		"/etc/ttytype"		/* terminal types */
#define UUCPNAME	"uucp"			/* uucp user in passwd file */

#ifdef SVR4
#define MAILER		"/bin/mailx"
#else
#define MAILER		"/usr/lib/sendmail -bm -oi -odb"
#endif
#define NOTIFY		"root"


/* const is buggy in SCO XENIX */

#ifdef M_XENIX
# define const
#endif

/*************************************************************************
**	Include files...  Just about everything you can get		**
*************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/types.h>
#if defined(__linux__) || defined(SVR4)
# include <sys/ioctl.h>
#endif
#ifndef M_XENIX
# include <sys/time.h>
#endif
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#ifdef SVR4LOCK
# include <sys/mkdev.h>
#endif

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <termio.h>
#include <time.h>
#include <utmp.h>
#include <math.h>
#ifdef _GNU_SOURCE
# include <getopt.h>
#endif

#ifdef __linux__				/* for highspeed serial stuff */
# ifdef OLD_TTY_DRIVER
#  include <linux/fs.h>				/* these are for the */
#  include <linux/tty.h>			/* ... old tty driver */
# else
#  include <linux/serial.h>			/* this is for new tty driver */
# endif
# include <sys/soundcard.h>			/* sound card access */
# ifdef SNDCTL_DSP_SPEED
#  define WITH_SOUND
# endif
#endif

#ifdef HAVE_SAS			/* only enable this if you have SAS-1.25.7 - RXTN */
#include <sys/sas.h>
#endif

/* some special defs for SysV R4 */
#ifdef SVR4
#define signal(signo,sighndlr)	sigset((signo),(sighndlr))
extern int select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
#endif

#ifndef __USE_BSD
/* directory scan functions not available in SysV */
extern int scandir(const char *__dir,
		   struct dirent ***__namelist,
		   int (*__select) (const struct dirent *),
		   int (*__cmp) (const void *, const void *));
extern int alphasort(const void *, const void *);
#endif

/* version/whatstring */

static const char *Id = "@(#)ZyXEL 2.0 (" __DATE__ " " __TIME__ ")   (c) 1993-2000 by Rob Janssen, PE1CHL";

/* macro definitions */

#define FALSE		(0)
#define TRUE		(1)

#define SUCCESS		(0)	/* normal return */
#define FAIL		(-1)	/* error return */

#define STDIN		0
#define STDOUT		1
#define STDERR		2

				/* modemlines to be monitored and state */
#define MODEMLINES	(TIOCM_DTR|TIOCM_RTS|TIOCM_CTS|TIOCM_CAR|TIOCM_DSR)
#define MODEMLN_OK	(TIOCM_DTR|TIOCM_RTS|TIOCM_CTS)

#ifndef CRTSCTS
#define CRTSCTS		0
#endif

#define SHOWMSG_AA	1	/* show messages present by turning AA on */
#define SHOWMSG_DSR	2	/* show messages present by turning DSR on */

#define BUTTON_ANSWER	0	/* answer the phone when DATA/VOICE button */
#define BUTTON_PLAYB	1	/* playback recorded messages when button */
#define BUTTON_EPADMP	2	/* dump EPA contents when button */

#define EXPFAIL		30	/* default num seconds to wait for expected input */

#define EXPECT		0	/* states for chat() */
#define SEND		1

#define INITIALIZE	0	/* indicators for settermios() */
#define INITIAL		1
#define DROPDTR		2
#define SHORTTIME	3
#define SHORT255	4
#define MEDMTIME	5
#define LONGTIME	6
#define XONXOFF		7
#define IMMEDIATE	8
#define WATCHDCD	9
#define SEVEN_E		10
#define FINAL		11

#define ONLINE		0	/* indicators for datamode(),faxmode(),voicemode() */
#define ANSWER		1
#define ANSWERDATA	2
#define FAXPOLL		3

#define NOZYXEL		1	/* Return values for initmodem() */

#define INITFAIL	1	/* Return values for wait_work() */
#define LOCKED		2
#define FIFO		3
#define CONNECT		4
#define NOCARRIER	5
#define RING		6
#define NORING		7
#define VCON		8
#define EVENT		9

#define INCORRECT	11	/* Return values for protect() */

#define NOANSWER	20	/* Return values for datamode(),voicemode() */
#define DATAFAX		21
#define DATA		22
#define FAX		23
#define VOICE		24
#define MESSAGE		25
#define DIALBACK	26
#define SILENCE		27
#define POTSCALL	28
#define NOLOCAL		29

#define EXIT		31	/* Return values for voicecommand() */
#define COMMAND		32
#define UNKNOWN		33
#define ERROR		34

#define DIALFAIL	41	/* Return values for faxpoll(),faxsend() */
#define CONNECTFAIL	42
#define NOPOLL		43
#define XFERFAIL	44	/* also from faxmode() */

#define BADSPEED	51	/* Return values for getlogname() */
#define BADCASE		52
#define NONAME		53
#define FIDOCALL	54
#define BBSCALL		55
#define PPPCALL		56

				/* Firmware features bits */
#define FW_PLUS		0x00001 /* is a PLUS model (U-1496E+, can do CELP) */
#define FW_2864		0x00002 /* is an 2864 model (Elite or Omni) */
#define FW_128		0x00004 /* is an 128 model (Omni TA 128) */
#define FW_LINUX	0x00008	/* is Linux ISDN modem emulator */
#define FW_ISDN		0x00010 /* is an ISDN model (2864I, Omni TA 128) */
#define FW_DTMFREC	0x00020 /* can record DTMF-only */
#define FW_RESYNC	0x00040 /* can do audio resync (S39.7=1) */
#define FW_AUDIOABORT	0x00080 /* can abort AUDIO playback */
#define FW_DIALTONE	0x00100 /* can detect DIALTONE */
#define FW_DTMFDET	0x00200 /* can set DTMF detection sensitivity */
#define FW_RTSCTSFLOW	0x00400 /* can RTS/CTS AUDIO and FAX data */
#define FW_NEWADPCM	0x00800 /* ADPCM3N/ADPCM4 recording */
#define FW_NEWLEAK	0x01000 /* new LEAKAGE coeff */
#define FW_FAXEC	0x02000 /* can do FAX EC */
#define FW_VBTI		0x04000 /* can set AT+VBT interval */
#define FW_DTMFPAUSE	0x08000 /* PAUSE when DTMF received */
#define FW_FAXBUG	0x10000 /* has FAX bug (no OK after FCLASS=2) */
#define FW_CLIR		0x20000 /* can do CLIR (ISDN models) */
#define FW_FASTC	0x40000 /* can do ATC$1 (ISDN models) */
#define FW_PLAYBUG	0x80000	/* has VOICE PLAY bug (U-1496 hangs when CND) */
#define FW_XLAW		0x100000 /* can do A-Law/U-Law audio (Linux ISDN) */

#define MAXDEF		90	/* max # lines in defaults file */
#define MAXLINE		256	/* max # chars in a line */
#define MAXBUF		1024	/* buffer size */
#define MAXPASSWD	20	/* max # chars in a password */
#define NARG		20	/* max # of args on a command line */

#define TSYNC		0xAE	/* Fido stuff (chars recognized at login:) */
#define YOOHOO		0xF1
#define ESC		0x1B
#define BS		0x08
#define DEL		0x7F

#define STX		002	/* ASCII control chars (for voice data etc) */
#define ETX		003
#define EOT		004
#define ACK		006
#define DLE		020
#define XON		021
#define DC2		022
#define XOFF		023
#define DC4		024
#define NAK		025
#define CAN		030
#define SUB		032
#undef FS
#define FS		034

#ifndef CEOF
#define CEOF		004	/* ^D */
#define CEOL		012	/* ^J */
#define CERASE		010	/* ^H */
#define CINTR		003	/* ^C */
#define CKILL		030	/* ^X */
#define CQUIT		034	/* ^\ */
#define CSTART		021	/* ^Q */
#define CSTOP		023	/* ^S */
#define CSUSP		032	/* ^Z */
#define CRPRNT		022	/* ^R */
#define CWERASE		027	/* ^W */
#define CLNEXT		026	/* ^V */
#endif

#define strncopy(s1,s2) (strncpy(s1,s2,sizeof(s1)))
#define getch()		(InPtr >= InLen && (InLen = read(STDIN,InBuf,BUFSIZ)) <= (InPtr = 0)? EOF : (int)(unsigned char)InBuf[InPtr++])
#define getapi()	(ApiPtr >= ApiLen && (ApiLen = read(FifoFd,ApiBuf,BUFSIZ)) <= (ApiPtr = 0)? EOF : (int)(unsigned char)ApiBuf[ApiPtr++])

/* debug levels
 */
#define D_OPT		0001	/* option settings */
#define D_DEF		0002	/* defaults file processing */
#define D_UTMP		0004	/* utmp/wtmp processing */
#define D_CHAT		0010	/* MODEM chat (send/expect/waitfor) */
#define D_FAX		0020	/* FAX responses */
#define D_GETL		0040	/* get login name routine */
#define D_RUN		0100	/* other runtime diagnostics */
#define D_LOCK		0200	/* lockfile processing */
#define D_EPA		0400	/* EPA dump chat */

#ifdef DEBUG

/* debug defs
 */
#define debug1(a,b)		{ if (Debug & a) dprint(b); }
#define debug2(a,b)		{ if (Debug & a) debug(b); }
#define debug3(a,b,c)		{ if (Debug & a) debug(b,c); }
#define debug4(a,b,c,d)		{ if (Debug & a) debug(b,c,d); }
#define debug5(a,b,c,d,e)	{ if (Debug & a) debug(b,c,d,e); }
#define debug6(a,b,c,d,e,f)	{ if (Debug & a) debug(b,c,d,e,f); }
#define debugtime(a)		{ if (Debug & a) debugt(); }

#else	/* DEBUG */

#define debug1(a,b)		/* define to nothing, disables debugging */
#define debug2(a,b)
#define debug3(a,b,c)
#define debug4(a,b,c,d)
#define debug5(a,b,c,d,e)
#define debug6(a,b,c,d,e,f)
#define debugtime(a)

#endif	/* DEBUG */

/* data definitions */

typedef int	boolean;	/* useless crap */
typedef void	sigh_t;		/* type of a signal handler */

/* lines in defaults file are in the form "NAME=value"
 */

typedef struct Default {
	char	*name;		/* name of the default */
	char	*value;		/* value of the default */
} DEF;

#ifdef ZFAX
/* header of .fax (ZyXEL FAX) files
 * this assumes lo-hi byte ordering and 16-bit shorts :-(
 */

typedef struct fax_header {
	char zyxel[6];		/* magic string */
#define FAX_MAGIC	"ZyXEL" /* includes \0 */
	short factory;		/* always 'Z' */
	short version;		/* always 2 */
	short rec_width;	/* recording width 1728/2048/2432 */
#define FAX_R0		1728
#define FAX_R1		2048
#define FAX_R2		2432
	short pages;		/* number of pages */
	short flags;
#define FAX_V1		0x0001	/* high resolution */
#define FAX_T1		0x0002	/* 2-D coding */
} FAX_HEADER;

const char FaxExt[] = ".fax";	/* the extension */
#else
/* CLASS 2 FAX information received in +F responses
 */

typedef struct fax_info {
	int cfr;		/* +FCFR received */
	int et;			/* +FET post message response */
	int hng;		/* +FHNG hangup response */
	int poll;		/* +FPOLL received */
	int dtc;		/* +FDTC: received (poll request) */

	/* info from +FPTS */
	int pts;		/* page transfer status */
	int lc;			/* line count */
	int blc;		/* bad line count */
	int cblc;		/* consecutive bad line count */
	int lbc;		/* lost byte count */

	/* info from +FDIS, +FDCS, +FDTC */
	int vr;			/* vertical resolution */
	int br;			/* bitrate */
	int wd;			/* page width */
	int ln;			/* page length */
	int df;			/* data format */
	int ec;			/* error correcting mode */
	int bf;			/* binary file transfer */
	int st;			/* scan time */

	/* info from +FCSI, +FTSI, +FCIG */
	char id[30];		/* identification string */

	/* info from +FNSC */
	int nstyp;		/* type of nonstandard frame (F, S, C) */
	int nslen;		/* length of nonstandard frame */
	unsigned char ns[64];	/* nonstandard frame */
} FAX_INFO;

const char FaxExt[] = ".g3";	/* the extension */
#endif

/* header of .zad (ZyXEL audio) files
 * this assumes lo-hi byte ordering and 16-bit shorts :-(
 */

typedef struct zad_header {
	char zyxel[6];		/* magic string */
#define ZAD_MAGIC	"ZyXEL\2"
	short version;
	short reserved1;
	short enc;		/* audio encoding (compression) scheme */
#define ENC_CELP	0x00
#define ENC_ADPCM2	0x01
#define ENC_ADPCM3	0x02
#define ENC_ADPCM4	0x03	/* AT+VSM=4 (from 6.13) */
#define ENC_RESYNC	0x04	/* this bit indicates resync is used */
#define ENC_ADPCM3N	0x08	/* AT+VSM=30 (from 6.13) */
#define ENC_S16		0x10	/* signed 16-bit (internal use only) */
#define ENC_ALAW	0x11	/* AT+VSM=5 (A-law Linux ISDN) */
#define ENC_ULAW	0x12	/* AT+VSM=6 (U-law Linux ISDN) */
	short reserved2;
	short reserved3;
} ZAD_HEADER;

const char AudioExt[] = ".zad"; /* the extension */

#define	ENC_BEST	ENC_ALAW /* preferable encoding for audio to modem */


/* variables */

#ifdef DEBUG
int	Debug;			/* debug value from command line */
FILE	*Dfp;			/* debugging file pointer */
#endif
#ifdef WITH_SOUND
FILE	*SoundFp;		/* file pointer for soundcard */
#endif

boolean Fax = TRUE;		/* support sending/receiving FAX */
boolean FinalXoff = TRUE;	/* set IXON and IXANY after logging in */
boolean Is_ZyXEL;		/* MODEM seems to be a ZyXEL */
boolean LockPoll;		/* poll for lockfiles while IDLE? */
boolean LoginEnv = TRUE;	/* pass environment as args to LOGIN? */
boolean ModemPoll;		/* poll the MODEM while IDLE? */
boolean NeedInit = FALSE;	/* need Modem Init after some operation? */
boolean NoHangUp;		/* don't hangup line before setting speed */
boolean No_nonblock;		/* Don't use O_NONBLOCK */
boolean No_ZyXEL;		/* Don't enforce ZyXEL MODEM */
boolean PlayCND;		/* sent CND OFF to work around PLAY bug? */
boolean Voice;			/* support VOICE */
boolean WarnCase;		/* controls display of bad case message */
char	*AnswerTone;		/* VOICE mode answer tone */
char	*Archive;		/* archiving directory */
char	*Beep;			/* beep spec being run */
#ifndef ZFAX
char	*Buffer;		/* Pathname of buffering program */
#endif
char	*CndFile;		/* file to read CND info (from other unit) */
char	*CndName;		/* last received CND NAME */
char	*CndNumber;		/* last received CND NUMBER */
char	*Connect;		/* last received CONNECT message from MODEM */
char	*DataFlags;		/* flags to pass to DATA mode program */
char	*DataProg;		/* program to handle DATA mode (instead of login) */
char	*DefName;		/* defaults file name */
char	*Device;		/* controlling line (minus "/dev/") */
char	*DevName;		/* controlling line including "/dev/" */
char	*DialPrefix;		/* sent before each phone number */
char	*DtmfDet;		/* DTMF detection sensitivity */
char	*EpaFile;		/* EPA dump filename */
char	*Error;			/* error beep for voice mode */
char	*FaxDir;		/* Incoming FAX directory */
char	*FaxFile = "test.g3";	/* Incoming FAX file */
char	*FaxId;			/* FAX ID (phone number) */
char	*FaxMail;		/* Address to send 'fax arrived' mail msgs */
#ifndef ZFAX
char	*FaxPoll;		/* file to send when polled for FAX */
char	*FaxRej;		/* Fax IDs to reject */
#endif
#ifdef FIDO
char	*Fido;			/* fido protocol program */
#endif	/* FIDO */
char	*Fifo;			/* API FIFO configured name */
char	*FifoFile;		/* API FIFO full pathname */
#ifndef ZFAX
char	*GhostScript;		/* Pathname of Ghostscript program */
#endif
char	*Greeting = "greeting"; /* Greeting for VOICE mode */
#ifndef ZFAX
char	*Gunzip;		/* Pathname of Gunzip program */
#endif
char	*Init;			/* value of INIT */
char	*Issue = ISSUE;		/* default issue file */
char	*LSOn = "M2";		/* Speaker-ON command sent in VOICE mode */
char	*LocalKey;		/* Voice DTMF keys file for local phone */
char	*Lock;			/* lockfile name */
char	*Login = LOGIN;		/* default login program */
char	*LoginFlags;		/* flags for login program */
char	*Mailer = MAILER;	/* program used to send mail (arg = destin.) */
char	*Match;			/* location in ChatBuf where waitfor matched */
char	*Message;		/* alternative Greeting/Message for VOICE */
char	*MidPlay;		/* MIDI player executable program */
char	*MidPlayFlags;		/* flags for MIDI player */
char	*ModPlay;		/* MOD player executable program */
char	*ModPlayFlags;		/* flags for MOD player */
char	*Mp3Play;		/* MP3 player executable program */
char	*Mp3PlayFlags;		/* flags for MP3 player */
char	*MsgArch;		/* beep for "message archived" */
char	*MsgDel;		/* beep for "message deleted" */
char	*Mute = "M1";		/* mute command (Lx/Mx/Nx) sent in DATA/FAX mode */
char	*MyName;		/* this program name */
char	*PPPFlags;		/* flags to pass to PPP program */
char	*PPPProg;		/* program to handle PPP (instead of login) */
char	*PagerCentral;		/* telephone number of PAGER service (IXO/TAP/PET) */
char	*PagerNum;		/* (telephone) number of PAGER */
char	*PagerRecMsg = "*%s*%d"; /* message sent to PAGER when recording made */
char	*PagerSource = "";	/* source (telephone) number sending PAGER msg */
char	*PlayDir;		/* directory of voice play files */
char	*ProdInfo;		/* MODEM product info (ATI) */
char	*Prompt1;		/* prompt beep for DTMF digit entry */
char	*Prompt2;		/* prompt beep for DTMF code entry */
char	*Protect;		/* name of "protected login" file */
#ifndef ZFAX
char	*PsText;		/* Pathname of PsText program */
#endif
char	*Ready;			/* ready beep (after startup) */
char	*RecAttn;		/* attention beep */
char	*RecDir;		/* recording directory */
char	*RecDone;		/* recording done beep */
char	*RecFile = "test";	/* recording file */
char	*RingType[8];		/* ringtypes corresponding to 8 numbers */
char	*Silence = "25,70";	/* silence-detect parameters */
#ifdef WITH_SOUND
char	*SoundDev;		/* device name for soundcard */
#endif
char	*StateFile = "/tmp/ZyXEL";	/* file holding current state */
char	*StateProg;		/* program to call to get current state */
char	*Synth;			/* voice synthesizer executable program */
char	*SynthFlags;		/* flags for synthesizer program */
char	*SysName;		/* nodename of system */
char	*ToNumber;		/* last received TO: NUMBER (ISDN) */
char	*Version;		/* value of VERSION */
char	*VoiceCmd;		/* Voice expert command file */
char	*VoiceKey;		/* Voice DTMF keys file */
char	*VoiceRing;		/* Sent after 1st RING in VOICE mode (volume) */
char	ApiBuf[BUFSIZ];		/* read() buffer for API FIFO */
char	ChatBuf[MAXBUF];	/* result from last chat() */
char	CurVSD[20];		/* current setting of VSD= */
char	InBuf[BUFSIZ];		/* read() buffer for STDIN */
char	MsgBuf[MAXLINE];	/* message buffer */
char	Modem[MAXLINE];		/* Modem description (make, version) */
char	Number[80];		/* from readnumber command */
char	Recordfile[MAXLINE];	/* 'current recording file' for voice mailbox */
#ifdef SETTERM
char	Term[16];		/* terminal type */
#endif	/* SETTERM */
double	Beep1;			/* Tone 1 for beep function */
double	Beep2;			/* Tone 2 for beep function */
int	ApiLen;			/* ApiBuf current length */
int	ApiPtr;			/* ApiBuf character ptr */
int	ApiUserLevel;		/* default userlevel for users on API FIFO */
int	BeepPause;		/* set when pausing after Beep */
int	ButtonAction;		/* action when DATA/VOICE button pressed */
int	Cbaud;			/* decoded baudrate field */
int	CheckLine;		/* time to check line before playing greeting */
int	CurLS;			/* currently set LS */
int	CurSync;		/* currently set resync mode (voice) */
int	CurVSM;			/* currently set VSM */
int	DialTime;		/* extra time when dialling (pulse) */
int	DialTries = 1;		/* retries when dialling out (protect) */
int	DropMsg;		/* drop messages at or below this duration */
int	EpaInit;		/* dump EPA before initializing? */
#ifndef ZFAX
int	FaxComp = TRUE;		/* supported FAX compression */
#endif
int	Fclass;			/* current FCLASS of modem */
int	FifoFd = -1;		/* API FIFO file descriptor */
int	Fsamp = 9600;		/* Voice sample rate (samples/s) */
int	Fw_beta;		/* Firmware betaversion (if any) */
int	Fw_features;		/* Firmware features bits */
int	Fw_version;		/* Firmware version number */
int	InLen;			/* InBuf current length */
int	InPtr;			/* InBuf character ptr */
int	Mode = DATAFAX;		/* answer mode */
int	ModeRB;			/* answer mode after RingBack */
#ifdef TIOCMGET
int	Modemlines = MODEMLINES; /* mask to test modem lines */
int	Modemln_ok = MODEMLN_OK; /* correct state of lines when idle */
#endif
int	Nrings = 4;		/* required number of rings to answer */
int	Nusers;			/* number of users currently logged in */
int	PagerType;		/* type of PAGER (dial algorithm) */
int	PhoneLS = 2;		/* LS to select phoneline (plus speaker) */
int	RecLS;			/* recording LS (0 = current) */
int	RecMode = ENC_CELP;	/* recording mode */
int	RingBack;		/* ringback datamode enabled (# rings) */
int	ShowMsg;		/* show "messages recorded" */
int	SoundFmt;		/* sound format currently on soundcard */
int	SoundRt;		/* sound samplingrate currently on soundcard */
int	Speed;			/* current baud rate */
int	TimeOut;		/* timeout value from command line */
int	Tstate;			/* last settermios state */
int	UserLevel;		/* userlevel for current VOICE user */
int	WFclass = 8;		/* FCLASS while waiting for work */
long	Bsamples;		/* samples for beep function */
long	Samples;		/* samples read from pipe in say() */
long	Signal;			/* last caught signals */
long	Leakage;		/* for ADPCM coder */
short	Delta;			/* for ADPCM coder */
short	EstMax;			/* for ADPCM coder */
struct termios Termios;		/* termios used for the tty */
struct timeval MatchTime;	/* time of last matched input */
time_t	HoldTime;		/* holdtime for selected state (mode) */
time_t	RecTime;		/* timestamp of last recording */
uid_t	UuUid;			/* UUCP's user-id */
gid_t	UuGid;			/* UUCP's group-id */
mode_t	FifoPerm = 0200;	/* API FIFO default permission (write by owner) */
#ifdef __linux__
struct serial_struct Serial;	/* high-speed and other serial details */
#endif

#ifdef ZFAX
/* RTC strings (end-of-page) for 1-D and 2-D FAX reception */

const unsigned char fax_rtc_1d[] = {0x00,0x08,0x80,0x00,0x08,0x80,0x00,0x08,0x80};
#define RTC_1D_LEN	9
const unsigned char fax_rtc_2d[] = {0x00,0x18,0x00,0x03,0x00,0x06,0x00,0x0c,0x80,0x01,0x30,0x00};
#define RTC_2D_LEN	12
#else				/* Class 2 */
FAX_INFO FaxInfo;		/* info about current FAX connection */

/* RTC string (end-of-page) */

const char fax_rtc[] = {0x00,0x10,0x01,0x00,0x10,0x01,0x00,0x10};
#define RTC_LEN		8

/* blank FAX_INFO structure used as a startoff */

const FAX_INFO blank_info = {
	0, -1, -1, 0, 0,	/* cfr, et, hng, poll, dtc */
	-1, -1, -1, -1, -1,	/* pts, lc, blc, cblc, lbc */
	-1, -1, -1, -1, -1, -1, -1, -1, /* dcs params */
	"",			/* ID */
	0, 0, ""		/* nstyp, nslen, ns */
};

/* nonstandard facility used to send polling ID when polling for FAXes */

const unsigned char poll_nsc[] = {
	0xad,0x00,0x12,0x00,0x00,0x50,0x4f,0x4c,0x4c,0x2d,0x49,0x44,0x3a
};

const char poll_hex[] = "AD00120000504F4C4C2D49443A";

/* bit reverse table (some firmware revisions can't do it...) */

const unsigned char rev_bits[] = {
	0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,0x10,0x90,0x50,0xd0,0x30,0xb0,0x70,0xf0,
	0x08,0x88,0x48,0xc8,0x28,0xa8,0x68,0xe8,0x18,0x98,0x58,0xd8,0x38,0xb8,0x78,0xf8,
	0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4,
	0x0c,0x8c,0x4c,0xcc,0x2c,0xac,0x6c,0xec,0x1c,0x9c,0x5c,0xdc,0x3c,0xbc,0x7c,0xfc,
	0x02,0x82,0x42,0xc2,0x22,0xa2,0x62,0xe2,0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2,
	0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,0x1a,0x9a,0x5a,0xda,0x3a,0xba,0x7a,0xfa,
	0x06,0x86,0x46,0xc6,0x26,0xa6,0x66,0xe6,0x16,0x96,0x56,0xd6,0x36,0xb6,0x76,0xf6,
	0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe,
	0x01,0x81,0x41,0xc1,0x21,0xa1,0x61,0xe1,0x11,0x91,0x51,0xd1,0x31,0xb1,0x71,0xf1,
	0x09,0x89,0x49,0xc9,0x29,0xa9,0x69,0xe9,0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9,
	0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,0x15,0x95,0x55,0xd5,0x35,0xb5,0x75,0xf5,
	0x0d,0x8d,0x4d,0xcd,0x2d,0xad,0x6d,0xed,0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,0x7d,0xfd,
	0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3,
	0x0b,0x8b,0x4b,0xcb,0x2b,0xab,0x6b,0xeb,0x1b,0x9b,0x5b,0xdb,0x3b,0xbb,0x7b,0xfb,
	0x07,0x87,0x47,0xc7,0x27,0xa7,0x67,0xe7,0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7,
	0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,0x1f,0x9f,0x5f,0xdf,0x3f,0xbf,0x7f,0xff
};
#endif

const char AudioResume[] = { DLE, 'r', '\0' };		/* resume playing */
const char AudioEnd[] = { DLE, ETX, '\0' };		/* end audio */
const char AudioRecAbort[] = { DLE, '!', '\r', '\0' };	/* abort recording */

char AudioAbort[] = { ETX, DLE, ETX, '\0' };		/* abort audio */
/*		      ^^^	^^^  this byte overwritten with DC4 when >= 6.11 */

const short MaxTbl[][8] =				/* used in ADPCM conversion */
{
	/* dummy entry to make ENC_XXX the index into this table */
	{ 0 },

	/* ADPCM2, 6.10 or later */
	{ 0x3800, 0x5600 },

	/* ADPCM3, 6.10 or later */
	{ 0x399A, 0x3A9F, 0x4D14, 0x6607 },

	/* ADPCM4, 6.13 or later */
	{ 0x3556, 0x3556, 0x399A, 0x3A9F, 0x4200, 0x4D14, 0x6607, 0x6607 },
};

const char *bad_case[] = {			/* UPPER-CASE-LOGIN warning */
	"\n",
	"If your terminal supports lower case letters, please\n",
	"use them.  Login again, using lower case if possible.\n",
	NULL
};


/* function prototypes */

DEF	**defbuild(char *filename);
DEF	*defread(FILE *fp);
FILE	*defopen(char *filename);
boolean checklock(char *name);
boolean expmatch(const char *got,const char *exp);
char	*apigetl(void);
char	*apigets(void);
char	*defvalue(DEF **deflist,const char *name);
char	*dtmfstring(void);
char	*faxident(void);
char	*getuname(void);
char	*make_fclass(int fclass);
char	*make_vls(int ls);
const char *unquote(const char *s,char *c);
int	Fputs(const char *s,int fd);
int	alaw2linear(unsigned char u_val);
int	api_eol(void);
int	api_mode(void);
int	apigetc(void);
int	auplay(char *filename);
int	auread(int fd,char *buf,size_t size);
int	ausend(int fd,int (*rdf)(int fd,char *buf,size_t size),int maxtime);
int	beep(const char *s);
int	chat(char *s);
int	checksync(char *where);
int	cndmessage(char *msg);
int	datamode(int answer);
int	defclose(FILE *fp);
int	dialup(const char *mode,int fclass,int timeout,const char *number);
int	dir_select(const struct dirent *dirent);
int	docommand(char *line,int (*inputdigit)(void),char *(*inputstring)(void));
int	dtmfcode(void);
int	dtmfdigit(void);
int	dtmfrecord(const char *end,int maxtime);
int	dump_epa(void);
int	execlogin(char *login,char *flags,char *username,boolean loginenv);
int	expect(const char *s);
int	fax2rsp(int logfd);
int	faxcpoll(int logfd);
int	faxmode(int answer);
int	faxpoll(char *number,char *pollid,int logfd);
int	faxrecv(int answer,int logfd);
int	faxsend(char *number,char *filename,int logfd);
int	filespresent(const char *dirname);
int	getlogname(struct termios *termios,char *name,int size);
int	getpasswd(char *passwd,int size);
int	initmodem(int dropdtr);
int	line_epa(char *line);
int	localvoice(void);
int	login_user(void);
int	main(int argc,char *argv[]);
int	makelock(char *name);
int	messagemode(int answer,char *message);
int	modemstat(void);
int	music(char *filename);
int	opendebug(void);
int	openline(void);
int	pager(const char *info);
int	play(const char *filename,const char *end);
int	playback(char *direc,int (*inputdigit)(void));
int	protect(char *username,int askpasswd);
int	push(pid_t *pid,int *fd,char *program,char *argv[]);
int	read_state(int ringtype,int altcount);
int	readbeep(int fd,char *buf,size_t size);
int	readlock(char *name);
int	readulaw(int fd,char *buf,size_t size);
int	readulaw8k(int fd,char *buf,size_t size);
int	readulaw9k6(int fd,char *buf,size_t size);
int	record(char *filename,boolean beeps,const char *end,const char *vsd,int enc,int maxtime);
int	rip(char *string);
int	run(boolean synth,char *argv[]);
int	say(const char *string);
int	send(const char *s);
int	set_vls(int ls);
int	set_voice(int ls);
int	set_vsm(int enc);
int	sound_mode(int enc,int fsamp);
int	sound_open(void);
int	startsynth(pid_t *pid,int *ifd,int *ofd);
int	toadpcm(int enc,short int edata);
int	ulaw2linear(unsigned char u_val);
int	voicecommand(char *filename,int code,int (*inputdigit)(void),char *(*inputstring)(void));
int	voicekey(char *filename,int firstdigit);
int	voicemode(int answer,char *message);
int	wait_work(int fclass);
int	wait_vcon(int state);
int	waitfor(const char *word,int state);
int	zmsp_test(int out);
sigh_t	expalarm(int);
sigh_t	huphandler(int);
sigh_t	inthandler(int);
sigh_t	sighandler(int);
sigh_t	timeout(int);
short int fromadpcm(int enc,char vdata);
unsigned char linear2alaw(int pcm_val);
unsigned char linear2ulaw(int pcm_val);
void	closeline(void);
void	cndfile(void);
void	debug(const char *fmt,...);
void	debugt(void);
void	dprint(const char *word);
void	exit_usage(int code);
void	gets_state(FILE *fp,int altcount);
void	getvalues(void);
void	hold_state(int duration,int mode,int nrings);
void	initadpcm(void);
void	linestat(void);
void	logmesg(const char *message);
void	rmlocks(void);
void	settermios(int state);
void	sound_close(void);
#ifdef ZFAX
void	faxheader(FAX_HEADER *header,char *connect);
#endif


/*************************************************************************
**	Functions...							**
*************************************************************************/

/*
** main: where it all starts
**
** we are called from "init" as: /etc/ZyXEL [options] device speed [term]
** example: /etc/ZyXEL ttyS1 38400
*/

int
main(argc,argv)
int argc;
char *argv[];
{
	FILE *fp;
	int c,mode,answer;
	struct passwd *pwd;
	struct stat st;
	char buf[MAXLINE+1];

	/* startup
	 */
	nice(-9);				/* real-time execution */
	srand((int)time(NULL));

	signal(SIGQUIT, SIG_DFL);		/* make sure we can quit */
	signal(SIGTERM, SIG_DFL);

	atexit(closeline);			/* close line when exit */

#ifdef SETTERM
	strcpy(Term, "unknown");
#endif	/* SETTERM */

	/* who am I?  ZyXEL, of course...
	 */
	if ((MyName = strrchr(argv[0],'/')) != NULL)
	    MyName++;
	else
	    MyName = argv[0];

	/* process the command line
	 */

	while ((c = getopt(argc, argv, "D:d:hlmnt:Z?")) != EOF) {
		switch (c) {
		case 'D':
#ifdef DEBUG
			sscanf(optarg, "%o", &Debug);
#else	/* DEBUG */
			logmesg("DEBUG not compiled in");
#endif	/* DEBUG */
			break;
		case 'd':
			DefName = optarg;
			break;
		case 'h':
			NoHangUp = TRUE;
			break;
		case 'l':
			LockPoll = TRUE;
			break;
		case 'm':
			ModemPoll = TRUE;
			break;
		case 'n':
			No_nonblock = TRUE;
			break;
		case 't':
			TimeOut = atoi(optarg);
			break;
		case 'Z':
			No_ZyXEL = TRUE;
			break;
		case '?':
			exit_usage(2);
			break;
		}
	}

	/* normal System V init options handling
	 */

	if (optind < argc)
	    Device = argv[optind++];
	if (optind < argc)
	    Speed = atoi(argv[optind++]);
#ifdef SETTERM
	if (optind < argc)
	    strncpy(Term, argv[optind++], sizeof(Term));
#endif	/* SETTERM */

	if (!Device) {
	    logmesg("no line given");
	    exit_usage(2);
	}

#ifdef DEBUG
	if (opendebug() != SUCCESS)
	    return EXIT_FAILURE;
#endif

#if defined(SETTERM) && defined(TTYTYPE)
	if (!strcmp(Term, "unknown")) {
	    if ((fp = fopen(TTYTYPE, "r")) == NULL) {
		sprintf(MsgBuf, "cannot open %s", TTYTYPE);
		logmesg(MsgBuf);
	    } else {
		char device[MAXLINE+1];

		while ((fscanf(fp, "%s %s", buf, device)) != EOF) {
		    if (!strcmp(device, Device)) {
			strncpy(Term,buf,sizeof(Term));
			break;
		    }
		}
		fclose(fp);
	    }
	}
#endif	/* SETTERM */

	/* need full name of the device
	 */
	sprintf(buf, "/dev/%s", Device);
	DevName = strdup(buf);

	/* command line parsed, now build the list of
	 * runtime defaults; this may override things set above.
	 */

	getvalues();			/* read defaults from file */

	/* generate the FIFO filename (when configured)
	 * create the FIFO when it doesn't exist yet
	 */
	if (Fifo != NULL) {
	    sprintf(buf, Fifo, Device);
	    FifoFile = strdup(buf);

	    if ((stat(FifoFile, &st) && errno == ENOENT) ||
		!S_ISFIFO(st.st_mode))
	    {
		unlink(FifoFile);
		if (mkfifo(FifoFile,FifoPerm) != 0) {
		    logmesg("cannot make FIFO");
		    return EXIT_FAILURE;
		}
	    }

	    chmod(FifoFile,FifoPerm);	/* make sure it is accessible... */
	    chown(FifoFile,0,0);	/* ... to root only */
	}

#ifdef DEBUG
	if (opendebug() != SUCCESS)
	    return EXIT_FAILURE;

	debug3(D_OPT, " debug    = (%o)\n", Debug);
	debug3(D_OPT, " defname  = (%s)\n", DefName);
	debug3(D_OPT, " NoHangUp = (%s)\n", (NoHangUp) ? "TRUE" : "FALSE");
	debug3(D_OPT, " LockPoll = (%s)\n", (LockPoll) ? "TRUE" : "FALSE");
	debug3(D_OPT, " TimeOut  = (%d)\n", TimeOut);
	debug3(D_OPT, " line     = (%s)\n", Device);
# ifdef SETTERM
	debug3(D_OPT, " type     = (%s)\n", Term);
# endif /* SETTERM */
#endif	/* DEBUG */

	/* get uid/gid of UUCP user to use for lockfiles and line.
	 * when not found, 0/0 will be used instead.
	 */
	if ((pwd = getpwnam(UUCPNAME)) != NULL) {
	    UuUid = pwd->pw_uid;
	    UuGid = pwd->pw_gid;
	}

	/* close the stdio descriptors, we will not use them
	 * because there are problems with re-opening
	 */
	fclose(stdin);
	fclose(stdout);
	fclose(stderr);

	debug2(D_RUN, "check for lockfiles\n");

	/* deal with the lockfiles; we don't want to charge
	 * ahead if uucp, kermit or something is already
	 * using the line.
	 */

	/* name the lock file(s)
	 */
#ifdef SVR4LOCK
	if (stat(DevName,&st)) {
	    logmesg("cannot stat line");
	    return EXIT_FAILURE;
	}
	sprintf(buf, LOCK, (unsigned)major(st.st_dev),
			   (unsigned)major(st.st_rdev),
			   (unsigned)minor(st.st_rdev));
#else
	sprintf(buf, LOCK, Device);
#endif
	Lock = strdup(buf);
	debug3(D_LOCK, "Lock = (%s)\n", Lock);

	atexit(rmlocks);			/* remove lockfile when exit */

	/* lock the line for our use
	 */
	while (makelock(Lock) != SUCCESS)
	    sleep(15);

	chmod(DevName, 0660);
	chown(DevName, 0, 0);

	/* make sure nobody has the tty still opened
	 */
	if (openline() == SUCCESS) {
	    signal(SIGHUP, SIG_IGN);
	    debug2(D_RUN, "vhangup\n");
	    vhangup();
	    closeline();
	    signal(SIGHUP, SIG_DFL);
	}

	/* the line is mine now ...
	 */

	if (openline() != SUCCESS) {
	    closeline();
	    rmlocks();
	    if (getppid() == 1)			/* called from INIT? */
		sleep(300);			/* prevent rapid respawn */

	    return EXIT_FAILURE;
	}

	if (EpaInit)				/* dump EPA result from */
	    dump_epa();				/* previous session? */

	/* branch back here when a call has been handled and
	 * we don't need to restart the program for some reason
	 */
re_init:
	/* reset and initialize the MODEM
	 */
#ifdef DEBUG
	linestat();
#endif
	if (initmodem(!NoHangUp)) {
	    send("ATZ\r");
	    closeline();
	    rmlocks();

	    if (getppid() == 1)			/* called from INIT? */
		sleep(300);			/* prevent rapid respawn */

	    return EXIT_FAILURE;
	}

	NeedInit = FALSE;			/* no need for re-init (yet) */

	/* allow uucp to access the device
	 */
	debug2(D_RUN, "chmod and chown to uucp\n");
	chmod(DevName, 0660);
	chown(DevName, UuUid, UuGid);

#ifdef LOGUTMP
	/* create UTMP entry when we were started by INIT
	 */
	if (getppid() == 1) {			/* only when from INIT */
	    struct utmp ut,*utmp;
	    pid_t pid;

	    debug2(D_RUN, "update utmp/wtmp files\n");

	    pid = getpid();
	    setutent();

	    while ((utmp = getutent()) != NULL)
		if (utmp->ut_type == INIT_PROCESS && utmp->ut_pid == pid)
		    break;

	    if (utmp != NULL)
		memcpy(&ut,utmp,sizeof(ut));
	    else {
		memset(&ut,0,sizeof(ut));
		strncopy(ut.ut_id,Device + 3);
	    }

	    endutent();

	    /* show login process in utmp
	     */
	    strncopy(ut.ut_line, Device);
	    strncopy(ut.ut_user, "LOGIN");
	    ut.ut_type = LOGIN_PROCESS;
	    ut.ut_time = time(NULL);
	    ut.ut_pid = pid;

	    setutent();
	    pututline(&ut);
	    endutent();
	    debug2(D_UTMP, "logutmp entry made\n");

	    /* write same record to end of wtmp
	     * if wtmp file exists
	     */
	    if (access(WTMP_FILE, W_OK) == 0 &&
		(fp = fopen(WTMP_FILE, "a")) != NULL)
	    {
		fseek(fp,0L,SEEK_END);
		fwrite((char *)&ut,sizeof(ut),1,fp);
		fclose(fp);
		debug2(D_UTMP, "logwtmp entry made\n");
	    }
	}
#endif	/* LOGUTMP */

	/* sound a short beep to indicate we are ready,
	 * optionally play a file as well.
	 */
	if (Voice) {
	    beep(Ready);
	    if (play("init",NULL) == SUCCESS)	/* check if "init" exists */
		if (Ready != NULL || set_vls(16) == SUCCESS)
		    play("init","#bdl");
	}

	signal(SIGINT, inthandler);
#ifdef DEBUG
	signal(SIGUSR1, sighandler);		/* for testing: USR1=RING */
#endif

wait_again:
	rmlocks();				/* allow others to get line */

	/* wait for incoming calls or data on application FIFO
	 */
	switch (wait_work(WFclass))
	{
	case FIFO:				/* data on application FIFO */
	    debug2(D_RUN, "data on FIFO - api_mode\n");

	    if ((c = api_mode()) != LOCKED)	/* handle it */
	    {
		close(FifoFd);
		debug3(D_RUN, "api_mode ends, result=%d\n",c);

		if (c == SUCCESS)
		    break;			/* OK, just loop around */

		send("ATZ\r");
		return EXIT_FAILURE;		/* some mishap, restart */
	    }

	    close(FifoFd);
	    /* it was LOCKED */
	case LOCKED:				/* someone else locked it */
	    closeline();			/* close the line */
	    debug2(D_RUN, "line locked now\n");

	    /* closing the line still leaves it as controlling terminal :-( */
	    /* it seems best to just exit, and take care of the lock at */
	    /* startup only, where we are in a new process group */

	    return EXIT_SUCCESS;		/* INIT will respawn us */

	case INITFAIL:				/* initialization failed */
	    send("ATZ\r");
	    closeline();

	    if (getppid() == 1)			/* called from INIT? */
		sleep(300);			/* prevent rapid respawn */
	    /* just as FAIL */
	case FAIL:				/* got some garbage */
	    return EXIT_FAILURE;		/* re-init from start */

	case VCON:				/* DATA/VOICE in FCLASS=8 */
						/* or BUTTON on TA-128 */
	    debug3(D_RUN, "BUTTON action: %d\n", ButtonAction);

	    switch (ButtonAction)
	    {
	    case BUTTON_ANSWER:			/* answer as if it ringed */
		if (read_state(0,0) != SUCCESS)
		    Mode = DATAFAX;		/* default to DATA/FAX mode */

		goto ring;			/* continue as if a RING */

	    case BUTTON_PLAYB:			/* playback messages */
		/* try to lock the line, we don't want others to use it
		 */
		if (makelock(Lock) == FAIL) {
		    closeline();		/* close the line */
		    return EXIT_SUCCESS;	/* INIT will respawn us */
		}

		if (set_voice(16))		/* select voice mode & speaker */
		    return EXIT_FAILURE;

		playback(NULL,NULL);		/* play all messages */

		if (set_vls(0))
		    return EXIT_FAILURE;

		rmlocks();			/* release the lock */
		break;

	    case BUTTON_EPADMP:			/* dump the EPA contents */
		dump_epa();
		break;
	    }
	    break;

	case RING:
ring:
	    if (!Is_ZyXEL)
		usleep(500000);			/* some small delay after RING */

	    answer = ANSWER;
	    switch (mode = Mode)		/* state file already read! */
	    {
	    case DATA:
		answer = ANSWERDATA;		/* answer in DATA-only mode */
	    case DATAFAX:			/* answer in DATA/FAX mode */
		switch (datamode(answer))
		{
		case VOICE:			/* was a VOICE call? */
		    if (Voice) {
			hold_state(120,VOICE,2);
			answer = ANSWER;
			goto voice;
		    }
		    break;

		case FAX:			/* was a FAX call */
		    if (Fax && faxmode(ONLINE) != SUCCESS)
			hold_state(120,FAX,2);
		    break;

		case POTSCALL:			/* POTS call on ISDN? */
		    waitfor("FLUSH",SHORTTIME);
		    
		    if (!NeedInit)		/* no re-init needed? */
			goto wait_again;
		    break;			/* do re-init */

		case VCON:			/* voice call in Linux ISDN */
		    if (Voice) {
			answer = ONLINE;
			goto voice;
		    }
		    break;

		case RING:			/* another ring? */
		case FAIL:			/* some real failure? */
		    return EXIT_FAILURE;	/* play safe and restart it */
		}
		break;

	    case FAX:
		if (faxmode(ANSWER) == DATA)	/* FAX-only answer */
		    if (datamode(ONLINE) == FAIL) /* it still connected DATA! */
			return EXIT_FAILURE;
		break;

	    case VOICE:
voice:		/* try voicemode */
		switch (voicemode(answer,(Message != NULL? Message:Greeting)))
		{
		case SILENCE:
		    if (mode != VOICE)		/* here from another mode? */
			break;
		    /* silence in VOICE mode, try DATA connection */
		case DATA:
		    /* DATA requested, continue in datamode */
		    /* no FAX in this case */

		    switch (datamode(ANSWERDATA)) /* answer in DATA-only mode */
		    {
		    case SUCCESS:		/* worked? */
			break;

		    case FAIL:			/* some real failure? */
			return EXIT_FAILURE;	/* play safe and restart it */

		    default:			/* NO CARRIER etc */
			hold_state(120,DATA,2); /* remain in DATA mode */
			break;
		    }
		    break;

		case FAX:
		    if (Fax && faxmode(ANSWER) != SUCCESS)
			hold_state(120,FAX,2);
		    break;

		case DIALBACK:			/* have dialled back to user */
		    if (voicemode(ONLINE,Message) != FAIL) /* handle voice */
			break;			/* but ignore DATA and FAX */

		case FAIL:			/* some real failure? */
		    return EXIT_FAILURE;	/* play safe and restart it */
		}
		break;

	    case MESSAGE:
		switch (messagemode(answer,(Message != NULL? Message:Greeting)))
		{
		case SUCCESS:
		    break;

		case FAIL:			/* some real failure? */
		    return EXIT_FAILURE;	/* play safe and restart it */
		}
		break;
	    }
	    break;

	case NORING:				/* not enough rings */
	    if (Mode == NOANSWER)		/* safe to just wait again? */
		goto wait_again;

	    break;				/* modem settings changed! */

	case EVENT:				/* event reported (2864) */
	    switch (Match[1])
	    {
	    case 'H':				/* local phone off hook */
		switch (localvoice())
		{
		case NOLOCAL:
		    goto wait_again;		/* wait for next event */

		case FAIL:
		    return EXIT_FAILURE;	/* failure, restart */
		}
		break;				/* else just re-init */

	    case 'h':				/* local phone on hook */
	    case 'L':				/* loop current pol. reversal */
	    case 'l':				/* loop current interruption */
		goto wait_again;

	    default:
		debug3(D_RUN, "Unexpected event 0x%02x", Match[1]);
		break;
	    }
	    break;

	case CONNECT:
	   /* we can enter here when DATA/VOICE is pressed
	    *	while MODEM is in FCLASS other than 8
	    */
	    switch (datamode(ONLINE))		/* already CONNECTed */
	    {
	    case FAX:				/* was a FAX call */
		if (faxmode(ONLINE) == FAIL)
		    return EXIT_FAILURE;
		break;

	    case POTSCALL:			/* button pressed */
		waitfor("FLUSH",SHORTTIME);
		    
		if (!NeedInit)			/* no re-init needed? */
		    goto wait_again;
		break;				/* do re-init */

	    case FAIL:				/* some real failure? */
		return EXIT_FAILURE;		/* play safe and restart it */
	    }
	    break;
	}

	if (makelock(Lock) != SUCCESS) {	/* acquire a new lock */
	    debug2(D_RUN, "line locked now\n");
	    return EXIT_SUCCESS;
	}

	goto re_init;				/* and do it all over again */
}


/*
**	exit_usage() - exit with usage display
*/

void
exit_usage(code)
int code;
{
	FILE *fp = stderr;

	/* when not started from INIT, just write to stderr */

	if (getppid() != 1 || (fp = fopen(CONSOLE, "w")) != NULL) {
		fprintf(fp, "\n%s\n\n",Id + 4);
		fprintf(fp, "This program is copyrighted by:\n");
		fprintf(fp, "\tR.E. Janssen\n");
		fprintf(fp, "\tNormandie 173\n");
		fprintf(fp, "\t3524 RJ   UTRECHT, NETHERLANDS\n\n");
		fprintf(fp, "Permission is granted only for noncommercial use.  Any commercial use or\n");
		fprintf(fp, "distribution via commercial services without the author's permission is\n");
		fprintf(fp, "prohibited.\n\n");
#ifdef SETTERM
		fprintf(fp, "Usage: %s [options] line speed [term]\n",MyName);
#else
		fprintf(fp, "Usage: %s [options] line speed\n",MyName);
#endif	/* SETTERM */
		fprintf(fp, "Options:\n");
		fprintf(fp, "\t-D<dbglvl>   set debugging level (octal)\n");
		fprintf(fp, "\t-d<defaults> set defaults file\n");
		fprintf(fp, "\t-h           don't hangup (DTR-toggle) during init\n");
		fprintf(fp, "\t-l           check for lockfiles even when no input\n");
		fprintf(fp, "\t-n           don't open line with NONBLOCK (require DCD)\n");
		fprintf(fp, "\t-t<timeout>  timeout (seconds) at login: prompt\n");
		fclose(fp);
	}
	exit(code);
}



/*
**	getvalues() - read all runtime values from /etc/defaults
*/

void
getvalues()

{
	FILE *fp;
	register DEF **def;
	char *p;
	int i;
	char buf[MAXLINE+1];

	def = defbuild(DefName);

#ifdef DEBUG

	/* debugging on?
	 */
	if ((p = defvalue(def, "DEBUG")) != NULL)
		sscanf(p, "%o", &Debug);

#endif	/* DEBUG */

	Init = defvalue(def, "INIT");
#ifdef TIOCMGET
	if ((p = defvalue(def, "IGNOREDSR")) != NULL && !strcmp(p,"YES")) {
	    Modemlines &= ~TIOCM_DSR;
	    Modemln_ok &= ~TIOCM_DSR;
	}
#endif
	if ((p = defvalue(def, "HANGUP")) != NULL)
	    NoHangUp = !strcmp(p, "NO");
	if ((p = defvalue(def, "LOCKPOLL")) != NULL)
	    LockPoll = !strcmp(p, "YES");
	if ((p = defvalue(def, "MODEMPOLL")) != NULL)
	    ModemPoll = !strcmp(p, "YES");
	if ((p = defvalue(def, "WFCLASS")) != NULL)
	    WFclass = atoi(p);
	if ((p = defvalue(def, "MUTE")) != NULL)
	    Mute = p;
	DialPrefix = defvalue(def, "DIALPREFIX");
	if ((p = defvalue(def, "DIALTIME")) != NULL)
	    DialTime = atoi(p);
	if ((p = defvalue(def, "DIALTRIES")) != NULL)
	    DialTries = atoi(p);
	Fifo = defvalue(def, "FIFO");
	if ((p = defvalue(def, "FIFOPERM")) != NULL) {
	    sscanf(p, "%o", &i);
	    FifoPerm = i;
	}
	if ((p = defvalue(def, "STATEFILE")) != NULL)
	    StateFile = p;
	StateProg = defvalue(def, "STATEPROG");
	CndFile = defvalue(def, "CNDFILE");
	for (i = 0; i < 8; i++) {
	    sprintf(buf,"RING%d",i + 1);
	    RingType[i] = defvalue(def, buf);
	}
	if ((EpaFile = defvalue(def, "EPADUMP")) != NULL && strchr(EpaFile,'%') != NULL) {
	    sprintf(buf,EpaFile,Device);
	    EpaFile = strdup(buf);
	}
	if ((p = defvalue(def, "EPAINIT")) != NULL)
	    EpaInit = atoi(p);
	if ((SysName = defvalue(def, "SYSTEM")) == NULL)
	    SysName = getuname();
	if ((Version = defvalue(def, "VERSION")) != NULL)
	    if (*Version == '/') {
		if ((fp = fopen(Version, "r")) != NULL) {
		    if (fgets(buf, sizeof(buf), fp) != NULL)
		       buf[strlen(buf)-1] = '\0';
		    else
		       buf[0] = '\0';
		    fclose(fp);
		    Version = strdup(buf);
		}
	    }
	if ((p = defvalue(def, "MAILER")) != NULL)
	    Mailer = p;
	if ((p = defvalue(def, "XOFF")) != NULL && !strcmp(p,"NO"))
	    FinalXoff = FALSE;
	if ((p = defvalue(def, "LOGIN")) != NULL)
	    Login = p;
	if ((p = defvalue(def, "LOGINENV")) != NULL && !strcmp(p,"NO"))
	    LoginEnv = FALSE;
	LoginFlags = defvalue(def, "LOGINFLAGS");
#ifdef FIDO
	Fido = defvalue(def, "FIDO");
#endif	/* FIDO */
	Protect = defvalue(def, "PROTECT");
	DataProg = defvalue(def, "DATAPROG");
	DataFlags = defvalue(def, "DATAFLAGS");
	PPPProg = defvalue(def, "PPPPROG");
	PPPFlags = defvalue(def, "PPPFLAGS");

	if ((p = defvalue(def, "ISSUE")) != NULL)
	    Issue = p;
	if ((p = defvalue(def, "TIMEOUT")) != NULL)
	    TimeOut = atoi(p);

	if ((p = defvalue(def, "FAX")) != NULL && !strcmp(p,"NO"))
	    Fax = FALSE;
	if ((FaxDir = defvalue(def, "FAXDIR")) != NULL && strchr(FaxDir,'%') != NULL) {
	    sprintf(buf,FaxDir,Device);
	    FaxDir = strdup(buf);
	}
	if ((p = defvalue(def, "FAXFILE")) != NULL)
	    FaxFile = p;
	FaxId = defvalue(def, "FAXID");
	FaxMail = defvalue(def, "FAXMAIL");
#ifndef ZFAX
	if ((p = defvalue(def, "FAXCOMP")) != NULL)
	    FaxComp = atoi(p);
	if ((p = defvalue(def, "FAXPOLL")) != NULL)
	    FaxPoll = p;
	if ((p = defvalue(def, "FAXREJ")) != NULL)
	    FaxRej = p;
	if ((p = defvalue(def, "GHOSTSCRIPT")) != NULL)
	    GhostScript = p;
	if ((p = defvalue(def, "GUNZIP")) != NULL)
	    Gunzip = p;
	if ((p = defvalue(def, "BUFFER")) != NULL)
	    Buffer = p;
	if ((p = defvalue(def, "PSTEXT")) != NULL)
	    PsText = p;
#endif

	if ((p = defvalue(def, "APIUSERLEVEL")) != NULL)
	    ApiUserLevel = atoi(p);
	if ((p = defvalue(def, "GREETING")) != NULL) {
	    if (strchr(p,'%') != NULL) {
		sprintf(buf,p,Device);
		Greeting = strdup(buf);
	    } else
		Greeting = p;
	}
	PlayDir = defvalue(def, "PLAYDIR");
	if ((RecDir = defvalue(def, "RECDIR")) != NULL && strchr(RecDir,'%') != NULL) {
	    sprintf(buf,RecDir,Device);
	    RecDir = strdup(buf);
	}
	if ((Archive = defvalue(def, "ARCHIVE")) != NULL && strchr(Archive,'%') != NULL) {
	    sprintf(buf,Archive,Device);
	    Archive = strdup(buf);
	}
	if ((p = defvalue(def, "RECFILE")) != NULL)
	    RecFile = p;
	if ((p = defvalue(def, "RECMODE")) != NULL)
	    RecMode = atoi(p);
	if ((p = defvalue(def, "CHECKLINE")) != NULL)
	    CheckLine = atoi(p);
	if ((p = defvalue(def, "DROPMSG")) != NULL)
	    DropMsg = atoi(p);

	DtmfDet = defvalue(def, "DTMFDET");
	LocalKey = defvalue(def, "LOCALKEY");
#ifdef WITH_SOUND
	SoundDev = defvalue(def, "SOUNDDEV");
#endif
	MidPlay = defvalue(def, "MIDPLAY");
	MidPlayFlags = defvalue(def, "MIDPLAYFLAGS");
	ModPlay = defvalue(def, "MODPLAY");
	ModPlayFlags = defvalue(def, "MODPLAYFLAGS");
	Mp3Play = defvalue(def, "MP3PLAY");
	Mp3PlayFlags = defvalue(def, "MP3PLAYFLAGS");
	if ((p = defvalue(def, "SILENCE")) != NULL)
	    Silence = p;
	Synth = defvalue(def, "SYNTH");
	SynthFlags = defvalue(def, "SYNTHFLAGS");
	VoiceCmd = defvalue(def, "VOICECMD");
	VoiceKey = defvalue(def, "VOICEKEY");
	VoiceRing = defvalue(def, "VOICERING");

	if ((p = defvalue(def, "SHOWMSG")) != NULL) {
	    if (!strcmp(p,"AA"))
		ShowMsg = SHOWMSG_AA;
	    if (!strcmp(p,"DSR"))
		ShowMsg = SHOWMSG_DSR;
	}
	if ((p = defvalue(def, "BUTTON")) != NULL) {
	    if (!strcmp(p,"ANSWER"))
		ButtonAction = BUTTON_ANSWER;
	    if (!strcmp(p,"PLAYBACK"))
		ButtonAction = BUTTON_PLAYB;
	    if (!strcmp(p,"EPADUMP"))
		ButtonAction = BUTTON_EPADMP;
	}

	AnswerTone = defvalue(def, "ANSWERTONE");
	Error = defvalue(def, "ERROR");
	MsgArch = defvalue(def, "MSGARCH");
	MsgDel = defvalue(def, "MSGDEL");
	Prompt1 = defvalue(def, "PROMPT1");
	Prompt2 = defvalue(def, "PROMPT2");
	Ready = defvalue(def, "READY");
	RecAttn = defvalue(def, "RECATTN");
	RecDone = defvalue(def, "RECDONE");

	PagerCentral = defvalue(def, "PAGERCENTRAL");
	PagerNum = defvalue(def, "PAGERNUM");
	if ((p = defvalue(def, "PAGERTYPE")) != NULL)
	    PagerType = atoi(p);
	if ((p = defvalue(def, "PAGERRECMSG")) != NULL)
	    PagerRecMsg = p;
	if ((p = defvalue(def, "PAGERSOURCE")) != NULL)
	    PagerSource = p;
}

/*
**	initmodem() - initialize MODEM for use by this program
*/

int
initmodem(dropdtr)
int dropdtr;

{
	char *p,*q;
#ifdef TIOCMGET
	int mstat;
#endif

	debug2(D_RUN, "perform MODEM initialization\n");

	signal(SIGINT, SIG_IGN);		/* it will trip when switched */

	/* toggle DTR
	 * when AT&D3 is still active this will reset the MODEM
	 */
	if (dropdtr) {
	    usleep(100000);			/* keep DTR on a while */
#ifdef CLOSE_OPEN
	    closeline();
	    usleep(500000);			/* keep DTR off a while */
	    if (openline() != SUCCESS)
		return FAIL;
#else
	    settermios(DROPDTR);
	    usleep(500000);			/* keep DTR off a while */
#endif
	}

	/* make sure the line is properly initialized
	 */
	settermios(INITIAL);

	/* flush any pending garbage & allow MODEM to recover from DTR drop
	 */
	waitfor("FLUSH",SHORTTIME);

	/* now try to RESET the MODEM
	 */
	send("ATZ\r");

	if (waitfor("OK",MEDMTIME)) {
	    /* no reaction to ATZ
	     * first attempt to end voice mode (DLE-ETX)
	     */
	    send(AudioAbort);
	    waitfor("FLUSH",SHORTTIME);

	    /* exit from data mode, when it is stuck in a mode that does
	     * not honour the DTR drop.	 this is done using:
	     * - sending Q to escape from EPA mode
	     * - sending +++ with delays around it (escape to command mode)
	     * - sending XON to allow dataflow in XON/XOFF mode
	     * - sending ATCM0 to end CAPI mode
	     * - sending ATZ to reset the modem
	     */
	    send("Q");
	    send("\\d+++\\d");
	    send("\\021\\p");
	    send("ATCM0\r");
	    waitfor("FLUSH",SHORTTIME);
	    send("ATZ\r");

#ifdef TIOCMGET
	    /* make sure the MODEM is present
	     * do this by checking CTS, the only line which is
	     * guaranteed to be active
	     */
	    if ((mstat = modemstat()) != -1 &&
		!(mstat & TIOCM_CTS) &&
		(sleep(3),!(modemstat() & TIOCM_CTS)))
	    {
		waitfor("FLUSH",SHORTTIME);	/* flush OK */

		debug2(D_RUN, "CTS is OFF\n");
		logmesg("MODEM not switched ON, waiting");

		while (!(modemstat() & TIOCM_CTS))
		    sleep(5);

		debug2(D_RUN, "CTS is ON\n");
		sleep(5);
		logmesg("MODEM switched ON, continuing");

		send("ATZ\r");
	    }
#endif

	    if (waitfor("OK",MEDMTIME)) {
		logmesg("MODEM RESET (ATZ) failed");
		debug2(D_RUN, "MODEM RESET (ATZ) failed -- aborting\n");
		return FAIL;
	    }
	}

	waitfor("FLUSH",SHORTTIME);			/* skip all garbage */
	Fclass = 0;

	if (ProdInfo == NULL) {
	    /* check that the MODEM is a ZyXEL U-1496 or 2864
	     * don't want to be responsible for destroying other MODEM's
	     * config..
	     */
	    send("ATI\r");
	    if (waitfor("OK",MEDMTIME)) {
		logmesg("MODEM Product Info inquiry (ATI) failed");
		debug2(D_RUN, "MODEM Product Info inquiry (ATI) failed -- aborting\n");
		return FAIL;
	    }

	    p = ChatBuf;

	    if (!strncmp(p,"AT",2)) {			/* skip the echo */
		if ((p = strchr(p,'\r')) == NULL)
		    p = ChatBuf;
		else
		    while (*p == '\r' || *p == '\n')
			p++;				/* skip to next line */
	    }

	    rip(p);					/* discard CR etc */
	    ProdInfo = strdup(p);			/* save prod. info */

	    if (!strncmp(p,"U1496",5)) {		/* really old U1496? */
		Is_ZyXEL = Voice = TRUE;
		goto prodinfo;				/* evaluate as ATI1 */
	    }

	    /* it looks like they can't make up their mind...  there are
	     * at least 3 different product info strings for the 2864!
	     */
	    if (!strcmp(p,"1496") ||
		!strcmp(p,"2864") ||
		!strcmp(p,"28641") || !strcmp(p,"28642") || !strcmp(p,"28643") ||
		!strncmp(p,"Elite 2864",10) ||
		!strcmp(p,"1281") || !strcmp(p,"1282") || !strcmp(p,"1283"))
	    {
		Is_ZyXEL = Voice = TRUE;		/* looks like ZyXEL */
	    } else {
#ifdef __linux__
		if (!strcmp(p,"Linux ISDN")) {		/* Linux kernel ISDN */
		    Voice = TRUE;
		    Fsamp = 8000;
		    LSOn = Mute = "";
# ifdef TIOCMGET
		    Modemlines &= ~TIOCM_DSR;
		    Modemln_ok &= ~TIOCM_DSR;
# endif
		    Fw_version = 100;
		    Fw_features = FW_LINUX | FW_ISDN | FW_XLAW | FW_DTMFDET |
		    		  FW_NEWADPCM | FW_NEWLEAK;
		} else
#endif
		    if (!No_ZyXEL) {
			sprintf(MsgBuf, "MODEM returns Product Info \"%s\", not a ZyXEL", p);
			logmesg(MsgBuf);
			debug3(D_RUN, "%s -- aborting\n",MsgBuf);
			return NOZYXEL;
		    }
	    }
	}

	/* get the firmware version, when not yet known
	 */
	if (Is_ZyXEL && Fw_version == 0) {
	    send("ATI1\r");
	    if (waitfor("OK",LONGTIME)) {
		logmesg("MODEM Version inquiry (ATI1) failed");
		debug2(D_RUN, "MODEM Version inquiry (ATI1) failed -- aborting\n");
		return FAIL;
	    }

	    p = ChatBuf;

	    if (!strncmp(p,"AT",2)) {			/* skip the echo */
		if ((p = strchr(p,'\r')) == NULL)
		    p = ChatBuf;
		else
		    while (*p == '\r' || *p == '\n')
			p++;				/* skip to next line */
	    }

	    while (*p != '\0' && !isalpha(*p)) {	/* skip the checksum */
		if ((p = strchr(p,'\r')) == NULL)
		    break;
		else
		    while (*p == '\r' || *p == '\n')
			p++;				/* skip to next line */
	    }

prodinfo:
	    switch ((p != NULL)? *p : 0)
	    {
	    case 'U':
		/* it looks like a U-1496, which returns a line like:
		 * U1496E  V 6.15 P  ^M
		 */

		rip(p);					/* take 1 line */

		if ((q = strchr(p,' ')) != NULL) {	/* get product info */
		    *q = '\0';
		    free(ProdInfo);
		    ProdInfo = strdup(p);		/* more accurate info */
		    *q = ' ';
		}

		if ((q = strchr(p,'V')) != NULL) {	/* find V x.xx [P] */
		    while (*q != '\0' && !isdigit(*q))
			q++;

		    Fw_version = atoi(q) * 100;		/* major version */

		    while (isdigit(*q))
			q++;

		    if (*q == '.' || *q == '-')
			Fw_version += atoi(++q);	/* add minor version */
		    else
			if (Fw_version > 9999)		/* beta version, no '.' */
			    Fw_version /= 100;

		    while (isdigit(*q))
			q++;

		    if (isalpha(*q))			/* beta version? */
			Fw_beta = *q;

		    if (strchr(q,'P') != NULL)
			Fw_features |= FW_PLUS;		/* PLUS version */
		}

		/* evaluate version number and set firmware feature bits */

		if (Fw_version >= 610) {
		    Fw_features |= FW_DTMFREC;		/* can record DTMF-only */
		    Fw_features |= FW_RESYNC;		/* can do audio resync */
		}

		if (Fw_version >= 611) {
		    Fw_features |= FW_AUDIOABORT;	/* can abort AUDIO playback */
		    Fw_features |= FW_DIALTONE;		/* can detect DIALTONE */
		}

		if (Fw_version >= 612)
		    Fw_features |= FW_DTMFDET;		/* can set DTMF sens */

		if (Fw_version >= 613) {
		    Fw_features |= FW_RTSCTSFLOW;	/* can RTS/CTS AUDIO/FAX */
		    Fw_features |= FW_NEWADPCM;		/* ADPCM3N/ADPCM4 recording */
		    Fw_features |= FW_NEWLEAK;		/* new LEAKAGE coeff */
		}

		if (Fw_version >= 614) {
		    Fw_features |= FW_FAXEC;		/* can do FAX EC */
		    Fw_features |= FW_VBTI;		/* can set VBT interval */
#if 0 /* this feature seems problematic */
		    Fw_features |= FW_DTMFPAUSE;	/* PAUSE when DTMF received */
#endif
		}

		Fw_features |= FW_PLAYBUG;		/* for now, assume all have it */
		break;

	    case 'E':
		/* it looks like an Elite 2864, which returns lines like:
		 * E2864  V 1.05
		 * or:
		 * Elite 2864I DSS1: v1.06
		 * Internal fax/modem:	V 1.05
		 */

		Fw_features |= FW_2864;

		if (strchr(p,'I') != NULL)
		    Fw_features |= FW_ISDN;

		if (p[1] == 'l') {			/* "Elite 2864" style */
		    if ((q = strchr(p,':')) != NULL) {	/* get product info */
			*q = '\0';
			free(ProdInfo);
			ProdInfo = strdup(p);		/* more accurate info */
			*q = ':';
		    } else {
			if ((q = strchr(p,'\r')) != NULL) {
			    *q = '\0';
			    free(ProdInfo);
			    ProdInfo = strdup(p);	/* more accurate info */
			    *q = '\r';
			}
		    }
		} else {				/* else "E2864" style */
		    if ((q = strchr(p,' ')) != NULL) {	/* get product info */
			*q = '\0';
			free(ProdInfo);
			ProdInfo = strdup(p);		/* more accurate info */
			*q = ' ';
		    }
		}

		/* find V x.xx */

		if ((q = strchr(p,'V')) != NULL || (q = strchr(p,'v')) != NULL) {
		    while (*q != '\0' && !isdigit(*q))
			q++;

		    Fw_version = atoi(q) * 100;		/* major version */

		    while (isdigit(*q))
			q++;

		    if (*q == '.' || *q == '-')
			Fw_version += atoi(++q);	/* add minor version */
		    else
			if (Fw_version > 9999)		/* beta version, no '.' */
			    Fw_version /= 100;

		    while (isdigit(*q))
			q++;

		    if (isalpha(*q))			/* beta version? */
			Fw_beta = *q;
		}

		/* these features are supported by all Elite modems */
		/* FW_RTSCTSFLOW is documented, but does not seem to work */

		Fw_features |= FW_DTMFREC | FW_AUDIOABORT | FW_DIALTONE |
			       FW_DTMFDET | FW_NEWADPCM | FW_NEWLEAK |
			       FW_FAXBUG;

		/* version 1.09 allows enabling speaker during phone line */
		/* recordings. */

		if (Fw_version >= 109)
		    PhoneLS = 16 + 2;			/* enable speaker */

		if (Fw_version >= 209)
		    Fw_features |= FW_CLIR;		/* can do CLIR */

		if (Fw_version >= 211)
		    Fw_features |= FW_FASTC;		/* has ATC$1 */

		break;

	    case 'S':
	    case 'O':
		/* It looks like an Omni, which returns a line like:
		 * S288	  V 1.09     ^M
		 * or:
		 * O288S  V 1.12     ^M
		 */

		Fw_features |= FW_2864;

		rip(p);					/* take 1 line */

		if ((q = strchr(p,' ')) != NULL) {	/* get product info */
		    *q = '\0';
		    free(ProdInfo);
		    ProdInfo = strdup(p);		/* more accurate info */
		    *q = ' ';
		}

		if ((q = strchr(p,'V')) != NULL) {	/* find V x.xx */
		    while (*q != '\0' && !isdigit(*q))
			q++;

		    Fw_version = atoi(q) * 100;		/* major version */

		    while (isdigit(*q))
			q++;

		    if (*q == '.' || *q == '-')
			Fw_version += atoi(++q);	/* add minor version */
		    else
			if (Fw_version > 9999)		/* beta version, no '.' */
			    Fw_version /= 100;

		    while (isdigit(*q))
			q++;

		    if (isalpha(*q))			/* beta version? */
			Fw_beta = *q;
		}

		/* these features are supported by all Omni modems */
		/* FW_RTSCTSFLOW is documented, but does not seem to work */

		Fw_features |= FW_DTMFREC | FW_AUDIOABORT | FW_DIALTONE |
			       FW_DTMFDET | FW_NEWADPCM | FW_NEWLEAK |
			       FW_FAXBUG;
		break;

	    case 'Z':
		/* It looks like an Omni TA-128, which returns a line like:
		 * ZyXEL Omni TA128 DSS1: V 2.00
		 */

		if (strncmp(p,"ZyXEL Omni TA128",16))	/* verify it */
		    break;

		Fw_features |= FW_128 | FW_ISDN;

		rip(p);					/* take 1 line */
		p += 6;					/* skip "ZyXEL" */

		if ((q = strchr(p,':')) != NULL) {	/* get product info */
		    *q = '\0';
		    free(ProdInfo);
		    ProdInfo = strdup(p);		/* more accurate info */
		    *q = ':';
		}

		if ((q = strchr(p,'V')) != NULL) {	/* find V x.xx */
		    while (*q != '\0' && !isdigit(*q))
			q++;

		    Fw_version = atoi(q) * 100;		/* major version */

		    while (isdigit(*q))
			q++;

		    if (*q == '.' || *q == '-')
			Fw_version += atoi(++q);	/* add minor version */
		    else
			if (Fw_version > 9999)		/* beta version, no '.' */
			    Fw_version /= 100;

		    while (isdigit(*q))
			q++;

		    if (isalpha(*q))			/* beta version? */
			Fw_beta = *q;
		}

		Fax = Voice = FALSE;			/* it has no FAX/VOICE */
		WFclass = -1;				/* no +FCLASS */

		if (Fw_version >= 205)
		    Fw_features |= FW_CLIR;		/* can do CLIR */

		if (Fw_version >= 207)
		    Fw_features |= FW_FASTC;		/* has ATC$1 */

		break;
	    }
	}

	/* make modem description string
	 */
	sprintf(Modem,"%s%s V %d.%02d",
		(Is_ZyXEL? "ZyXEL " : ""),ProdInfo,
		Fw_version / 100,Fw_version % 100);

	if (Fw_beta)
	    sprintf(Modem + strlen(Modem),"%c",Fw_beta);

	if (Fw_features & FW_PLUS)
	    strcat(Modem," P");

	debug4(D_RUN, "Modem: \"%s\", Fw_features: 0x%04x\n", Modem, Fw_features);

#if 0 /* just do this in the INIT sequence when desired! */
	/* on ISDN model, select fax/modem/voice DTE channel
	 * I read that this feature is going to be dropped in the future,
	 * maybe code will need to be changed a bit (when AT&O0 returns ERROR)
	 */
	if (Fw_features & FW_ISDN) {
	    if (Fw_features & FW_128)
		send("AT&O2\r");
	    else
		send("AT&O0\r");

	    if (waitfor("OK",MEDMTIME)) {
		logmesg("Set DTE channel failed");
		debug2(D_RUN, "Set DTE channel failed -- aborting\n");
		return FAIL;
	    }
	}
#endif

	/* handle init sequence if requested
	 */
	CurVSD[0] = '\0';			/* VSD setting gets reset */
	CurLS = CurVSM = 0;			/* VLS/VSM setting as well */
	CurSync = -1;
	PlayCND = FALSE;			/* Play CND bug workaround */

	if (Init != NULL) {
	    if (chat(Init) == FAIL){
		logmesg("INIT sequence failed");
		debug2(D_RUN, "Modem initialization failed -- aborting\n");
		return FAIL;
	    }
	    waitfor(NULL,SHORTTIME);		/* eat all remaining chars */
	}

#ifdef TIOCMGET
	/* check the MODEM line status to make sure INIT has worked
	 */
	if ((mstat = modemstat()) != -1 && (mstat & Modemlines) != Modemln_ok) {
	    logmesg("Invalid line status after INIT sequence");
	    debug2(D_RUN, "Invalid line status after INIT sequence -- aborting\n");
	    return FAIL;
	}
#endif

	if (Fax) {
#ifndef ZFAX
	    /* send default CLASS 2 FAX parameters
	     */
	    sprintf(ChatBuf,"AT%s%s+FBOR=0+FCR=1+FDCC=1,,0,2,%d,%d,,0\r",
		    make_fclass(2),(Fw_features & FW_RTSCTSFLOW? "+FLO=2":""),
		    FaxComp,((Fw_features & FW_FAXEC)? 1 : 0));
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME) && !(Fw_features & FW_FAXBUG)) {
		logmesg("Set FAX params failed");
		debug2(D_RUN, "Set FAX params failed -- aborting\n");
		return FAIL;
	    }

	    /* not sure if all versions accept this one, so send it separately...
	     */
	    send("AT+FCQ=2\r");
	    waitfor("OK",MEDMTIME);
#endif
	    /* send FAX ID when defined
	     */
	    if (FaxId != NULL) {
#ifdef ZFAX
		sprintf(ChatBuf,"AT#P%.25s\r",FaxId);
#else
		sprintf(ChatBuf,"AT+FLID=\"%.20s\"\r",FaxId);
#endif
		send(ChatBuf);
		if (waitfor("OK",MEDMTIME)) {
		    logmesg("Set FAX ID failed");
		    debug2(D_RUN, "Set FAX ID failed -- aborting\n");
		    return FAIL;
		}
	    }
	}

	/* override "abort audio" sequence to react quicker on modems
	 * that support the DLE-DC4 command to clear transmit buffer
	 */
	if (Fw_features & FW_AUDIOABORT)
	    AudioAbort[0] = AudioAbort[2] = DC4;

	/* init VOICE params
	 * also: when possible with this version, enable RTS/CTS flow
	 * control for VOICE mode (and FAX mode as well)
	 */
	if (Voice && set_voice(0)) {
	    logmesg("Set VOICE params failed");
	    debug2(D_RUN, "Set VOICE params failed -- aborting\n");
	    return FAIL;
	}

	/* for version 6.14 and up, set S55.4 to indicate pausing of play
	 * when a DTMF tone is received
	 * (AudioAbort=DC4 is required in this case)
	 */
	if (Fw_features & FW_DTMFPAUSE) {
	    AudioAbort[0] = AudioAbort[2] = DC4;
	    send("ATS55.4=1\r");
	    if (waitfor("OK",MEDMTIME)) {
		logmesg("Set S55.4=1 failed");
		debug2(D_RUN, "Set S55.4=1 failed -- aborting\n");
		return FAIL;
	    }
	}

	return SUCCESS;
}

/*
**	openline() - open line to STDIN/STDOUT/STDERR
*/

int
openline(void)

{
	int fd;
	int mode;

	debug2(D_RUN, "open stdin, stdout and stderr\n");

	/* open the line; don't wait around for carrier-detect
	 */
	mode = O_RDWR;
	if (!No_nonblock)
	    mode |= O_NONBLOCK;
	if ((fd = open(DevName,mode)) != STDIN) {
	    logmesg("cannot open line");
	    return FAIL;
	}

	/* make stdout and stderr, too
	 */
	if ((fd = dup2(STDIN,STDOUT)) != STDOUT) {
	    debug4(D_RUN, "dup stdout: %d %d\n", fd, errno);
	    logmesg("cannot open stdout");
	    return FAIL;
	}
	if ((fd = dup2(STDIN,STDERR)) != STDERR) {
	    debug4(D_RUN, "dup stderr: %d %d\n", fd, errno);
	    logmesg("cannot open stderr");
	    return FAIL;
	}

	/* setup tty characteristics
	 */
	debug3(D_RUN, "setup tty characteristics (%d baud)\n", Speed);
#ifdef __linux__
	mode = N_TTY;				/* select TTY discipline */
	ioctl(STDIN,TIOCSETD,&mode);
#endif
	settermios(INITIALIZE);

	/* clear O_NONBLOCK flag now, we need it for blocking reads
	 */
	if (!No_nonblock) {
	    int flags;

	    if ((flags = fcntl(STDIN, F_GETFL, 0)) >= 0)
		fcntl(STDIN, F_SETFL, flags & ~O_NONBLOCK);
	}

	/* create a new process group (maybe INIT does it?)
	 */
	if (getppid() == 1) {			/* only when from INIT */
#ifdef TIOCSPGRP
	    int rv;
	    pid_t pid;
#endif

	    setsid();				/* become session leader */
	    setpgrp();				/* new process group */

#ifdef TIOCSPGRP
	    pid = getpid();
	    if ((rv = ioctl(STDIN, TIOCSPGRP, &pid)) < 0) {
		debug5(D_RUN, "TIOCSPGRP returns %d errno=%d pid=%d\n", rv, errno, pid);
		logmesg("cannot set process group");
		return FAIL;
	    }
#endif
	}

	return SUCCESS;
}

/*
**	closeline() - close the line (STDIN/STDOUT/STDERR)
*/

void
closeline(void)

{
#ifdef TIOCNOTTY
	unsigned int arg;
#endif

	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);

#ifdef TIOCNOTTY
	ioctl(STDIN,TIOCNOTTY,&arg);		/* no controlling tty */
#endif

	close(STDIN);
	close(STDOUT);
	close(STDERR);
}


/*
**	wait_work() - wait for data from MODEM
*/

int
wait_work(fclass)
int fclass;					/* which +FCLASS to use */

{
	int sel,res;
	char *p;
#ifdef DEBUG
	time_t now;
#endif
	int messages = 0;			/* messages indicated on AA? */
	int ringtype = 0;			/* (distinctive) ring type */
#ifndef ZFAX
	int faxpoll = 0;			/* fax available for poll? */
#endif
#ifdef TIOCMGET
	int modemln_ok = Modemln_ok;		/* correct state of lines */
#endif
	int maxfd;
	fd_set select_set;			/* which fd's we have to select() on */
	fd_set read_set;			/* copy to be used for read */
	struct timeval timeval;			/* timeout for select() */
	struct termios termios;

again:
	if (ShowMsg)
	    messages = filespresent(RecDir);	/* messages recorded? */

#ifndef ZFAX
	if (FaxPoll != NULL) {
	    sprintf(ChatBuf,FaxPoll,"0000");	/* create default doc filename */
	    if (!access(ChatBuf,R_OK))
		faxpoll = 1;			/* we can respond to poll */
	}
#endif

	free(CndName);				/* release CND data */
	CndName = NULL;
	free(CndNumber);
	CndNumber = NULL;
	free(ToNumber);
	ToNumber = NULL;

	/* now prepare the MODEM for incoming calls:
	 * ATH		hangup the line (seems to be required after beeps etc)
	 * S0=99	when messages present, to light AA LED
	 * +FLPL=1	when FAX available for polling
	 * +FCLASS=0	so DATA/VOICE attempts DATA connection and sends
	 *		CONNECT when successful (ORG/ANS also works)
	 * +FCLASS=8	so DATA/VOICE button immediately sends VCON, but
	 *		ORG/ANS cannot be checked, always ANSWER mode
	 */
	strcpy(ChatBuf,"AT");
	if (Voice) {
	    strcat(ChatBuf,"H");
	    CurLS = 0;
	    if (Is_ZyXEL)
		Fclass = 0;
	}
	if (Voice && ShowMsg && messages) {
	    debug2(D_RUN, "messages present in RecDir!\n");

	    switch (ShowMsg)
	    {
	    case SHOWMSG_AA:
		strcat(ChatBuf,"S0=99");	/* AA on, but no answering! */
		break;
	    case SHOWMSG_DSR:
		strcat(ChatBuf,"&S0");		/* DSR on */
#ifdef TIOCMGET
		modemln_ok |= (TIOCM_DSR & Modemlines);	/* take DSR ON as OK later */
#endif
		break;
	    }
	}
	if (fclass >= 0) {
#ifndef ZFAX
	    if (Fax && faxpoll)
		sprintf(ChatBuf + strlen(ChatBuf),"%s+FLPL=1",make_fclass(2));
#endif
	    strcat(ChatBuf,make_fclass(fclass));
	}
	if (ChatBuf[2] != '\0') {		/* any command? (TA-128!) */
	    strcat(ChatBuf,"\r");
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME) && !(Fw_features & FW_FAXBUG))
		return INITFAIL;
	}

	/* setup API FIFO and select() set
	 */
	FD_ZERO(&select_set);			/* all bits off */
	FD_SET(STDIN,&select_set);		/* stdin */
	maxfd = 1;

	if (FifoFile != NULL) {
	    if ((FifoFd = open(FifoFile,O_RDONLY|O_NONBLOCK)) < 0) {
		logmesg("cannot open FIFO");
		debug3(D_RUN, "cannot open FIFO... errno=%d\n", errno);
	    } else {
		int flags;

		if ((flags = fcntl(FifoFd, F_GETFL, 0)) >= 0)
		    fcntl(FifoFd, F_SETFL, flags & ~O_NONBLOCK);

		FD_SET(FifoFd,&select_set);	/* the API FIFO */
		maxfd = FifoFd + 1;
	    }
	}

	ioctl(STDIN,TCFLSH,TCIFLUSH);		/* flush input */
	InPtr = InLen = 0;
	Signal = 0;

	debug2(D_RUN, "waiting for character from MODEM ...\n");

	/* wait for incoming character on MODEM and/or FIFO...
	 */
	do
	{
	    Match = NULL;			/* nothing Matched... */

	    read_set = select_set;
	    timeval.tv_sec = 15;		/* check locks every 15 sec? */
	    timeval.tv_usec = 0;

	    sel = select(maxfd,&read_set,NULL,NULL,
			 ((LockPoll || ModemPoll)? &timeval : NULL));

#ifdef DEBUG
	    /* check for SIGUSR1
	     * simulate a RING response
	     */
	    if (sel < 0 && errno == EINTR && (Signal & (1 << SIGUSR1))) {
		debug2(D_RUN, "SIGUSR1: simulate RING\n");
		strcpy(Match = ChatBuf,"RING\r\n");
		break;
	    }
#endif

	    if (sel < 0) {
		sprintf(MsgBuf,"select failed, errno=%d",errno);
		logmesg(MsgBuf);
		debug3(D_RUN, "%s...\n", MsgBuf);
		return FAIL;
	    }

	    /* check for locks from others
	     */
	    if ((LockPoll || sel > 0) && checklock(Lock)) {
		close(FifoFd);
		return LOCKED;
	    }

	    memset(&termios,0,sizeof(termios));
	    ioctl(STDIN, TCGETS, &termios);	/* check if others changed line */

	    if (memcmp(&termios,&Termios,sizeof(struct termios))) {
		if (!(LockPoll || sel > 0) && checklock(Lock)) {
		    close(FifoFd);
		    return LOCKED;
		}

		logmesg("someone changed termios");
		debug2(D_RUN, "someone changed termios...\n");
		close(FifoFd);
		return FAIL;
	    }

#ifdef TIOCMGET
	    /* make sure the handshake lines are still valid
	     * when not, the MODEM probably was powered off or disconnected...
	     */
	    if ((res = modemstat()) != -1 && (res & Modemlines) != modemln_ok)
	    {
		if (!(LockPoll || sel > 0) && checklock(Lock)) {
		    close(FifoFd);
		    return LOCKED;
		}

		logmesg("lost MODEM while waiting for work");
		debug3(D_RUN, "lost MODEM while waiting for work\n", MsgBuf);
		close(FifoFd);
		return FAIL;
	    }
#endif

	    /* when data available on FIFO, return FIFO and leave it OPEN
	     */
	    if (FifoFd > 0 && FD_ISSET(FifoFd,&read_set)) {
		/* when nothing on stdin, turnoff AA led
		 */
		if (ShowMsg && messages && !FD_ISSET(STDIN,&read_set))
		    switch (ShowMsg)
		    {
		    case SHOWMSG_AA:
			send("ATS0=0\r");	/* AA off */
			if (waitfor("OK",MEDMTIME))
			    return FAIL;
			break;
		    case SHOWMSG_DSR:
			send("AT&S1\r");	/* DSR off */
			if (waitfor("OK",MEDMTIME))
			    return FAIL;
			break;
		    }

		return FIFO;
	    }

	    /* when requested, check if AT still results in a response
	     * (not merely an echo of the AT command)
	     * we can't do a check for OK, because RING or other events
	     * may come in just at this moment...
	     */
	    if (ModemPoll && !FD_ISSET(STDIN,&read_set))
	    {
		send("AT\r");

		if (waitfor(NULL,MEDMTIME) || !strncmp(Match,"AT",2)){
		    logmesg("MODEM died while waiting for work");
		    debug3(D_RUN, "MODEM died while waiting for work\n", MsgBuf);
#if defined(DEBUG) && defined(TIOCMGET)
		    modemstat();
#endif
		    close(FifoFd);
		    return FAIL;
		}
	    }
	}
	while (!FD_ISSET(STDIN,&read_set));	/* continue until MODEM data */

	close(FifoFd);				/* close the FIFO */

#ifdef DEBUG
	now = time(NULL);
	debug3(D_RUN, "... got one! - %s", ctime(&now));
#endif

	/* try to lock the line, we don't want to
	 * read chars if it is locked by someone else, and neither
	 * do we want anyone else to pickup the MODEM while the phone
	 * is ringing...
	 */
	if (makelock(Lock) == FAIL)
	    return LOCKED;

	/* read the first keyword (RING, VCON, NO CARRIER...) */

	if (Match == NULL && waitfor(NULL,SHORTTIME))
	{
	    /* Event reports (sent only by 2864 modem) */

	    if (Match[0] == DLE) {
		debug2(D_RUN, "returned EVENT\n");
		return EVENT;
	    }

	    goto garbage;
	}

	/* save the reply, remove trailing CR/LF garbage and spaces */

	rip(Match);
	strcpy(MsgBuf,Match);

	/* ISDN firmware sometimes sends DISPLAY info, catch it first */

	if (Fw_features & FW_ISDN)
	    if (MsgBuf[0] == 'N' && MsgBuf[strlen(MsgBuf) - 1] == '#') {
		/* string of the form N21X0*0302817450# sent as DISPLAY info */
		sprintf(MsgBuf,"DISPLAY info: %s",Match);
		logmesg(MsgBuf);
		debug3(D_RUN, "%s\n", MsgBuf);
		send("ATD?\r");			/* it seems to need this... */
		waitfor("ERROR",MEDMTIME);	/* ... to recover after info */
		rmlocks();
		goto again;
	    }

	Match = NULL;

	/* see what it is... CONNECT checked first because it changes mode */

	if (!strncmp(MsgBuf,"CONNECT",7)) {
	    cndfile();
	    return CONNECT;
	}

#ifndef ZFAX
	if (!strcmp(MsgBuf,"+FCON")) {
	    cndfile();
	    return CONNECT;
	}
#endif

	/* turn-off AA/DSR to be really sure (cannot be done when CONNECT) */

	if (Voice && ShowMsg && messages)
	    switch (ShowMsg)
	    {
	    case SHOWMSG_AA:
		send("ATS0=0\r");		/* AA off */
		if (waitfor("OK",MEDMTIME))
		    return FAIL;
		if (Match != ChatBuf)		/* something else came back! */
		    ringtype = cndmessage(ChatBuf); /* process as CND message */
		break;
	    case SHOWMSG_DSR:
		send("AT&S1\r");		/* DSR off */
		if (waitfor("OK",MEDMTIME))
		    return FAIL;
		if (Match != ChatBuf)		/* something else came back! */
		    ringtype = cndmessage(ChatBuf); /* process as CND message */
		break;
	    }

	/* check for remaining result strings */

	if (!strcmp(MsgBuf,"NO CARRIER"))
	    return NOCARRIER;

	if (!strcmp(MsgBuf,"VCON")) {
	    cndfile();
	    return VCON;
	}

	if ((Fw_features & FW_ISDN) && !strcmp(MsgBuf,"OK")) {
	    cndfile();
	    return VCON;
	}

	if (!strncmp(MsgBuf,"RING",4))
	{
	    int rings;

	    /* first see if CND data is pending, to avoid some trouble */

	    if (CndNumber == NULL && ToNumber == NULL &&
		waitfor(NULL,SHORTTIME) == SUCCESS)
		ringtype = cndmessage(Match);

	    /* distinctive ring may result in messages RING1..RING4 */
	    /* process them when no (ISDN) ringtype determined yet */

	    if (ringtype == 0)
		ringtype = atoi(MsgBuf + 4);	/* get type of ringing */

	    /* Linux ISDN returns raw caller info in S-registers :-( */

	    if ((Fw_features & FW_LINUX) && CndNumber != NULL) {
		send("ATS21?S22?\r");
		if (waitfor("OK",MEDMTIME))
		    return FAIL;
		if (Match != ChatBuf) {
		    int oct3,oct3a;

		    MsgBuf[0] = '\0';
		    oct3 = atoi(ChatBuf);

		    for (p = ChatBuf; *p != '\0';)
			if (*++p == '\n')
			    break;

		    oct3a = atoi(p);

		    /*
		    bit3 2 1 0 Numbering plan
		       0 0 0 0 Unknown
		       0 0 0 1 ISDN/Telephony
		       0 0 1 1 Data
		       0 1 0 0 Telex
		       1 0 0 0 National standard
		       1 0 0 1 Private
		       1 1 1 1 Reserved
		    */
		    switch (oct3 & 0x0f)
		    {
		    case 0:			/* Unknown, probably E.164 */
			strcpy(MsgBuf,"0");	/* assume a national number */
			break;

		    case 1:			/* E.164 Telephony */
			/*
			 bit7 6 5 Type of number
			    0 0 0 Unknown
			    0 0 1 International number
			    0 1 0 National number
			    0 1 1 Network specific number
			    1 0 0 Subscriber number
			    1 1 0 Abbreviated number
			    1 1 1 Reserved
			*/
			switch ((oct3 >> 4) & 7)
			{
			case 1:			/* international number */
			    strcpy(MsgBuf,"00");
			    break;
			case 2:			/* national number */
			    strcpy(MsgBuf,"0");
			    break;
			case 4:			/* subscriber number */
			    break;
			default:
			    sprintf(MsgBuf,"%d-",(oct3 >> 4) & 7);
			    break;
			}
			break;

		    default:
			sprintf(MsgBuf,"%d:%d-",oct3 & 0x0f,(oct3 >> 4) & 7);
			break;
		    }

		    /*
		    bit6 5 Presentation indicator
		       0 0 Presentation allowed
		       0 1 Presentation restricted
		       1 0 Number not available
		       1 1 Reserved
		    */
		    switch (((oct3a >> 5) & 3))
		    {
		    case 1:			/* presentation restricted */
			free(CndNumber);	/* no caller ID */
			CndNumber = NULL;
			break;

		    case 2:			/* number not available */
			strcpy(MsgBuf,"N/A");
			CndNumber[0] = '\0';	/* fall through (screening) */

		    default:
			strcat(MsgBuf,CndNumber);

			/*
			bit1 0 Screening indicator
			   0 0 User-provided, not screened
			   0 1 User-provided, verified and passed
			   1 0 User-provided, verified and failed
			   1 1 Network provided
			*/
			switch (oct3a & 3)
			{
			case 0:			/* not screened */
			    strcat(MsgBuf,"?");
			    break;
			case 2:			/* verify failed */
			    strcat(MsgBuf,"!");
			    break;
			}

			if (strcmp(MsgBuf,CndNumber)) {
			    free(CndNumber);
			    CndNumber = strdup(MsgBuf);
			    sprintf(MsgBuf,"CndNumber: %s",CndNumber);
			    logmesg(MsgBuf);
			}
			break;
		    }
		}
	    }

	    cndfile();				/* get CND from file */

	    /* find out how many rings we need (for this type of ringing) */
	    /* 'messages present' indicator is passed to select tollsaver */

	    if (read_state(ringtype,messages) != SUCCESS)
		Nrings = 4;

	    /* when voice mode, adjust ring volume when requested */

	    if (Mode == VOICE && VoiceRing != NULL && Nrings != 1)
	    {
		sprintf(MsgBuf,"AT%s\r",VoiceRing);
		send(MsgBuf);
		if (waitfor("OK",MEDMTIME))
		    return FAIL;
		if (Match != ChatBuf)		/* something else came back! */
		    cndmessage(ChatBuf);	/* process as CND message */
	    }

	    rings = 1;

	    while (rings < Nrings || Nrings == 0)
	    {
		Signal = 0;

		/* wait for the next ring, or CND message from MODEM */

		if (waitfor(NULL, LONGTIME))
		    strcpy(Match = ChatBuf,"RUNG");

		/* remove trailing CR/LF garbage and spaces */

		if (rip(Match) > 80)
		    goto garbage;

		/* RUNG when nothing received or ringing stopped (Linux ISDN) */

		if (!strcmp(Match,"RUNG")) {
		    /* check for SIGUSR1
		     * simulate a RING response
		     */
		    if (Signal & (1 << SIGUSR1)) {
			debug2(D_RUN, "SIGUSR1: simulate RING\n");
			strcpy(Match = ChatBuf,"RING");
		    } else {
			/* nothing received
			 * assume not enough rings
			 */
			if (Nrings != 0) {
			    sprintf(MsgBuf, "%d RING%s", rings, (rings == 1? "":"s"));
			    if (ringtype != 0)
				sprintf(MsgBuf + strlen(MsgBuf)," (%d)", ringtype);
			    sprintf(MsgBuf + strlen(MsgBuf), ", %d required", Nrings);
			    if (RingBack != 0 && rings <= RingBack) {
				strcat(MsgBuf,", ");
				switch (ModeRB)
				{
				case NOANSWER:
				    strcat(MsgBuf,"NOANSWER");
				    break;
				case DATAFAX:
				    strcat(MsgBuf,"DATAFAX");
				    break;
				case DATA:
				    strcat(MsgBuf,"DATA");
				    break;
				case FAX:
				    strcat(MsgBuf,"FAX");
				    break;
				case VOICE:
				    strcat(MsgBuf,"VOICE");
				    break;
				case MESSAGE:
				    strcat(MsgBuf,"MESSAGE");
				    break;
				}
				strcat(MsgBuf," mode for 60 sec");
				hold_state(60,ModeRB,2);
			    }
			    logmesg(MsgBuf);
			    debug3(D_RUN, "%s\n", MsgBuf);
			}
			rmlocks();
			return NORING;
		    }
		}

		/* again check for these statuses, it may be that a human */
		/* answered the phone after some RINGs, and then pressed */
		/* DATA/VOICE because she detected it was a FAX or DATA call */

		if (!strncmp(Match,"CONNECT",7))
		    return CONNECT;

		if (!strcmp(Match,"NO CARRIER"))
		    return NOCARRIER;

		if (!strcmp(Match,"VCON"))
		    return VCON;

		if ((Fw_features & FW_ISDN) && !strcmp(MsgBuf,"OK"))
		    return VCON;

		/* check if it was indeed a RING that we received */

		if (!strncmp(Match,"RING",4)) {
		    int newringtype;

		    if ((newringtype = atoi(Match + 4)) != 0 &&
			newringtype != ringtype)
		    {
			/* ringtype has changed (can happen with EDR) */

			ringtype = newringtype;

			/* check state again, maybe number of rings is */
			/* different for this ringtype... */

			if (read_state(ringtype,messages) != SUCCESS)
			    Nrings = 4;
		    }

		    rings++;
		    continue;
		}

		/* it most likely is a CND message...  */

		cndmessage(Match);
	    }

	    /* we have got the requested rings */

	    debug4(D_RUN, "%d RING%s, SUCCESS\n", rings, (rings == 1? "":"s"));
	    return RING;
	}

	/* ISDN firmware after 2.07b for Omni can send caller-ID without
	 * preceding RING message (when calling A/B port), apparently to
	 * satisfy Microsoft.  We'll have to handle it...
	 */

	if ((Fw_features & FW_ISDN) &&
	    (!strncmp(MsgBuf,"FM:",3) || !strncmp(MsgBuf,"TO:",3)))
	{
	    /* just process the info as if a RING came in, so a possible
	     * state program will be called and can save it to a cndfile.
	     */

	    ringtype = cndmessage(MsgBuf);
	    read_state(ringtype,messages);
	    Mode = NOANSWER;
	    rmlocks();
	    return NORING;
	}

garbage:
	strcpy(MsgBuf,"garbage from MODEM: ");
	for (p = ChatBuf; *p != '\0' && strlen(MsgBuf) < sizeof(MsgBuf) - 3; p++)
	    sprintf(MsgBuf + strlen(MsgBuf),
		    (((unsigned char)*p < ' ') ? "^%c" : "%c"),
		    (((unsigned char)*p < ' ') ? *p | 0100 : *p));
	logmesg(MsgBuf);
	debug3(D_RUN, "%s\n", MsgBuf);
	waitfor("FLUSH",SHORTTIME);		/* flush garbage */
	return FAIL;
}

/*
**	cndmessage() - handle CND message
**
**	save information and write it in the log
**	when TO: number present, may return a "ringtype"
*/

int
cndmessage(msg)
char *msg;

{
	char *p;
	int ringtype = 0;

	rip(msg);
	debug3(D_RUN, "cndmessage(\"%s\")\n", msg);

	if ((p = strchr(msg,':')) != NULL)
	{
	    /* TIME: message (not logged) */

	    if (!strncmp(msg,"TIME:",5))
		return ringtype;

	    /* xxxNAME: and xxxNUMBER: messages */

	    if (p > (msg + 6)) {
		if (!strncmp(p - 4,"NAME",4)) {
		    while (*++p == ' ')
			;
		    free(CndName);
		    CndName = strdup(p);
		}
		if (!strncmp(p - 6,"NUMBER",6)) {
		    while (*++p == ' ')
			;
		    free(CndNumber);
		    CndNumber = strdup(p);
		}
	    }

	    /* FM: a-number TO: b-number message (ISDN) */

	    if (p == (msg + 2) && !strncmp(msg,"FM",2))
	    {
		char *q = p;

		while (*++q == ' ')
		    ;

		if ((p = strchr(q,' ')) != NULL)
		    *p = '\0';

		free(CndNumber);
		CndNumber = strdup(q);

		if (p != NULL) {
		    *p = ' ';
		    p = strchr(p + 1,':');
		}
	    }

	    /* TO: b-number message (when call not from ISDN?) */

	    if (p != NULL && p >= (msg + 2) && !strncmp(p - 2,"TO",2))
	    {
		int i;

		while (*++p == ' ')
		    ;

		free(ToNumber);
		ToNumber = strdup(p);

		for (i = 0; i < 8; i++)
		    if (RingType[i] != NULL && !strcmp(RingType[i],p)) {
			ringtype = i + 1;
			break;
		    }
	    }
	}

	/* MM-DD hh:mm number */

	if (isdigit(msg[0]) && isdigit(msg[1]) && msg[2] == '-')
	{
	    free(CndNumber);
	    CndNumber = strdup(msg + 12);
	}

	/* log it in the user logfile */

	if (strcmp(msg,"OK") && strcmp(msg,"RING"))
	    logmesg(msg);

	if (ringtype != 0) {
	    debug3(D_RUN, "ringtype %d\n", ringtype);
	}

	return ringtype;
}

/*
**	cndfile() - read CND data from file
**
**	read CND information from a file specified in CndFile when not
**	received from MODEM before.
**	(e.g. created by another process watching ISDN)
**	make sure it is recent, not some old leftover
*/

void
cndfile()

{
	FILE *fp;
	char *p;
	struct stat st;
	char buf[MAXLINE];

	if (CndNumber == NULL && CndFile != NULL)
	{
	    /* we have no CND info, see if there is a file */

	    sprintf(buf, CndFile, Device);	/* make filename */

	    if ((fp = fopen(buf,"r")) != NULL) {
		if (fstat(fileno(fp),&st) == 0 &&
		    st.st_mtime > (time(NULL) - 60) &&
		    fgets(buf,sizeof(buf)-1,fp) != NULL &&
		    buf[0] != '\n' && buf[0] != '\0')
		{
		    if ((p = strchr(buf,'\n')) != NULL)
			*p = '\0';

		    if ((p = strchr(buf,' ')) != NULL)
			*p = '\0';		/* found space, split it */

		    CndNumber = strdup(buf);
		    debug3(D_RUN, "CndNumber from file (%s)\n", CndNumber);

		    if (p != NULL)
		    {
			while (*++p != '\0' && !isspace(*p))
			    ;

			if (*p != '\0') {
			    free(CndName);
			    CndName = strdup(p);
			    debug3(D_RUN, "CndName from file (%s)\n", CndName);
			}
		    }
		}
		fclose(fp);
	    }
	}
}


/*
**	read_state() - read the statefile to find mode, number of rings etc
**
**	ringtype selects how many lines are skipped in the file, to
**	support distinctive ringing.  (RING, RING1, RING2 etc)
*/

int
read_state(ringtype,altcount)
int ringtype;
int altcount;

{
	FILE *fp;
	char name[MAXLINE];

	/* check if we a still holding a previous state
	 * (e.g. after manual entry into datamode, the phone is
	 *  answered in datamode for some time)
	 */
	if (HoldTime != 0 && time(NULL) < HoldTime) {
	    debug4(D_RUN, "Held state: Mode=%d Nrings=%d\n", Mode, Nrings);
	    return SUCCESS;
	}

	HoldTime = 0;

	/* see if a state-determination program is defined */

	if (StateProg != NULL) {
	    int fd;
	    pid_t pid;
	    char *argv[8];
	    char buf1[20],buf2[20];

	    /* we have a StateProg, call it to read the current state */
	    /* args are: device name, */
	    /*		 ringtype, */
	    /*		 number of messages, */
	    /*		 calling number, */
	    /*		 calling name */

	    sprintf(name, StateProg, Device);

	    fd = open("/dev/null",O_RDWR);	/* /dev/null as input */

	    memset(argv,0,sizeof(argv));	/* build argvector to exec */
	    sprintf(buf1,"%d",ringtype);
	    argv[1] = Device;
	    argv[2] = buf1;
	    sprintf(buf2,"%d",altcount);
	    argv[3] = buf2;
	    argv[4] = CndNumber != NULL? CndNumber : "";
	    argv[5] = CndName != NULL? CndName : "";

	    /* start the program, put its output on fd */

	    debug6(D_RUN, "Running StateProg: %s %s %s %s",name,argv[1],argv[2],argv[3]);
	    debug4(D_RUN, " \"%s\" \"%s\"\n",argv[4],argv[5]);

	    if (push(&pid,&fd,name,argv) == SUCCESS) {
		if ((fp = fdopen(fd,"r")) != NULL) {
		    alarm(10);			/* guard against hangs */
		    gets_state(fp,altcount);	/* read first output line */
		    fclose(fp);
		    waitpid(pid,&fd,0);
		    alarm(0);
		    debug5(D_RUN, "Mode=%d Nrings=%d RingBack=%d\n", Mode, Nrings, RingBack);
		    return SUCCESS;
		}

		debug2(D_RUN, "fdopen() failed - ");
		alarm(10);
		waitpid(pid,&fd,0);
		alarm(0);
	    } else
		debug2(D_RUN, "can't run program - ");

	    debug2(D_RUN, "Falling back to StateFile\n");
	    close(fd);
	}

	sprintf(name, StateFile, Device);

	if ((fp = fopen(name,"r")) == NULL) {
	    debug3(D_RUN, "no state file (%s)\n", name);
	    return FAIL;
	}

	debug4(D_RUN, "reading state(%d) from (%s) ", ringtype, name);

	do
	{
	    gets_state(fp,altcount);
	} while (ringtype--);

	fclose(fp);
	debug5(D_RUN, "Mode=%d Nrings=%d RingBack=%d\n", Mode, Nrings, RingBack);
	return SUCCESS;
}


/*
**	gets_state() - read a line from the statefile and set global vars
*/

void
gets_state(fp,altcount)
FILE *fp;
int altcount;

{
	char *p,*q;
	char buf[MAXLINE];

	RingBack = 0;
	free(Message);
	Message = NULL;

	if (fgets(buf,sizeof(buf),fp) != NULL) {
	    switch (buf[0])
	    {
	    case 'N':				/* No Answer */
		Mode = NOANSWER;
		break;

	    case 'D':				/* DATA/FAX mode */
		Mode = DATAFAX;
		break;

	    case 'F':				/* FAX-only mode */
		Mode = FAX;
		break;

	    case 'V':				/* VOICE mode */
		Mode = VOICE;
		break;

	    case 'M':				/* MESSAGE mode */
		Mode = MESSAGE;
		break;
	    }

	    p = buf;

	    while (*p != '\0' && !isspace(*p))	/* skip keyword */
		p++;

	    while (*p != '\0' && isspace(*p))	/* skip blanks */
		p++;

	    Nrings = atoi(p);			/* number of rings before answer */

	    while (isdigit(*p))			/* skip number */
		p++;

	    while (*p != '\0' && isspace(*p))	/* skip blanks */
		p++;

	    if (altcount && isdigit(*p)) {	/* try to get 2nd (tollsaver) count? */
		Nrings = atoi(p);

		while (isdigit(*p))		/* skip number */
		    p++;

		while (*p != '\0' && isspace(*p)) /* skip blanks */
		    p++;
	    }

	    if ((q = strchr(p,'R')) != NULL) {	/* ringback specified? */
		RingBack = atoi(++q);		/* get number of rings */

		while (isdigit(*q))		/* skip #rings digits */
		    q++;

		ModeRB = DATA;			/* default is DATA mode */

		if (*q == ':')			/* explicit mode specified? */
		    switch (q[1])
		    {
		    case 'N':			/* No Answer */
			ModeRB = NOANSWER;
			break;

		    case 'D':			/* DATA/FAX mode */
			ModeRB = DATAFAX;
			break;

		    case 'F':			/* FAX-only mode */
			ModeRB = FAX;
			break;

		    case 'V':			/* VOICE mode */
			ModeRB = VOICE;
			break;

		    case 'M':			/* MESSAGE mode */
			ModeRB = MESSAGE;
			break;
		    }

		while (*q != '\0' && !isspace(*q)) /* skip keyword */
		    q++;

		p = q;
	    }

	    if ((q = strchr(p,'M')) != NULL) {	/* message specified? */
		p = q;

		while (*p != '\0' && !isspace(*p)) /* find end of name */
		    p++;

		*p++ = 0;
		Message = strdup(q + 1);	/* get message name */
	    }

	    if ((q = strchr(p,'K')) != NULL) {	/* keysfile specified? */
		p = q;

		while (*p != '\0' && !isspace(*p)) /* find end of name */
		    p++;

		*p++ = 0;
		VoiceKey = strdup(q + 1);	/* get filename */
	    }
	}
}

/*
**	hold_state() - lock the state for specified duration
*/

void
hold_state(duration,mode,nrings)
int duration;
int mode;
int nrings;

{
	HoldTime = time(NULL) + duration;	/* end-time for this state */

	Mode = mode;
	Nrings = nrings;
}

/*
**	dump_epa() - dump the (ISDN) Embedded Protocol Analyzer.
*/

int
dump_epa()

{
	FILE *dump;
	int rv,try;
	time_t tim;
	char line[MAXLINE];

	if (EpaFile == NULL)			/* no EPA dump requested? */
	    return SUCCESS;

	if (!(Fw_features & FW_ISDN))		/* must have ISDN firmware */
	    return FAIL;

	debug3(D_RUN, "dump_epa(\"%s\")\n", EpaFile);

	if ((dump = fopen(EpaFile,"a")) == NULL)
	    return FAIL;

	tim = time(NULL);			/* write the time */
	fprintf(dump, "\n%s", ctime(&tim));

	debugtime(D_CHAT);
	waitfor("FLUSH",SHORTTIME);		/* flush garbage */
	debugtime(D_CHAT);

	if (Fw_features & FW_FASTC)
	    send("ATC$1\r");			/* send EPA without wait */
	else
	    send("ATC$\r");			/* send EPA page-by-page */

	if (waitfor("OK",MEDMTIME) != SUCCESS)	/* initially it says OK */
	    return FAIL;

	settermios(LONGTIME);			/* first screen takes time */

	for (;;) {
	    try = 1;

	    while ((rv = line_epa(line)) != SUCCESS && try--) {
		if (!(Fw_features & FW_FASTC)) {
		    debugtime(D_CHAT);
		    send("3");			/* send "next page" command */
		    settermios(MEDMTIME);	/* allow it some time */
		}
	    }

	    settermios(SHORTTIME);		/* react quickly to pause */

	    if (rv != SUCCESS)
		break;

	    fputs(line,dump);
	    putc('\n',dump);

	    if (!strcmp(line,"End of buffer"))
		break;
	}

	fclose(dump);

	debugtime(D_CHAT);

	if (Fw_features & FW_FASTC) {
	    send("AT\r");
	    return waitfor("OK",MEDMTIME);
	} else {
	    send("Q");
	    return waitfor("QUIT",MEDMTIME);
	}
}

/*
**	line_epa() - read a single Embedded Protocol Analyzer output line
*/

int
line_epa(line)
char *line;

{
	int ch;
	int esc = 0;

	debugtime(D_EPA);

	/* discard any leading CR and LF characters (from previous line) */

	while ((ch = getch()) != EOF && (ch == '\r' || ch == '\n' || ch == '\0'))
	    debug3(D_EPA, "^%c", ch | 0100);

	if (ch == EOF)
	    return FAIL;

	/* keep filling the linebuf until the read times out */

	do
	{
	    debug3(D_EPA, ((ch < ' ') ? "^%c" : "%c"),
			   ((ch < ' ') ? ch | 0100 : ch));

	    if (esc) {				/* skip ANSI ESC sequence */
		if (isalpha(ch))
		    esc = 0;			/* it ends at a LETTER */

		continue;
	    }

	    if (ch == ESC) {			/* find ANSI ESC sequence */
		esc++;
		continue;
	    }

	    if (ch == '\n')			/* stop at CR/LF */
		break;

	    if (ch >= ' ')			/* skip any ctrl chars */
		*line++ = ch;
	}
	while ((ch = getch()) != EOF);

	debug2(D_EPA, "\n");
	*line = '\0';
	return SUCCESS;
}

/*
**	zmsp_test() - some experiments in multiple stream mode (CAPI)
*/

int
zmsp_test(out)
int out;

{
	int ch,type,applid,length;
	char buf[MAXLINE];

	settermios(LONGTIME);			/* set timeout for read */
	
	send("\001\001\002\001");		/* register appl 0102 */

	for (;;)
	{
	    Fputs("\n",out);

	    while ((ch = getch()) != EOF && ch != 0x01) {
		sprintf(buf," %02x",ch);
		Fputs(buf,out);
	    }

	    if (ch != 0x01)
		break;

	    if ((ch = getch()) == EOF)
		break;

	    type = ch;

	    if ((ch = getch()) == EOF)
		break;

	    applid = ch;

	    if ((ch = getch()) == EOF)
		break;

	    applid |= ch << 8;

	    sprintf(buf,"type 0x%02x applid 0x%04x",type,applid);
	    Fputs(buf,out);

	    switch (type)
	    {
	    case 0x03:
		if ((ch = getch()) == EOF)
		    break;

		length = ch;

		if ((ch = getch()) == EOF)
		    break;

		length |= ch << 8;

		sprintf(buf,"length %d",length);
		Fputs(buf,out);
		length -= 2;

		while (length > 0)
		{
		    if ((ch = getch()) == EOF)
			break;

		    sprintf(buf," %02x",ch);
		    Fputs(buf,out);
		}
		break;

	    case 0x06:
	    }
	}

	return SUCCESS;
}


/*************************************************************************
**	API mode handling						**
*************************************************************************/

int
api_mode()

{
	int ch,quit,out,playLS,result;
	char *p;
	char buf[MAXLINE],buf2[MAXLINE];
	sigh_t (*oldint)();

	/* try to lock the line, we don't want others to use it
	 */
	if (makelock(Lock) == FAIL)
	    return LOCKED;

	/* ignore sig_int during api_mode, the interface
	 * uses this signal to stop playing/recording, and
	 * we don't want this program to quit if the signal
	 * comes a little to late and playing/recording already ended
	 */
	oldint = signal(SIGINT,SIG_IGN);

	ApiPtr = ApiLen = 0;			/* flush buffer */
	quit = 0;
	out = -1;
	playLS = 16;

	if (Fw_features & FW_LINUX)
	    playLS = 0;

	while (!quit && (ch = getapi()) != EOF) {
#ifdef DEBUG
	    if (isprint(ch))
		debug3(D_RUN, "API cmd '%c'\n",ch)
	    else
		debug3(D_RUN, "API cmd 0x%02x\n",ch);
#endif
	    switch (toupper(ch))
	    {
	    case '\n':				/* newline, give menu */
		Fputs(Id + 4,out);		/* my name and version */
		Fputs("\n\nCommands are:\n",out);
		Fputs("!                 Print ZyXEL pid\n",out);
		Fputs("#                 Print program version\n",out);
		Fputs("$                 Print MODEM version\n",out);
		Fputs(">                 Set output fifo/file/tty\n",out);
		Fputs("S                 Use SPEAKER for playback (default)\n",out);
		Fputs("L                 Use LINE for playback\n",out);
		Fputs("V                 Access VoiceMailBox\n",out);
		Fputs("C<command>        Execute VoiceMailBox command\n",out);
		Fputs("P<file>           Playback\n",out);
		Fputs("R<n>              Record over line <n>\n",out);
		Fputs("X<number>         Call <number> and record\n",out);
		Fputs("F<number> <file>  Send a FAX\n",out);
		Fputs("Q<number> <id>    Poll for FAX\n",out);
		Fputs("D                 Hold DATA/FAX mode for 5 minutes\n",out);
		Fputs("M<message>        Send pager message\n",out);
		Fputs("O                 Setup outgoing call\n",out);
		Fputs("AT<cmd>           Issue MODEM command\n",out);
		if (Fw_features & FW_ISDN)
		    Fputs("E                 EPA Dump\n",out);
		if (Fw_features & FW_CLIR)
		    Fputs("*                 Momentary CLIR enable\n",out);
		Fputs("\n",out);
		break;

	    case '!':				/* report pid */
		quit = api_eol();
		sprintf(buf,"%d\n",(int)getpid());
		Fputs(buf,out);
		break;

	    case '#':				/* program version */
		quit = api_eol();
		Fputs(Id + 4,out);		/* my name and version */
		Fputs("\n",out);
		break;

	    case '$':				/* MODEM version */
		quit = api_eol();
		Fputs(Modem,out);
		Fputs("\n",out);
		break;

	    case '*':				/* CLIR temporary enable */
		quit = api_eol();
		if (Fw_features & FW_CLIR) {
		    send("ATS0=99S120.6=1\r");
		    waitfor("OK",MEDMTIME);
		    Fputs(ChatBuf,out);
		} else
		    Fputs("?\n",out);
		break;

	    case '>':				/* output */
		if ((p = apigets()) != NULL) {
		    if (out >= 0)
			close(out);
		    out = open(p,O_WRONLY);
		} else
		    quit = 1;
		break;

	    case 'A':				/* issue MODEM command */
		p = apigets();
		sprintf(ChatBuf,"%c%s\r",ch,p);
		send(ChatBuf);
		waitfor("OK",MEDMTIME);
		Fputs(ChatBuf,out);
		break;

	    case 'C':				/* VoiceMailBox command */
		if ((p = apigetl()) != NULL)	/* actually something entered? */
		{
		    strcpy(buf,p);

		    if (set_voice(playLS)) {	/* select voice mode & speaker */
			Fputs("FAIL\n",out);
			return FAIL;
		    }

		    switch (result = docommand(buf,apigetc,apigets))
		    {
		    case DIALBACK:
			voicemode(ONLINE,Message);

		    case SUCCESS:
		    case EXIT:
			Fputs("OK\n",out);
			break;

		    default:
			Fputs("ERROR\n",out);
			break;
		    }

		    if (set_vls(0)) {
			Fputs("FAIL\n",out);
			return FAIL;
		    }
		}
		break;

	    case 'D':				/* hold DATA/FAX mode */
		quit = api_eol();
		hold_state(300,DATAFAX,2);
		Fputs("DATA/FAX mode enabled for 5 minutes\n",out);
		break;

	    case 'E':				/* EPA dump */
		quit = api_eol();
		if (Fw_features & FW_ISDN) {	/* must have ISDN firmware */
		    dump_epa();
		    Fputs("OK\n",out);
		} else
		    Fputs("?\n",out);
		break;

	    case 'F':				/* fax sending */
		if ((p = apigets()) == NULL)	/* read number */
		    break;
		strcpy(buf,p);

		if ((p = apigets()) == NULL)	/* read filename */
		    break;

		if (!Fax) {
		    Fputs("FAX disabled\n",out);
		    break;
		}
		strcpy(buf2,p);
		Fputs("Sending FAX to ",out);
		Fputs(buf,out);
		Fputs(", file ",out);
		Fputs(buf2,out);
		Fputs("\n",out);

		switch (faxsend(buf,buf2,out))
		{
		case SUCCESS:
		    Fputs("OK\n",out);
		    break;

		case DIALFAIL:
		    Fputs("DIAL Failed\n",out);
		    break;

		case CONNECTFAIL:
		    Fputs("CONNECT Failed or incompatible\n",out);
		    break;

		case XFERFAIL:
		    Fputs("Transfer failed\n",out);
		    break;

		default:
		    Fputs("Failed\n",out);
		    break;
		}
		break;

	    case 'L':				/* line */
		quit = api_eol();
		Fputs("LINE playback\n",out);
		playLS = 2;
		break;

	    case 'M':				/* send pager message */
		if ((p = apigetl()) != NULL) {
		    strcpy(buf,p);
		    if (pager(buf) == SUCCESS)
			Fputs("OK\n",out);
		    else
			Fputs("FAIL\n",out);
		} else
		    quit = 1;
		break;

	    case 'O':				/* setup outgoing data call */
		if ((p = apigets()) == NULL)	/* read initstring */
		    break;
		strcpy(buf,p);

		if ((p = apigets()) == NULL)	/* read number */
		    break;
		strcpy(buf2,p);

		strcpy(ChatBuf,"ATZE0");	/* load defaults, echo off */
		if ((Fw_features & FW_ISDN) && !(Fw_features & FW_LINUX))
		    strcat(ChatBuf,"CB1CC1CD1CT"); /* start EPA */
		strcat(ChatBuf,"\r");
		send(ChatBuf);

		if (waitfor("OK",MEDMTIME) != SUCCESS || /* init the modem */
		    dialup(buf,0,2,buf2) != SUCCESS) /* dial in data mode */
		{
		    if ((Fw_features & FW_ISDN) && /* ISDN firmware? */
			Match != NULL &&
			!strcmp(Match,"NO CARRIER") && /* TEI bug? */
			dialup(buf,0,2,buf2) == SUCCESS) /* re-try it */
		    {
			goto isdn_bug;		/* went OK, just continue */
		    }

		    if (Match != NULL && Match[0] != '\0') {
			Fputs(Match,out);
			Fputs("\n",out);
		    } else
			Fputs("FAIL\n",out);

		    dump_epa();			/* failed, dump the EPA */
		    return FAIL;
		}

isdn_bug:	Fputs("@C\n",out);

		if ((p = apigetl()) != NULL)	/* read commandline */
		{
		    if (out >= 0)
			close(out);		/* close output channel */

		    close(FifoFd);
		    settermios(INITIAL);
		    debug3(D_RUN, "execl: sh -c %s\n", MsgBuf);
#ifdef DEBUG
		    if (Dfp != NULL) {
			fclose(Dfp);
			Dfp = NULL;
		    }
#endif	/* DEBUG */
		    execl("/bin/sh", "sh", "-c", MsgBuf, NULL);
		    exit(EXIT_FAILURE);
		}

		quit = 1;			/* quit from API mode */
		break;

	    case 'P':				/* play back */
		p = apigets();			/* read filename */
		strcpy(buf,p);

		if (set_voice(playLS)) {	/* select voice mode & speaker */
		    Fputs("FAIL\n",out);
		    return FAIL;
		}

		if (play(buf,NULL) != SUCCESS)
		    Fputs("ERROR invalid file\n",out);
		else
		    if (play(buf,"#bdl") != SUCCESS)
			Fputs("ERROR play\n",out);
		    else
			if (strpbrk(MsgBuf,"bdhl") != NULL) /* 'busy', 'dialtone', 'hangup' */
			    Fputs("ERROR busy or dialtone\n",out);
			else
			    Fputs("OK\n",out);

		if (set_vls(0)) {
		    Fputs("FAIL\n",out);
		    return FAIL;
		}
		break;

	    case 'Q':				/* fax polling */
		if ((p = apigets()) == NULL)	/* read number */
		    break;
		strcpy(buf,p);

		if ((p = apigetl()) == NULL)	/* read POLL ID into buf2 */
		    strcpy(buf2,"0000");
		else
		    strcpy(buf2,p);

		if (!Fax) {
		    Fputs("FAX disabled\n",out);
		    break;
		}
		Fputs("Polling ",out);
		Fputs(buf,out);
		Fputs(" ID ",out);
		Fputs(buf2,out);
		Fputs("\n",out);

		switch (faxpoll(buf,buf2,out))
		{
		case SUCCESS:
		    Fputs("OK\n",out);
		    break;

		case DIALFAIL:
		    Fputs("DIAL Failed\n",out);
		    break;

		case NOPOLL:
		    Fputs("No FAX POLLING available\n",out);
		    break;

		case XFERFAIL:
		    Fputs("Transfer failed\n",out);
		    break;

		default:
		    Fputs("Failed\n",out);
		    break;
		}
		break;

	    case 'R':				/* record */
		RecLS = atoi(apigets());
		strcpy(buf,RecFile);
		switch (result = record(buf,TRUE,"0123456789ABCD*#abcdelq",Silence,RecMode,120))
		{
		case 0:
		    Fputs("ERROR silence\n",out);
		    break;

		case FAIL:
		    Fputs("ERROR in record\n",out);
		    break;

		default:
		    sprintf(ChatBuf,"%dsec message recorded (%s)\n",result,MsgBuf);
		    Fputs(ChatBuf,out);
		    Fputs("OK\n",out);
		    break;
		}
		RecLS = 0;
		break;

	    case 'S':				/* speaker */
		quit = api_eol();
		Fputs("SPEAKER playback\n",out);
		playLS = 16;
		break;

	    case 'V':				/* voice mailbox */
		quit = api_eol();

		if (set_voice(playLS)) {	/* select voice mode & speaker */
		    Fputs("FAIL\n",out);
		    return FAIL;
		}

		Fputs("VoiceMailBox, enter numeric command\n",out);
		Fputs("  0: leave VoiceMailBox\n",out);
		Fputs("\nIn Playback mode, commands are:\n",out);
		Fputs("  0: leave playback mode\n",out);
		Fputs("  2: tell message date/time\n",out);
		Fputs("  3: next message\n",out);
		Fputs("  4: delete message\n",out);
		Fputs("  5: archive message\n",out);
		Fputs("  7: first message\n",out);
		Fputs("  8: repeat message\n",out);
		Fputs("  9: last message\n",out);

		UserLevel = ApiUserLevel;	/* API default user level */
		Number[0] = '\0';

		while (Fputs("\nVMB> ",out), (p = apigets()) != NULL)
		{
		    char *file;

		    if (p[0] == '.') {
			file = VoiceKey;
			p++;
		    } else
			file = VoiceCmd;

		    switch (voicecommand(file,atoi(p),apigetc,apigets))
		    {
		    case DIALBACK:
			voicemode(ONLINE,Message);
			break;

		    case EXIT:
			goto endvmb;

		    case UNKNOWN:
			Fputs("?\n",out);
			break;

		    case ERROR:
			Fputs("ERROR\n",out);
			break;

		    case FAIL:
			Fputs("FAIL\n",out);
			return FAIL;
		    }
		}
endvmb:
		if (set_vls(0)) {
		    Fputs("FAIL\n",out);
		    return FAIL;
		}

		Fputs("\nexit from VMB\n",out);
		break;

	    case 'X':				/* dialout record mode */
		if ((p = apigetl()) == NULL)	/* read number */
		    break;
		strcpy(buf,p);

		if (set_voice(playLS)) {	/* select voice mode & speaker */
		    Fputs("FAIL\n",out);
		    return FAIL;
		}

		sound_open();			/* enable soundcard */

		Fputs("Dialling ",out);
		Fputs(buf,out);
		Fputs("\n",out);

		if (dialup("",8,5,buf) != SUCCESS) { /* dial in voice mode */
		    Fputs("FAIL\n",out);
		    sound_close();
		    break;
		}

		Fputs("VCON\n",out);

		ch = 1;

		do {
		    sprintf(buf,"/tmp/rec%02d.zad",ch++);

		    Fputs("record ",out);
		    Fputs(buf,out);

		    if (record(buf,FALSE,"bdq","20,20",ENC_BEST,60) < 0)
			Fputs(" FAIL",out);
		    else
			Fputs(" OK",out);

		    Fputs("\nDTMF> ",out);
		    
		    if ((p = apigetl()) != NULL && *p != '\0')
			beep(p);
		} while (p != NULL && toupper(*p) != 'Q');

		sound_close();

		if (set_vls(0)) {
		    Fputs("FAIL\n",out);
		    return FAIL;
		}

		Fputs("\nexit\n",out);
		break;

	    case 'Z':				/* ZMSP test */
		quit = api_eol();
		if (Fw_features & FW_ISDN) {	/* must have ISDN firmware */
		    Fputs("ZMSP test\n",out);
		    send("ATCM3\r");		/* MSP for CAPI 2.0 */
		    zmsp_test(out);
		    send("ATCM0\r");		/* SSP back */
		    waitfor("OK",MEDMTIME);
		    Fputs("DONE\n",out);
		} else
		    Fputs("Not ISDN\n",out);
		break;

	    default:				/* unknown command */
		quit = api_eol();
		Fputs("?\n",out);
		break;
	    }
	}

	if (out >= 0)
	    close(out);				/* close output channel */
	signal(SIGINT,oldint);
	rmlocks();				/* release the lock */
	return SUCCESS;
}

/*
**	apigets() - get space-separated string from API FIFO
**
**	Will put the string in MsgBuf and return pointer to it
**	Reads ahead one character (which is lost)
*/

char *
apigets()

{
	int ch;
	char *p;

	/* skip initial spaces
	 */
	while ((ch = getapi()) != EOF && isspace(ch))
	    ;

	p = MsgBuf;
	*p++ = ch;

	/* read characters until EOF, space, or buffer full
	 */
	while ((ch = getapi()) != EOF && !isspace(ch) &&
		p < (MsgBuf + sizeof(MsgBuf) - 1))
	    *p++ = ch;

	*p = 0;

	return (ch == EOF? NULL : MsgBuf);
}

/*
**	apigetl() - get remainder of line from API FIFO
**
**	Will put the string in MsgBuf and return pointer to it
**	Reads ahead one character (which is lost)
*/

char *
apigetl()

{
	int ch;
	char *p;

	/* skip initial spaces
	 */
	while ((ch = getapi()) != EOF && isspace(ch))
	    ;

	p = MsgBuf;
	if (ch != '\n' && ch != '\r') {
	    *p++ = ch;

	    /* read characters until EOF, \r, \n, or buffer full
	     */
	    while ((ch = getapi()) != EOF && ch != '\n' && ch != '\r' &&
		    p < (MsgBuf + sizeof(MsgBuf) - 1))
		*p++ = ch;

	} else
	    ch = EOF;

	*p = 0;

	return (ch == EOF? NULL : MsgBuf);
}

/*
**	apigetc() - get a single char from the API FIFO (with beep prompt)
*/

int
apigetc()

{
	beep(Prompt1);				/* issue a prompting beep */
	return getapi();
}

/*
**	api_eol() - skip until newline, return 1 when EOF
*/

int
api_eol()

{
	int ch;

	while ((ch = getapi()) != EOF && ch != '\n' && ch != '\r')
	    ;

	return (ch == EOF);
}


/*************************************************************************
**	DATA mode handling						**
*************************************************************************/


/*
**	datamode() - answer an incoming call in DATA mode
**
**	'answer' can be one of:
**
**	ONLINE		the CONNECT is assumed to be already
**			received, and stored in the ChatBuf pointed by Match.
**	ANSWER		answer the call first, wait for CONNECT
**	ANSWERDATA	same, but don't allow FAX calls
*/

int
datamode(answer)
int answer;					/* answer the phone? */

{
	int n;
	char *p;
	sigh_t (*oldhup)(),(*oldint)(),(*oldalarm)() = SIG_DFL;

	/* BREAK should not trip inthandler, getlogname returns BADSPEED.
	 * actually we cannot have a bad speed as the MODEM does the
	 * interspeeding, so we effectively ignore BREAK but re-issue
	 * the login prompt to ease life for UUCP chat scripts...
	 */
	oldint = signal(SIGINT, SIG_IGN);

	if (answer != ONLINE) {
	    debug2(D_RUN, "DATA mode, perform connect sequence\n");

	    /* +VNH=1	no hangup when connect attempt fails (ZyXEL) */
	    /* Mute	sets muting for answer (e.g. M0) */

	    sprintf(ChatBuf,"AT%s%s",(Voice? "+VNH=1" : ""),Mute);

	    if (WFclass > 0) {
		/* prepare MODEM for answer
		 */
		if (answer == ANSWERDATA || !Fax) {
		    if (Is_ZyXEL)
			strcat(ChatBuf,"S38.4=1");	/* no-FAX answer */
#ifdef ZFAX
		    strcat(ChatBuf,"+FCLASS=0");	/* DATA mode */
#else
		    strcat(ChatBuf,"+FCLASS=0+FAA=0");	/* go to DATA mode */
#endif
		    Fclass = 0;
		} else {
#ifdef ZFAX
		    strcat(ChatBuf,"#B0+FCLASS=6");	/* go to DATA/FAX mode */
		    Fclass = 6;
#else
		    strcat(ChatBuf,"+FCLASS=0+FAA=1");	/* go to DATA/FAX mode */
		    Fclass = 0;
#endif
		}
	    }

	    if (ChatBuf[2] != '\0')		/* anything to be sent? */
		NeedInit = TRUE;

	    strcat(ChatBuf,"\r");
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME))
		return FAIL;

	    send("ATA\r");

	    /* wait some time to get the connect, 60 seconds set by MODEM */
	    /* read next answer from MODEM and make sure it is CONNECT */
	    /* (done this way to proceed quickly on other responses) */

	    for (n = 5; n > 0; n--)
		if (!waitfor(NULL,LONGTIME)) {
		    if (rip(Match) > 80) {	/* garbage? */
			logmesg("garbage from MODEM in DATA mode answer");
			n = 0;			/* terminate loop, hangup */
			break;
		    }

		    if (!strncmp(Match,"CONNECT",7))
			break;

#ifndef ZFAX
		    if (!strcmp(Match,"+FCON")) /* is it a FAX connect? */
			break;
#endif

		    if (Fw_features & FW_ISDN) {
			if (!strcmp(Match,"OK")) { /* no reaction to ATA? */
			    signal(SIGINT, oldint);
			    return POTSCALL;	/* -> not a data call */
			}

			if (!strcmp(Match,"RING")) { /* another RING message */
			    signal(SIGINT, oldint);
			    return RING;	/* (probably a second call) */
			}
		    }

		    /* got something else than CONNECT, +FCON, OK or RING */

		    logmesg(Match);		/* log what we got */

		    if (!strcmp(Match,"VCON") || !strcmp(Match,"NO CARRIER")) {
			signal(SIGINT, oldint);
			return VCON;		/* go try VOICE mode */
		    }

		    n = 0;			/* break from loop */
		    break;
		}

	    if (!n) {				/* nothing or error received */
		NeedInit = TRUE;
		return FAIL;
	    }
	}

	free(Connect);				/* save the CONNECT */
	Connect = strdup(Match);
	NeedInit = TRUE;

#ifdef ZFAX
	if (!strncmp(Match + 8, "FAX", 3))	/* is it a FAX connect? */
#else
	if (!strcmp(Match,"+FCON"))		/* is it a FAX connect? */
#endif
	{
	    signal(SIGINT, oldint);
#ifdef ZFAX
	    logmesg(Match);			/* log the CONNECT when ZFAX */
#endif
	    return FAX;
	}

	logmesg(Match);				/* log the CONNECT we got! */
	settermios(INITIAL);			/* select normal handling */
	usleep(1500000);			/* allow line to settle */

	/* from now, watch for loss of DCD
	 */
	oldhup = signal(SIGHUP, huphandler);
	settermios(WATCHDCD);			/* not CLOCAL */

#ifdef TIOCMGET
	/* check if DCD already dropped before we got here...
	 */
	if ((n = modemstat()) != -1 && !(n & TIOCM_CAR)) {
	    logmesg("CARRIER lost within 1 second");
	    settermios(INITIAL);		/* back to CLOCAL mode */
	    signal(SIGHUP, oldhup);
	    signal(SIGINT, oldint);
	    waitfor(NULL,SHORTTIME);		/* flush "NO CARRIER" */
	    waitfor(NULL,SHORTTIME);		/* and other pending garbage */
	    return VOICE;			/* could be a whistle... */
	}
#endif

	/* when PPP connect and PPP program defined, start it
	 * (ISDN versions only)
	 */
	if (PPPProg != NULL && (Fw_features & FW_ISDN) &&
	    (p = strchr(Connect,'P')) != NULL && !strncmp(p,"PPP",3))
	{
	    if ((n = protect("ppp",FALSE)) == SUCCESS)
		n = execlogin(PPPProg,PPPFlags,NULL,0);
	    else
		logmesg("PPP call - not allowed");
	} else {
	    /* when DATA program defined, start it
	     */
	    if (DataProg != NULL) {
		n = execlogin(DataProg,DataFlags,NULL,0);
	    } else {
		/* start timer, if required
		 */
		if (TimeOut > 0) {
		    oldalarm = signal(SIGALRM, timeout);
		    alarm((unsigned) TimeOut);
		}

		/* attempt to get someone logged in
		 */
		n = login_user();

		/* stop alarm clock
		 */
		if (TimeOut > 0) {
		    alarm((unsigned) 0);
		    signal(SIGALRM, oldalarm);
		}
	    }
	}

	signal(SIGHUP, oldhup);
	signal(SIGINT, oldint);
	settermios(INITIAL);
	return n;
}


/*
**	login_user() - attempt to get someone logged in (DATA mode)
*/

int
login_user()

{
	struct utmp *utmp;
	FILE *fp;
	char *login = NULL;
	char *flags = NULL;
	boolean loginenv = FALSE;
	int logintry = 3;
	int rv;
	struct termios termios;
#ifdef TIOCGWINSZ
	struct winsize winsize;
#endif
	char username[MAXLINE];

	debug2(D_RUN, "entering login loop\n");
	WarnCase = TRUE;

	Fputs("\021",STDOUT);		/* send a XON */

	/* loop until a successful login is made
	 */
	for (;;)
	{
	    /* set Nusers value
	     */
	    Nusers = 0;
	    setutent();
	    while ((utmp = getutent()) != NULL) {
#ifdef USER_PROCESS
		if (utmp->ut_type == USER_PROCESS)
#endif	/* USER_PROCESS */
		{
		    Nusers++;
		    debug3(D_UTMP, "utmp entry (%s)\n",
				    utmp->ut_name);
		}
	    }
	    endutent();
	    debug3(D_UTMP, "Nusers=%d\n", Nusers);

	    Fputs("\n\n\n\n\n",STDOUT);	/* just in case */

	    /* display ISSUE, if present
	     */
	    if (*Issue != '/') {
		Fputs(Issue,STDOUT);
		Fputs("\n",STDOUT);
	    } else {
		if ((fp = fopen(Issue, "r")) != NULL) {
		    while (fgets(username, sizeof(username), fp) != NULL)
			Fputs(username,STDOUT);

		    fclose(fp);
		}
	    }

login_prompt:
	    /* eat any chars from line noise
	     */
	    ioctl(STDIN,TCFLSH,TCIFLUSH);
	    InPtr = InLen = 0;

#ifdef FIDO
	    /* put EMSI request
	     */
	    if (Fido != NULL)
		Fputs("**EMSI_REQA77E\r\021              \r",STDOUT);
#endif

login_prompt_noemsi:
	    /* display login prompt
	     */
	    Fputs("login: ",STDOUT);

	    /* take current termios to modify during logname entry
	     * take default login program settings (modified when FIDO)
	     */
	    termios = Termios;
	    login = Login;
	    flags = LoginFlags;
	    loginenv = LoginEnv;

	    /* handle the login name
	     */
	    switch (getlogname(&termios, username, 80))
	    {
#ifdef FIDO
	    case FIDOCALL:		/* username holds the type of FIDO */
		if (Fido != NULL) {	/* program specified? */
		    login = Fido;	/* use it instead of login */
		    flags = NULL;	/* don't pass args */
		    loginenv = FALSE;
		}

		logmesg("Fido Call Detected!");
#endif
	    case SUCCESS:
#ifdef FIDO
		if (Fido != NULL) {
		    if (!strncmp(username,"**EMSI_CLI",10) && logintry--)
			goto login_prompt_noemsi;

		    if (!strncmp(username,"**EMSI_INQ",10)) {
			login = Fido;		/* use Fido instead of login */
			flags = NULL;		/* don't pass args */
			loginenv = FALSE;
			logmesg("EMSI Call Detected!");
		    }
		}
#endif
	    case BBSCALL:
		/* screen the username and optionally ask for password
		 * and/or dial back
		 */
		if ((rv = protect(username,TRUE)) != SUCCESS) {
		    if (rv == INCORRECT) {
			Fputs("Login incorrect\n",STDOUT);
			sleep(5);

			if (logintry--)
			    goto login_prompt;	/* try again */
		    }

		    return rv;
		}

		/* setup terminal
		 */
		Termios = termios;	/* take suggestions by getlogname */
		settermios(FINAL);
#ifdef TIOCGWINSZ
		if (ioctl(STDIN,TIOCGWINSZ,&winsize) == 0) {
		    if (winsize.ws_row == 0)
			winsize.ws_row = 25;

		    if (winsize.ws_col == 0)
			winsize.ws_col = 80;

		    ioctl(STDIN,TIOCSWINSZ,&winsize);
		}
#endif
		/* exec the login program
		 */
		return execlogin(login,flags,username,loginenv);

	    case PPPCALL:
		/* PPP-like input detected, start PPP program */
		if (PPPProg != NULL) {
		    if ((rv = protect("ppp",FALSE)) != SUCCESS) {
			Fputs("PPP call - not allowed\n",STDOUT);
			logmesg("PPP call - not allowed");
			sleep(3);
			return rv;
		    } else
			return execlogin(PPPProg,PPPFlags,NULL,0);
		}
		goto login_prompt;

	    case BADSPEED:
		goto login_prompt;

	    case BADCASE:
		/* first try was all uppercase
		 */
		for (rv=0; bad_case[rv] != NULL; rv++)
		    Fputs(bad_case[rv],STDOUT);
		goto login_prompt;

	    case NONAME:
		/* no login name entered
		 * re-display the issue and try again
		 */
		break;

	    case FAIL:
		return FAIL;
	    }
	}
}

/*
**	execlogin() - run the login program
*/

int
execlogin(login,flags,username,loginenv)
char *login;
char *flags;
char *username;
boolean loginenv;

{
	int argc,i;
	char *argv[NARG];
	char *lbase;
	char flagbuf[80];

	/* guard against bogus username that may confuse "login"
	 */
	if (username != NULL && strchr(username,'-') != NULL) {
	    sprintf(MsgBuf, "bogus username %s", username);
	    logmesg(MsgBuf);
	    return FAIL;
	}

	/* change terminal modes/owner for login purposes
	 */
	debug2(D_RUN, "chmod and chown to root\n");
	chmod(DevName, 0622);
	chown(DevName, 0, 0);

	/* make sure all signals are set to SIG_DFL
	 */
	for (i = 0; i < NSIG; i++)
	    signal(i, SIG_DFL);

	/* log the login attempt
	 */
	sprintf(MsgBuf, "exec %s", login);

	if (flags != NULL)
	    sprintf(MsgBuf + strlen(MsgBuf), " %s", flags);

	if (username != NULL)
	    sprintf(MsgBuf + strlen(MsgBuf), " %s", username);

	logmesg(MsgBuf);

	/* get basename of login program
	 */
	if ((lbase = strrchr(login,'/')) != NULL)
	    lbase++;
	else
	    lbase = login;

	/* build the argument vector
	 */
	argc = 0;
	memset(argv,0,sizeof(argv));

	argv[argc++] = lbase;

	/* handle multiple flags passed in the LOGINFLAGS value.
	 * the great John F. Haugh II couldn't get the parsing right
	 * in his shadow login, and so we have to pass -p and -f as
	 * separate flags, instead of -pf.  sigh.
	 */
	if (flags != NULL && flags[0] != '\0') {
	    char *p,*q;

	    strcpy(q = flagbuf,flags);

	    while ((p = strchr(q,' ')) != NULL) {
		*p++ = '\0';
		argv[argc++] = q;
		q = p;
	    }

	    argv[argc++] = q;
	}

	if (username != NULL)
	    argv[argc++] = username;

	/* add strings to environment as required
	 * some "init" programs export "HOME=/", remove it
	 */
	putenv("HOME");

#ifdef SETTERM
	sprintf(MsgBuf, "TERM=%s", Term);
	putenv(strdup(MsgBuf));
	if (loginenv)
	    argv[argc++] = strdup(MsgBuf);
#endif	/* SETTERM */

	if (Connect != NULL) {
	    sprintf(MsgBuf, "CONNECT=%s", Connect);
	    putenv(strdup(MsgBuf));
	    if (loginenv)
		argv[argc++] = strdup(MsgBuf);
	}

	if (CndName != NULL) {
	    sprintf(MsgBuf, "CNDNAME=%s", CndName);
	    putenv(strdup(MsgBuf));
	    if (loginenv)
		argv[argc++] = strdup(MsgBuf);
	}

	if (CndNumber != NULL) {
	    sprintf(MsgBuf, "CNDNUMBER=%s", CndNumber);
	    putenv(strdup(MsgBuf));
	    if (loginenv)
		argv[argc++] = strdup(MsgBuf);
	}

	if (ToNumber != NULL) {
	    sprintf(MsgBuf, "TONUMBER=%s", ToNumber);
	    putenv(strdup(MsgBuf));
	    if (loginenv)
		argv[argc++] = strdup(MsgBuf);
	}

#ifdef DEBUG
	if (Dfp != NULL) {
	    char **p;
	    extern char **environ;

	    for (p = environ; *p != NULL; p++)
		debug3(D_RUN, "environment: %s\n", *p);

	    for (p = argv; *p != NULL; p++)
		debug3(D_RUN, "argv: %s\n", *p);

	    debug2(D_RUN, "exec login program...\n");
	    fclose(Dfp);
	    Dfp = NULL;
	}
#endif	/* DEBUG */

	/* hand off to login, which can be a shell script!
	 * some systems cannot execv shell scripts directly...
	 */
	alarm((unsigned) 0);
	nice(9);
	execv(login,argv);
#ifndef __linux__
	strcpy(ChatBuf,login);
	for (argc = 1; argv[argc] != NULL; argc++)
	    sprintf(ChatBuf + strlen(ChatBuf)," '%s'",argv[argc]);
	execl("/bin/sh", "sh", "-c", ChatBuf, NULL);
#endif

	/* when we get here, exec has failed
	 */
	sprintf(MsgBuf, "cannot execute %s", login);
	logmesg(MsgBuf);
	return FAIL;
}


/*
**	protect() - check username and optionally ask for dialup password
**		    and/or dial back
**
**		    when askpasswd is FALSE, no password is requested and
**		    only entries without password can be valid.
**		    (useful to validate ppp access by phonenumber)
*/

int
protect(username,askpasswd)
char *username;
int askpasswd;

{
	FILE *fp;
	int match,rv,i;
	struct termios termios;
	char name[MAXLINE+1],passwd[MAXLINE+1],dialback[MAXLINE+1];
	char buf[MAXLINE+1];

	/* when PROTECT not defined, no validation
	 */
	if (Protect == NULL)
	    return SUCCESS;

	/* find logged-in username in PROTECT file
	 */
	if ((fp = fopen(Protect,"r")) == NULL) {
	    sprintf(MsgBuf, "cannot open %s", Protect);
	    logmesg(MsgBuf);
	    Fputs(MsgBuf,STDOUT);
	    Fputs("\n",STDOUT);
	    return FAIL;
	}

	while (match = -1, fgets(buf,MAXLINE,fp) != NULL &&
	       (buf[0] == '#' || (match = sscanf(buf,"%[^:\n]:%[^:\n]:%[^:\n]",
						 name,passwd,dialback)) > 0))
	{
	    if (match > 0)
	    {
		char *cnd;
		int l1,l2;

		debug5(D_GETL, "protect file: match %d name=(%s)\n",
			    match, name, passwd);

		if ((cnd = strchr(name,'/')) != NULL)	/* name/cnd ? */
		    *cnd++ = '\0';		/* strip cnd and point to it */

		/* check if name equal and calling number equal to */
		/* cnd info (after /).	tail of number is matched */

		if (!strcmp(name,username) &&
		    (cnd == NULL ||
		     (CndNumber != NULL &&
		      (l1 = strlen(CndNumber)) >= (l2 = strlen(cnd)) &&
		      !strcmp(CndNumber + (l1 - l2),cnd))
		    )
		   ) {
		    debug2(D_GETL, "got it!\n");
		    break;
		}
	    }
	}

	fclose(fp);

	if (match <= 0) {
	    if (askpasswd) {
		debug2(D_GETL, "not found, ask for passwd anyway...\n");

		Fputs("Password:",STDOUT);	/* ask for password */
		getpasswd(buf, MAXPASSWD);
	    }

	    sprintf(MsgBuf, "login \"%s\" attempted, unknown", username);
	    logmesg(MsgBuf);
	    return INCORRECT;
	}

	/* we have found a matching entry
	 * see if a password has to be entered for it
	 */
	if (match > 1 && strcmp(passwd,"*")) {
	    if (!askpasswd)
		return INCORRECT;
passwd_prompt:
	    Fputs("Password:",STDOUT);

	    switch (getpasswd(buf, MAXPASSWD))
	    {
	    case NONAME:
		goto passwd_prompt;

	    case SUCCESS:
		if (strcmp(passwd,buf)) {
		    debug3(D_GETL, "incorrect passwd (%s)\n", buf);
		    sprintf(MsgBuf, "login \"%s\" attempted, incorrect passwd",
				username);
		    logmesg(MsgBuf);
		    return INCORRECT;
		}
		break;

	    default:
		return FAIL;
	    }
	}

	/* now the entered password was okay (or there wasn't any)
	 * check if we need to dial back to this user...
	 */
	if (match > 2 && strcmp(dialback,"UUCP_U")) {
	    if (!strcmp(dialback,"CNDNUMBER")) {
		if (CndNumber != NULL) {
		    strcpy(dialback,CndNumber);
		} else {
		    Fputs("No number to call back\n",STDOUT);
		    sleep(3);
		    return FAIL;
		}
	    }

	    Fputs("Hangup the phone, you will be called back\n",STDOUT);
	    sleep(3);

	    rv = FAIL;

	    for (i = 0; i < DialTries; i++) {
		/* hangup and init the MODEM
		 */
		if (initmodem(TRUE) != SUCCESS)
		    return FAIL;

		/* make sure hangup is complete and defeat hackers who try
		 * to collide with the dialback
		 */
		sleep(5 + ((unsigned)rand() % 30));

		/* dial back in datamode
		 */
		rv = dialup("",(WFclass < 0? -1 : 0),3,dialback);
		debug3(D_RUN, "protect: dialup() returned %d\n",rv);

		if (rv == SUCCESS || rv == RING)
		    break;
	    }

	    if (rv != SUCCESS)
		return rv;

	    /* now ask again for a username, as he may want to login
	     * under another name
	     */
	    WarnCase = TRUE;		/* again warn for uppercase */
	    settermios(INITIAL);	/* back to normal handling */

login_prompt:
	    Fputs("@S!login: ",STDOUT);
	    termios = Termios;
	    switch (getlogname(&termios, username, 80))
	    {
	    case SUCCESS:
	    case BBSCALL:
		break;			/* okay, continue with this name */

	    case BADSPEED:
	    case NONAME:
		goto login_prompt;

	    case BADCASE:
		/* first try was all uppercase
		 */
		for (i=0; bad_case[i] != NULL; i++)
			Fputs(bad_case[i],STDOUT);
		goto login_prompt;

	    default:
		return FAIL;		/* some problem, give up */
	    }
	}

	return SUCCESS;
}


/*
**	timeout() - handles SIGALRM while in login loop
*/

sigh_t
timeout(sig)
int sig;
{
#ifdef TIOCMGET
	int mstat;
#endif

	/* log it
	 */
	sprintf(MsgBuf, "Timed out after %d seconds", TimeOut);
	logmesg(MsgBuf);

#ifdef TIOCMGET
	/* make sure CTS is active so we can still send some data
	 */
	if ((mstat = modemstat()) != -1 &&
	    !(mstat & TIOCM_CTS) &&
		(sleep(3),!(modemstat() & TIOCM_CTS)))
	{
	    /* CTS not active, just exit without message */
	    exit(EXIT_FAILURE);
	}
#endif

	/* say bye-bye
	 */
	Fputs("\n",STDOUT);
	Fputs(MsgBuf,STDOUT);
	Fputs(".\nBye Bye.\n",STDOUT);
	sleep(3);

	/* force a hangup
	 */
	settermios(DROPDTR);

	exit(EXIT_FAILURE);
}


/*
**	getlogname() - get the users login response
**
**	Returns int value indicating success.
**	If a fido-protocol is detected, return the name of the protocol.
**	Modify passed termios struct according to input conventions used.
*/

int
getlogname(termios, name, size)
struct termios *termios;
char *name;
int size;
{
	int count;
	int lower = 0;
	int upper = 0;
	char ch,*p;

	debug2(D_GETL, "getlogname() called\n");

	p = name;	/* point to beginning of buffer */
	count = 0;	/* nothing entered yet */

	do
	{
	    if (read(STDIN,&ch,1) != 1) {	/* get one char */
		debug2(D_GETL, "returned EOF\n");
		return FAIL;
	    }

	    if ((ch & 0200) && (unsigned char)ch != TSYNC && (unsigned char)ch != YOOHOO) {
		termios->c_iflag |= ISTRIP;	/* better strip his bytes... */
		ch &= 0x7f;
	    }

	    switch ((unsigned char)ch)
	    {
	    case CQUIT:				/* user wanted out, i guess */
		debug2(D_GETL, "returned QUIT\n");
		return FAIL;

	    case '\0':				/* BREAK? */
	    case CINTR:
		debug2(D_GETL, "returned BADSPEED\n");
		return BADSPEED;

#ifdef FIDO
	    case TSYNC:
		if (Fido != NULL) {
		    strcpy(name,"tsync");
		    debug2(D_GETL, "returned TSYNC\n");
		    return FIDOCALL;
		}
		break;

	    case YOOHOO:
		if (Fido != NULL) {
		    strcpy(name,"yoohoo");
		    debug2(D_GETL, "returned YOOHOO\n");
		    return FIDOCALL;
		}
		break;
#endif

	    case ESC:				/* for frontdoor addicts */
		termios->c_iflag |= ICRNL;	/* turn on cr/nl xlate */
		termios->c_oflag |= ONLCR;
		Fputs("bbs\n",STDOUT);
		strcpy(name,"bbs");
		debug2(D_GETL, "returned ESC\n");
		return BBSCALL;

	    case BS:
#if CINTR != DEL
	    case DEL:				/* alternate way to backspace */
#endif
#if CERASE != BS && CERASE != DEL
	    case CERASE:
#endif
		termios->c_cc[VERASE] = ch;
		if (count) {
		    if (!(termios->c_lflag & ECHOE))
			    Fputs("\b \b",STDOUT);
		    --p;
		    --count;
		}
		break;

	    case CKILL:
	    case CWERASE:
		if (!(termios->c_lflag & ECHOK))
		    Fputs("\n",STDOUT);

		p = name;
		count = 0;
		break;

	    case '\r':
		termios->c_iflag |= ICRNL;	/* turn on cr/nl xlate */
		termios->c_oflag |= ONLCR;
		break;

	    case '\n':
		break;

	    case 0x7e:				/* Async PPP 'flag' */
		if (p == name) {		/* only as first char */
		    strcpy(name,"ppp");
		    debug2(D_GETL, "returned PPPCALL\n");
		    return PPPCALL;
		}
		/* fall through... */

	    case CEOF:
		if (p == name)			/* ctrl-d was first char */
		    return FAIL;
		/* fall through... */

	    default:				/* any normal character... */
		if (isspace(ch) || !isprint(ch))
		    break;			/* ignore dirty stuff */

		if (!(termios->c_lflag & ECHO))
		    write(STDOUT, &ch, 1);

		*p++ = ch;
		count++;
		if (islower(ch))
		    lower++;
		if (isupper(ch))
		    upper++;
		break;
	    }
#ifdef FIDO
	    /* it looks like not all software sends a CR after this... */

	    if (Fido != NULL && count == 14 && !strncmp(name,"**EMSI_INQC816",14))
		break;
#endif
	} while ((ch != '\n') && (ch != '\r') && (count < size));

	*p = '\0';				/* terminate buffer */
	Fputs("\n",STDOUT);

	if (name[0] == '\0') {
		debug2(D_GETL, "returned NONAME\n");
		return NONAME;
	}

	if (upper && !lower && strncmp(name,"**EMSI",6)) {
		if (WarnCase) {
			WarnCase = FALSE;
			debug3(D_GETL, "returned BADCASE (%s)\n", name);
			return BADCASE;
		}

		termios->c_iflag |= IUCLC;
		termios->c_oflag |= OLCUC;
		termios->c_lflag |= XCASE;
	}

	debug3(D_GETL, "returned SUCCESS, name=(%s)\n", name);
	return SUCCESS;
}


/*
**	getpasswd() - get the users password response
**		      it is not echoed, but erase/kill are handled
*/

int
getpasswd(passwd, size)
char *passwd;
int size;
{
	int count;
	char ch,*p;

	debug2(D_GETL, "getpasswd() called\n");

	p = passwd;	/* point to beginning of buffer */
	count = 0;	/* nothing entered yet */

	do
	{
	    if (read(STDIN,&ch,1) != 1)		/* get one char */
		return FAIL;

	    ch &= 0x7f;				/* strip high bit */

	    switch (ch)
	    {
	    case CQUIT:				/* user wanted out, i guess */
		return FAIL;

	    case '\0':				/* BREAK? */
	    case CINTR:
		debug2(D_GETL, "returned (BADSPEED)\n");
		return BADSPEED;

	    case BS:
#if CINTR != DEL
	    case DEL:
#endif
#if CERASE != BS && CERASE != DEL
	    case CERASE:
#endif
		if (count) {
		    --p;
		    --count;
		}
		break;

	    case CKILL:
	    case CWERASE:
		p = passwd;
		count = 0;
		break;

	    case '\r':
	    case '\n':
		break;

	    case CEOF:
		if (p == passwd)		/* ctrl-d was first char */
		    return FAIL;
		/* fall through... */

	    default:				/* any normal character... */
		if (isspace(ch) || !isprint(ch))
		    break;			/* ignore dirty stuff */

		*p++ = ch;
		count++;
		break;
	    }
	} while ((ch != '\n') && (ch != '\r') && (count < size));

	*p = '\0';				/* terminate buffer */
	Fputs("\n",STDOUT);

	if (passwd[0] == '\0') {
		debug2(D_GETL, "returned NONAME\n");
		return NONAME;
	}

	debug2(D_GETL, "returned SUCCESS\n");
	return SUCCESS;
}


#ifdef ZFAX
/*************************************************************************
**	ZFAX mode handling						**
*************************************************************************/

/*
**	faxmode() - main handling of FAX calls (ZFAX mode)
**
**	'answer' can be one of:
**
**	ONLINE		the CONNECT is assumed to be already received
**	ANSWER		answer the call first, wait for CONNECT
*/

int
faxmode(answer)
int answer;

{
	debug2(D_RUN, "FAX mode\n");

	return faxrecv(answer,-1);
}

/*
**	faxpoll() - poll a FAX and receive any data it may have (ZFAX mode)
*/

int
faxpoll(number,pollid,logfd)
char *number;
char *pollid;
int logfd;

{
	if (!Is_ZyXEL || !Fax)			/* ZFAX requires ZyXEL */
	    return FAIL;

	/* dialup in FAX POLLING mode
	 */
	if (dialup("H#F#B2#V1#T1#L2#C7",6,5,number) != SUCCESS) {
	    if (!strncmp(Match,"DISCONNECT1",11))
		return NOPOLL;

	    return DIALFAIL;
	}

	return faxrecv(FAXPOLL,logfd);
}

/*
**	faxrecv() - receive a FAX and write it to a file (ZFAX mode)
*/

int
faxrecv(answer,logfd)
int answer;
int logfd;

{
	register FILE *fp;
	register int ch;
	const unsigned char *rtc;
	int ematch,rmatch,rtclen;
	FAX_HEADER header;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	char filename[MAXLINE];
	static const char end[] = "\r\nDISCONNECT";	/* result code */
#define NMATCH	12

	if (!Is_ZyXEL || !Fax)			/* ZFAX requires ZyXEL */
	    return FAIL;

	if (answer == ANSWER) {
	    int n;

	    /* set parameters and select FAX mode
	     * +VNH=1	no hangup when connect attempt fails (ZyXEL)
	     * Mute	sets muting for answer (e.g. M0)
	     */
	    sprintf(ChatBuf,"AT%s%sS38.4=0#F#B0+FCLASS=6\r",
			(Is_ZyXEL? "+VNH=1" : ""),Mute);
	    send(ChatBuf);
	    Fclass = 6;
	    if (waitfor("OK",MEDMTIME))
		return FAIL;

	    /* answer the call
	     */
	    send("ATA\r");

	    /* now wait for connection, may take some time
	     */
	    for (n = 5; n > 0; n--)
		if (!waitfor(NULL,LONGTIME)) {
		    rip(Match);

		    if (!strncmp(Match,"CONNECT",7)) /* prefix match only */
			break;

		    /* got something else than expected */

		    logmesg(Match);		/* log what we got (NO CARRIER?) */
		    debug3(D_RUN, "faxmode() - answer failed (%s)\n", Match);
		    return FAIL;
		}

	    if (!n)				/* nothing received */
		return FAIL;

	    logmesg(Match);			/* log the CONNECT we got! */
	    free(Connect);
	    Connect = strdup(Match);		/* and save it */
	}

	/* check if we have a FAX connection */

	if (strncmp(Connect + 8, "FAX", 3))	/* is it a FAX connect? */
	    return DATA;			/* no, forget FAX receive */

	if (logfd >= 0) {
	    sprintf(ChatBuf,"\t%s\r\n",Connect);
	    write(logfd,ChatBuf,strlen(ChatBuf));
	}

	/* create filename for recording, add extension */

	if (FaxDir != NULL && FaxFile[0] != '/') {
	    sprintf(filename,"%s/%s",FaxDir,FaxFile);

	    if (strcmp(filename + strlen(filename) - 4,FaxExt))
		strcat(filename,FaxExt);
	} else
	    strcpy(filename,FaxFile);

	/* %-macro's will be expanded using strftime, to generate */
	/* a time-dependent recording file */

	if (strchr(filename,'%') != NULL) {
	    char buf[MAXLINE];

	    RecTime = time(NULL);
	    strftime(buf,sizeof(buf),filename,localtime(&RecTime));
	    strcpy(filename,buf);
	}

	if ((fp = fopen(filename,"w")) == NULL) {
	    sprintf(MsgBuf, "cannot create faxfile (%s)", filename);
	    logmesg(MsgBuf);
	    return FAIL;
	}

	/* trap interrupts received during receive to exit neatly */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	faxheader(&header,Connect);		/* prepare a FAX file header */
	fwrite(&header,sizeof(header),1,fp);

	settermios(LONGTIME);			/* prepare for 15-sec timeout */
	ematch = rmatch = 0;

	if (header.flags & FAX_T1) {
	    rtc = fax_rtc_2d;
	    rtclen = RTC_2D_LEN;
	} else {
	    rtc = fax_rtc_1d;
	    rtclen = RTC_1D_LEN;
	}

	if (logfd >= 0)
	    write(logfd,"\t[page data]\r\n",14);

	/* copy data to outputfile until timeout
	 * match the RTC string, and increment pagecount when it is found
	 * match the end string, and terminate when it is found
	 */
	while ((ch = getch()) != EOF && !Signal) {
	    if (ch == rtc[rmatch]) {
		if (++rmatch == rtclen) {
		    /* matched entire RTC string, end of page */

		    header.pages++;
		    rmatch = 0;
		}
	    } else {
		if (ch == rtc[rmatch = 0])
		    rmatch++;
	    }

	    if (ch == end[ematch]) {
		/* match of endstring, check if we have matched entire
		 * string
		 */
		if (++ematch == NMATCH) {
		    char *p;

		    /* matched entire end string, assume FAX has ended
		     * transfer end string to MsgBuf (up to CR)
		     */
		    strcpy(MsgBuf,end + 2);	/* copy except CR/LF */
		    p = MsgBuf + NMATCH - 2;

		    while ((ch = getch()) != EOF && ch != '\r')
			if (p < (MsgBuf + sizeof(MsgBuf) - 2))
			    *p++ = ch;

		    *p = '\0';

		    while (--p >= MsgBuf && *p == ' ')
			*p = '\0';

		    logmesg(MsgBuf);		/* log the end string */

		    if (logfd >= 0) {
			sprintf(ChatBuf,"\t%s\r\n",MsgBuf);
			write(logfd,ChatBuf,strlen(ChatBuf));
		    }

		    waitfor(NULL,SHORTTIME);	/* flush CR/LF */
		    break;
		}
	    } else {
		/* mismatch, write possibly-matched data
		 */
		if (ematch != 0) {
		    fwrite(end,1,ematch,fp);

		    if (ch == end[ematch = 0]) {
			ematch++;
			continue;
		    }
		}

		putc(ch,fp);
	    }
	}

	fflush(fp);
	ch = ftell(fp) - sizeof(header);	/* see how much we wrote */

	fseek(fp,0L,SEEK_SET);			/* re-write (modified) header */
	fwrite(&header,sizeof(header),1,fp);

	if (!!ferror(fp) | fclose(fp) == EOF)
	    logmesg("Error writing FAX file");

	if (ch == 0) {
	    unlink(filename);
	    if (answer != FAXPOLL)
		logmesg("FAX call, no data recorded");
	} else {
	    sprintf(ChatBuf,"FAX %s, %d pages received",
		    (answer == FAXPOLL? "poll":"call"),header.pages);
	    logmesg(ChatBuf);
	}

	send("ATH\r");
	waitfor("OK",MEDMTIME);
	CurLS = Fclass = 0;

	if (ch > 0 && FaxMail != NULL) {
	    sprintf(ChatBuf, "%s %s", Mailer, FaxMail);
	    if ((fp = popen(ChatBuf, "w")) != NULL) {
		fprintf(fp, "From: %s %s\n", MyName, Device);
		fprintf(fp, "To: %s\n", FaxMail);
		fprintf(fp, "Subject: FAX arrived\n\n");
		fprintf(fp, "A %d-page FAX has arrived from: %s\n",
				header.pages, Connect + 28);
		if (CndName != NULL)
		    fprintf(fp, "NAME: %s\n", CndName);
		if (CndNumber != NULL)
		    fprintf(fp, "NUMBER: %s\n", CndNumber);
		if (ToNumber != NULL)
		    fprintf(fp, "TO: %s\n", ToNumber);
		fprintf(fp, "\n%d bytes have been recorded to file %s\n",
				ch, filename);
		pclose(fp);
	    }
	}

	signal(SIGINT, oldint);
	signal(SIGQUIT, oldquit);
	signal(SIGTERM, oldterm);
	signal(SIGALRM, oldalarm);
	Signal = 0;

	return (MsgBuf[10] == '0')? SUCCESS : XFERFAIL;
}

/*
**	faxsend() - send a FAX (ZFAX mode)
*/

int
faxsend(number,filename,logfd)
char *number;
char *filename;
int logfd;

{
	FILE *fp;
	int res,num;
	long arg;
	FAX_HEADER header,connect;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	char buf[BUFSIZ];

	if (!Is_ZyXEL || !Fax)			/* ZFAX requires ZyXEL */
	    return FAIL;

	/* open the file */

	if ((fp = fopen(filename,"r")) == NULL) {
	    sprintf(MsgBuf, "cannot open faxfile (%s)", filename);
	    logmesg(MsgBuf);
	    return FAIL;
	}

	/* read and check the file header */

	if (fread(&header,sizeof(header),1,fp) != 1 ||
	    memcmp(header.zyxel,FAX_MAGIC,sizeof(header.zyxel)))
	{
	    fclose(fp);
	    return FAIL;
	}

	/* prepare the mode string from FAX header data */

	switch (header.rec_width)
	{
	case FAX_R1:
	    res = 1;
	    break;

	case FAX_R2:
	    res = 2;
	    break;

	default:
	    res = 0;
	    break;
	}

	sprintf(buf,"H#F#B0#V%d#T%d#R%d#L2#C7",
		((header.flags & FAX_V1)? 1 : 0),
		((header.flags & FAX_T1)? 1 : 0),
		res);

	/* dialup in FAX SENDING mode
	 */
	if (dialup(buf,6,5,number) != SUCCESS)
	    return DIALFAIL;

	if (logfd >= 0) {
	    sprintf(ChatBuf,"\t%s\r\n",Connect);
	    write(logfd,ChatBuf,strlen(ChatBuf));
	}

	/* analyze the CONNECT to see if it is compatible with our header
	 */
	faxheader(&connect,Connect);

	if (header.rec_width > connect.rec_width ||
	    header.flags != connect.flags)
	{
	    fclose(fp);
	    logmesg("FAX CONNECT incompatible with file");
	    return CONNECTFAIL;
	}

	/* trap interrupts received during sending to exit neatly */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	/* everything OK, start sending the FAX
	 */
	settermios(IMMEDIATE);			/* immediate return on read */

	if (logfd >= 0)
	    write(logfd,"\t[page data]\r\n",14);

	while ((num = fread(buf,1,sizeof(buf),fp)) > 0 && !Signal) {
	    if (write(STDOUT,buf,num) != num)
		break;

	    switch (res = getch())		/* something returned? */
	    {
	    case EOF:				/* nothing */
	    case XON:				/* leaked flow control char */
	    case XOFF:
		break;				/* ignore it and go on */

	    default:				/* something else */
		debug3(D_RUN, "faxsend() terminated by unexpected char %02x\n", res);
		InPtr--;			/* push it back */
		waitfor(NULL,SHORTTIME);	/* and flush the input */
		goto quit;			/* stop sending */
	    }
	}
quit:
	fclose(fp);

	/* end of file, send an extra RTC in case the file did not
	 * contain one (hope this does not result in an empty page...)
	 */
	if (header.flags & FAX_T1)
	    write(STDOUT,fax_rtc_2d,RTC_2D_LEN);
	else
	    write(STDOUT,fax_rtc_1d,RTC_1D_LEN);

	/* all sent, now we must lower RTS to signal the end of the FAX
	 * it is better to first wait until all data sent...  I'm afraid
	 * we'll interrupt the transfer when RTS is made low while sending.
	 */

	settermios(IMMEDIATE);			/* wait for data to be sent */
	settermios(XONXOFF);			/* no RTS/CTS for now... */
	usleep(20000);
	arg = TIOCM_RTS;			/* RTS OFF */
	ioctl(STDIN,TIOCMBIC,&arg);

	/* now wait for the result, and see what it is
	 */
	for (res = 0; res < 5; res++)
	    if (waitfor(NULL,LONGTIME) != FAIL) /* read result */
		break;

	arg = TIOCM_RTS;			/* RTS ON after response */
	ioctl(STDIN,TIOCMBIS,&arg);
	settermios(INITIAL);			/* enable RTS/CTS handshake */

	rip(Match);
	strcpy(MsgBuf,Match);			/* save result */
	logmesg(MsgBuf);			/* log the result string */

	if (logfd >= 0) {
	    sprintf(ChatBuf,"\t%s\r\n",MsgBuf);
	    write(logfd,ChatBuf,strlen(ChatBuf));
	}

	send("ATH\r");				/* just to make sure... */
	waitfor("OK",MEDMTIME);
	CurLS = Fclass = 0;

	signal(SIGINT, oldint);
	signal(SIGQUIT, oldquit);
	signal(SIGTERM, oldterm);
	signal(SIGALRM, oldalarm);
	Signal = 0;

	return ((int)strlen(MsgBuf) > 10 && MsgBuf[10] == '0')? SUCCESS : XFERFAIL;
}


/*
**	faxheader() - convert CONNECT info to FAX file header (ZFAX mode)
**
**	page count will be left at zero
*/

void
faxheader(header,connect)
FAX_HEADER *header;
char *connect;

{
	char *p;

	/* clear header and put fixed values */

	memset(header,0,sizeof(FAX_HEADER));
	memcpy(header->zyxel,FAX_MAGIC,sizeof(header->zyxel));
	header->factory = 'Z';
	header->version = 2;

	if (connect != NULL) {
	    if ((p = strchr(connect + 11,'V')) != NULL)
		if (p[1] != '0')
		    header->flags |= FAX_V1;	/* high resolution */

	    if ((p = strchr(connect + 11,'T')) != NULL)
		if (p[1] != '0')
		    header->flags |= FAX_T1;	/* 2-D encoding */

	    if ((p = strchr(connect + 11,'R')) != NULL)
		switch (p[1])
		{
		case '0':
		    header->rec_width = FAX_R0; /* 1728 pixels */
		    break;

		case '1':
		    header->rec_width = FAX_R1; /* 2048 pixels */
		    break;

		case '2':
		    header->rec_width = FAX_R2; /* 2432 pixels */
		    break;
		}
	}
}
#else
/*************************************************************************
**	CLASS2 FAX mode handling					**
*************************************************************************/

/*
**	faxmode() - main handling of FAX calls (CLASS 2 mode)
**
**	'answer' can be one of:
**
**	ONLINE		the +FCON is assumed to be already received
**	ANSWER		answer the call first, wait for +FCON
**	FAXPOLL		used when polling
*/

int
faxmode(answer)
int answer;

{
	char *name;
	int logfd = -1;
	int rv;
	char tmpname[MAXLINE];

	debug2(D_RUN, "FAX mode\n");

	if (!Fax)
	    return FAIL;

	/* create a logfile for the FAX mode responses */

	sprintf(tmpname,"%s/%sXXXXXX",P_tmpdir,MyName);

	if ((name = mktemp(tmpname)) != NULL && name[0] != '\0')
	    logfd = open(name,O_RDWR|O_CREAT|O_TRUNC,0600);

	/* receive the FAX and remove the logfile */

	rv = faxrecv(answer,logfd);

	if (logfd >= 0) {
	    close(logfd);
	    unlink(name);
	}

	return rv;
}

/*
**	faxpoll() - poll a FAX and receive any data it may have (CLASS 2 mode)
*/

int
faxpoll(number,pollid,logfd)
char *number;
char *pollid;
int logfd;

{
	int res;

	if (!Fax)
	    return FAIL;

	sprintf(ChatBuf,"AT%s\r",make_fclass(2));
	if (waitfor("OK",MEDMTIME) && !(Fw_features & FW_FAXBUG))
	    return FAIL;

	/* set polling station ID
	 */
	if (pollid != NULL) {
	    send("AT+FNS=\"\"\r");		/* reset to null string */
	    if (waitfor("OK",MEDMTIME))
		return FAIL;

	    sprintf(ChatBuf,"AT+FNS=\"%s",poll_hex);
	    while (*pollid != '\0')
		sprintf(ChatBuf + strlen(ChatBuf),"%02X",*pollid++);
	    strcat(ChatBuf,"\"\r");

	    send(ChatBuf);			/* send nonstandard string */
	    if (waitfor("OK",MEDMTIME))
		return FAIL;
	}

	/* set local ID for polling (necessary?)
	 */
	if (FaxId != NULL) {
	    sprintf(ChatBuf,"AT+FCIG=\"%.20s\"\r",FaxId);
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME))
		return FAIL;
	}

	/* dialup in FAX POLLING mode
	 */
	if (dialup("H+FSPL=1",2,5,number) != SUCCESS)
	    return DIALFAIL;

	res = faxrecv(FAXPOLL,logfd);

	send("AT+FSPL=0\r");
	waitfor("OK",MEDMTIME);

	return res;
}

/*
**	faxrecv() - receive a FAX and write it to a file (CLASS2 mode)
*/

int
faxrecv(answer,logfd)
int answer;
int logfd;

{
	register FILE *fp;
	register int ch;
	int pages,bytes;
	char *id;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	char dirname[MAXLINE],filename[MAXLINE];

	if (!Fax)
	    return FAIL;

	if (answer == ANSWER) {
	    int n;

	    /* ignore BREAK during connection
	     * it seems to be triggered sometimes?
	     */
	    oldint = signal(SIGINT, SIG_IGN);

	    /* set parameters and select FAX mode
	     * +VNH=1	no hangup when connect attempt fails (ZyXEL)
	     * Mute	sets muting for answer (e.g. M0)
	     */
	    sprintf(ChatBuf,"AT%s%s\r",(Is_ZyXEL? "S38.4=0+VNH=1":""),Mute);
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME)) {
#ifdef DEBUG
		linestat();
#endif
		return FAIL;
	    }
	    sprintf(ChatBuf,"AT%s+FAA=0\r",make_fclass(2));
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME) && !(Fw_features & FW_FAXBUG)) {
#ifdef DEBUG
		linestat();
#endif
		return FAIL;
	    }

	    /* answer the call
	     */
	    send("ATA\r");

	    /* now wait for connection, may take some time
	     */
	    for (n = 5; n > 0; n--)
		if (!waitfor(NULL,LONGTIME)) {
		    signal(SIGINT, oldint);
		    rip(Match);

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\t%s\r\n",Match);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    if (!strcmp(Match,"+FCON"))
			break;

		    /* got something else than expected */

		    logmesg(Match);		/* log what we got (NO CARRIER?) */
		    debug3(D_RUN, "faxmode() - answer failed (%s)\n", Match);
#ifdef DEBUG
		    linestat();
#endif
		    return FAIL;
		}

	    signal(SIGINT, oldint);

	    if (!n)				/* nothing received */
		return FAIL;
	}

	/* get the connection info
	 */
	FaxInfo = blank_info;			/* set everything to unknown */

	if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"OK"))
	    return FAIL;

	/* log it for later reference
	 */
	sprintf(MsgBuf,"FAX connect: DCS=%d,%d,%d,%d,%d,%d,%d,%d ID=%s %s",
		FaxInfo.vr,FaxInfo.br,FaxInfo.wd,FaxInfo.ln,
		FaxInfo.df,FaxInfo.ec,FaxInfo.bf,FaxInfo.st,
		FaxInfo.id,(id = faxident()));
	logmesg(MsgBuf);

	if (id[0]) {
	    sprintf(MsgBuf,"\t%s\r\n",id);
	    write(logfd,MsgBuf,strlen(MsgBuf));
	}

	/* are we polling?
	 */
	if (answer == FAXPOLL) {
	    if (!FaxInfo.poll) {
		send("ATH\r");
		CurLS = Fclass = 0;
		fax2rsp(logfd);
		return NOPOLL;			/* no fax available for polling */
	    }
	} else {
	    int rv;

	    /* we are not polling, check if remote is polling us */

	    if ((rv = faxcpoll(logfd)) != NOPOLL)
		return rv;
	}

	/* create directory name for storing recorded pages */

	if (FaxDir != NULL && FaxFile[0] != '/')
	    sprintf(dirname,"%s/%s",FaxDir,FaxFile);
	else
	    strcpy(dirname,FaxFile);

	/* %-macro's will be expanded using strftime, to generate */
	/* a time-dependent recording directory */

	if (strchr(dirname,'%') != NULL) {
	    char buf[MAXLINE];

	    RecTime = time(NULL);
	    strftime(buf,sizeof(buf),dirname,localtime(&RecTime));
	    strcpy(dirname,buf);
	}

	/* make the directory to store the pages
	 */
	if (mkdir(dirname,0755) != 0) {
	    sprintf(MsgBuf, "cannot create faxdir (%s)", dirname);
	    logmesg(MsgBuf);
	    return FAIL;
	}

	/* trap interrupts received during receive to exit neatly */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	pages = bytes = 0;

	if (FaxRej != NULL)
	{
	    char *p;

	    strcpy(ChatBuf,FaxRej);
	    p = strtok(ChatBuf,",");

	    while (p != NULL) {
		if (!strcmp(p,"NULL")) {
		    if (!FaxInfo.id[0] && CndNumber == NULL) {
			logmesg("FAX call rejected (no ID)");
			goto end;
		    }
		} else {
		    if (!strncmp(FaxInfo.id,p,strlen(p)) ||
			(CndNumber != NULL && !strncmp(CndNumber,p,strlen(p)))) {
			sprintf(MsgBuf, "FAX call rejected (%s)", p);
			logmesg(MsgBuf);
			goto end;
		    }
		}
		
		p = strtok(NULL,",");
	    }
	}

	/* loop to receive all pages in the FAX message...
	 */
	do
	{
	    send("AT+FDR\r");			/* ready to receive a page */

	    if (logfd >= 0)
		write(logfd,"\tAT+FDR\r\n",9);

	    if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"CONNECT"))
		break;

	    if (faxcpoll(logfd) != NOPOLL)
		break;				/* fax polling detected */

	    if (!FaxInfo.cfr && pages == 0 && answer != FAXPOLL)
		logmesg("FAX receive: no +FCFR (trying to continue)");

	    sprintf(filename,"%s/p%02d_%c%d%s",
		dirname,pages + 1,(FaxInfo.vr? 'f':'n'),FaxInfo.df + 1,FaxExt);

	    if ((fp = fopen(filename,"w")) == NULL) {
		sprintf(MsgBuf, "cannot create faxfile (%s)", filename);
		logmesg(MsgBuf);
		break;
	    }

	    if (logfd >= 0)
		write(logfd,"\t[page data]\r\n",14);

	    /* don't switch to XON/XOFF mode here.  we are receiving */
	    /* binary data and eating these characters is not a good idea... */
	    /* let's simply assume we are fast enough to keep up */

	    send("\022");			/* send DC2 */
	    settermios(SHORT255);
	    debugtime(D_FAX);
	    debug2(D_FAX, "started FAX data receive\n");

	    /* read the FAX data, it is terminated by DLE-ETX */
	    /* or a timeout on receiving FAX data */

	    for (;;)
	    {
		switch (ch = getch())
		{
		case EOF:
		    debugtime(D_FAX);
		    debug2(D_FAX, "FAX data receive timeout\n");
		    goto done;

		case DLE:
		    switch (ch = getch())
		    {
		    case EOF:
			debugtime(D_FAX);
			debug2(D_FAX, "FAX data receive timeout\n");
			goto done;

		    case ETX:			/* end of message! */
			debugtime(D_FAX);
			debug2(D_FAX, "Received DLE-ETX in FAX data\n");
			goto done;

		    case DLE:			/* doubled DLE, just write */
			putc(ch,fp);
			break;

		    case SUB:			/* DLE-DLE (2864) */
			putc(DLE,fp);
			putc(DLE,fp);
			break;

		    default:			/* something else... */
			debugtime(D_FAX);
			debug3(D_FAX, "Received DLE-%02x in FAX data\n", ch);
			break;
		    }
		    break;

		default:
		    putc(ch,fp);		/* normal FAX data */
		    break;
		}
	    }
done:

	    ch = ftell(fp);			/* bytes received */

	    if (fclose(fp) == EOF) {
		logmesg("Error writing FAX file");
		ch = 0;
	    }

	    if (ch == 0)
		unlink(filename);
	    else {
		bytes += ch;
		pages++;
	    }

	    if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"OK"))
		break;

	    switch (FaxInfo.et)
	    {
	    case 1:				/* end of doc, another follows */
		FaxInfo.et = 0;			/* ignore, just continue */
		break;

	    case 2:				/* end of doc, no more follows */
		break;				/* stop */

	    case 4:				/* procedure interrupt, another page */
	    case 5:				/* procedure interrupt, end of doc, another follows */
	    case 6:				/* procedure interrupt, end of doc */
		sprintf(MsgBuf,"Procedure interrupt %d",FaxInfo.et);
		logmesg(MsgBuf);
		break;
	    }
	} while (FaxInfo.hng < 0 && FaxInfo.pts < 4 && FaxInfo.et == 0 && !Signal);

	/* see if they are polling us now...
	 */
	if (!Signal)
	    faxcpoll(logfd);

	/* when call not yet terminated, request completion status
	 */
	if (FaxInfo.hng < 0) {
	    send("AT+FDR\r");

	    if (logfd >= 0)
		write(logfd,"\tAT+FDR\r\n",9);

	    fax2rsp(logfd);
	}

	sprintf(ChatBuf,"FAX %s, %d pages received, result=%d",
		(answer == FAXPOLL? "poll":"call"),pages,FaxInfo.hng);
	logmesg(ChatBuf);

end:	if (bytes == 0)
	    rmdir(dirname);

	/* end of call, hangup
	 */
	send("ATH\r");
	fax2rsp(logfd);
	CurLS = Fclass = 0;

	/* generate mail when requested
	 */
	if (bytes > 0 && FaxMail != NULL) {
	    sprintf(ChatBuf, "%s %s", Mailer, FaxMail);
	    if ((fp = popen(ChatBuf, "w")) != NULL) {
		fprintf(fp, "From: %s %s\n", MyName, Device);
		fprintf(fp, "To: %s\n", FaxMail);
		fprintf(fp, "Subject: FAX arrived\n\n");
		fprintf(fp, "A %d-page FAX has arrived from: %s\n",
				pages, FaxInfo.id);
		if (CndName != NULL)
		    fprintf(fp, "NAME: %s\n", CndName);
		if (CndNumber != NULL)
		    fprintf(fp, "NUMBER: %s\n", CndNumber);
		if (ToNumber != NULL)
		    fprintf(fp, "TO: %s\n", ToNumber);
		fprintf(fp, "\n%d bytes have been recorded into directory %s\n",
				bytes, dirname);

		if (logfd >= 0 && !isatty(logfd) && lseek(logfd,0L,SEEK_SET) == 0)
		{
		    int n;

		    fprintf(fp,"\nTransfer log:\n");

		    while ((bytes = read(logfd,ChatBuf,sizeof(ChatBuf))) > 0)
			for (n = 0; n < bytes; n++)
			    if (ChatBuf[n] != '\r')
				fputc(ChatBuf[n],fp);
		}

		pclose(fp);
	    }
	}

	signal(SIGINT, oldint);
	signal(SIGQUIT, oldquit);
	signal(SIGTERM, oldterm);
	signal(SIGALRM, oldalarm);
	Signal = 0;

	return (FaxInfo.hng == 0)? SUCCESS : XFERFAIL;
}

/*
**	faxcpoll() - check if remote is polling for a FAX (CLASS 2 mode)
*/

int
faxcpoll(logfd)
int logfd;

{
	int pass,rv;
	char pollid[5],orgpoll[5];
	char filename[MAXLINE];

	if (FaxInfo.hng >= 0 || !FaxInfo.dtc)
	    return NOPOLL;			/* no poll request received */

	if (FaxPoll == NULL)
	    return NOPOLL;			/* poll request, but no doc */

	strcpy(pollid,"0000");			/* default poll-id */
	strcpy(MsgBuf,"FAX poll");
	orgpoll[0] = '\0';

	/* check for optional 4-digit poll id, which is passed as a
	 * nonstandard command.	 keep default 0000 when not found.
	 * the format of the nonstandard command is unclear. there is
	 * an example in a ZyXEL doc which uses the string in poll_nsc,
	 * but I have seen at least one FAX which sent something completely
	 * different.  in an attempt to decode many possibilities, a
	 * binary value 04 followed by 4 ASCII digits is also recognized
	 * as a POLL-ID.  The 4 digits are sent least-significant-first...
	 */

	if (FaxInfo.nstyp == 'C')
	{
	    unsigned char *p;
	    int idset = 0;

	    if (FaxInfo.nslen > sizeof(poll_nsc) && !memcmp(FaxInfo.ns,poll_nsc,sizeof(poll_nsc))) {
		memcpy(pollid,FaxInfo.ns + sizeof(poll_nsc),4);
		idset = 1;
	    } else
		for (p = FaxInfo.ns; p < FaxInfo.ns + FaxInfo.nslen - 4; p++)
		    if (p[0] == 0x04 &&
			isdigit(p[1]) && isdigit(p[2]) &&
			isdigit(p[3]) && isdigit(p[4]))
		    {
			pollid[0] = p[4];
			pollid[1] = p[3];
			pollid[2] = p[2];
			pollid[3] = p[1];
			idset = 1;
			break;
		    }

	    if (idset)
	    {
		strcpy(orgpoll,pollid);
		sprintf(MsgBuf + strlen(MsgBuf),", POLL-ID: %s",pollid);
	    }
	}

	logmesg(MsgBuf);

	if (logfd >= 0) {
	    sprintf(ChatBuf,"\t[%s]\r\n",MsgBuf);
	    write(logfd,ChatBuf,strlen(ChatBuf));
	}

	for (pass = 0; pass < 2; pass++)
	{
	    sprintf(filename,FaxPoll,pollid);	/* generate filename */

	    if (!access(filename,R_OK))
	    {
		rv = faxsend(NULL,filename,logfd);

		/* generate mail when requested
		 */
		if (FaxMail != NULL) {
		    FILE *fp;

		    sprintf(ChatBuf, "%s %s", Mailer, FaxMail);
		    if ((fp = popen(ChatBuf, "w")) != NULL) {
			fprintf(fp, "From: %s %s\n", MyName, Device);
			fprintf(fp, "To: %s\n", FaxMail);
			fprintf(fp, "Subject: FAX polled\n\n");
			fprintf(fp, "FAX has been polled by: %s\n", FaxInfo.id);
			if (orgpoll[0] != '\0')
			    fprintf(fp, "POLL-ID: %s\n", orgpoll);
			if (CndName != NULL)
			    fprintf(fp, "NAME: %s\n", CndName);
			if (CndNumber != NULL)
			    fprintf(fp, "NUMBER: %s\n", CndNumber);
			if (ToNumber != NULL)
			    fprintf(fp, "TO: %s\n", ToNumber);

			if (logfd >= 0 && !isatty(logfd) && lseek(logfd,0L,SEEK_SET) == 0)
			{
			    int n,bytes;

			    fprintf(fp,"\nTransfer log:\n");

			    while ((bytes = read(logfd,ChatBuf,sizeof(ChatBuf))) > 0)
				for (n = 0; n < bytes; n++)
				    if (ChatBuf[n] != '\r')
					fputc(ChatBuf[n],fp);
			}

			pclose(fp);
		    }
		}

		return rv;
	    }

	    strcpy(pollid,"0000");		/* try default poll-id */
	}

	return NOPOLL;				/* no pollfile found */
}

/*
**	faxsend() - send a FAX (CLASS 2 mode)
*/

int
faxsend(number,name,logfd)
char *number;
char *name;
int logfd;

{
	char *p,*q;
	int fd,ptr,num,rmatch,file,nfiles,vr,df,retry,fet,c;
	int rv = FAIL;
	int skip = 0;
	pid_t gpid,bpid,cpid;
	off_t pagepos = 0;
	struct dirent **namelist = NULL;
	struct stat st;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	char filename[MAXLINE];
	char buf[BUFSIZ];

	if (!Fax)
	    return FAIL;

	/* check if the passed name exists, and
	 * if it is a directory (in that case it contains the
	 * files that have to be sent)
	 */
	if (stat(name,&st) != 0) {
	    sprintf(MsgBuf, "cannot open faxfile (%s)", name);
	    logmesg(MsgBuf);
	    return FAIL;
	}

	if (S_ISDIR(st.st_mode)) {
	    /* get the file list in the directory passed to us */

	    /* this line may generate a warning for arg #3, depending on */
	    /* the version of <dirent.h> */
	    nfiles = scandir(name,&namelist,dir_select,alphasort);

	    if (nfiles == 0)
		return FAIL;
	} else {
	    strcpy(filename,name);
	    nfiles = -1;
	}

	/* trap interrupts received during sending to exit neatly */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	if (number != NULL) {
	    /* dialup in FAX SENDING mode
	     */
	    if (dialup("H",2,5,number) != SUCCESS) {
		rv = DIALFAIL;
		goto fail;
	    }

	    /* get the connection info
	     */
	    FaxInfo = blank_info;		/* set everything to unknown */

	    if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"OK") || Signal)
		goto fail;

	    /* log it for later reference
	     */
	    sprintf(MsgBuf,"FAX connect: DCS=%d,%d,%d,%d,%d,%d,%d,%d ID=%s %s",
		    FaxInfo.vr,FaxInfo.br,FaxInfo.wd,FaxInfo.ln,
		    FaxInfo.df,FaxInfo.ec,FaxInfo.bf,FaxInfo.st,
		    FaxInfo.id,(p = faxident()));
	    logmesg(MsgBuf);

	    if (p[0]) {
		sprintf(MsgBuf,"\t%s\r\n",p);
		write(logfd,MsgBuf,strlen(MsgBuf));
	    }
	}

	file = 0;

	do
	{
	    /* see what file we should send */

	    if (nfiles > 0) {
		sprintf(filename,"%s/%.*s",name,(int)namelist[file]->d_reclen,
						     namelist[file]->d_name);
		file++;
	    }

	    /* open the file */

	    if ((fd = open(filename,O_RDONLY)) < 0) {
		sprintf(MsgBuf, "cannot open faxfile (%s) errno=%d", filename, errno);
		logmesg(MsgBuf);
		goto fail;
	    }

	    /* read first chunk of data */

	    num = read(fd,buf,sizeof(buf));
	    ptr = 0;
	    gpid = bpid = cpid = 0;

	    /* determine file type from first data byte(s) */

	    switch (buf[0])
	    {
	    case 0:
		/* looks like a raw G3 file */

		if (logfd >= 0) {
		    sprintf(MsgBuf,"\tG3 file: %s\r\n",filename);
		    write(logfd,MsgBuf,strlen(MsgBuf));
		}

		/* guess vertical resolution and data format from name */

		if ((p = strrchr(filename,'_')) != NULL &&
		    (p[1] == 'n' || p[1] == 'f') && isdigit(p[2]))
		{
		    vr = (p[1] == 'f');
		    df = p[2] - '1';
		}
		else
		    vr = df = 0;
		break;

	    case 0x1f:
		if (Gunzip != NULL && GhostScript != NULL &&
		    ((unsigned char)buf[1] == 0x8b || (unsigned char)buf[1] == 0x9d))
		{
		    char *argv[2];

		    /* looks like a gzipped or compressed file */

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\tCompressed file: %s\r\n",filename);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    lseek(fd,0L,SEEK_SET);	/* rewind the file */
		    num = 0;
		    memset(argv,0,sizeof(argv)); /* build argvector for GUNZIP */

		    if (push(&cpid,&fd,Gunzip,argv) != SUCCESS) {
			sprintf(MsgBuf,"Cannot start %s",Gunzip);
			logmesg(MsgBuf);
			close(fd);
			goto fail;
		    } else {
			/* as we cannot rewind the pipe, we have to assume */
			/* that compressed files are always of the same type */

			goto postscript;
		    }
		} else
		    goto othertype;

		break;

	    case '%':
		if (GhostScript != NULL && !strncmp(buf,"%!PS-Adobe",10)) {
		    int argc;
		    char *argv[10];

		    /* looks like a PS file */

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\tPostscript file: %s\r\n",filename);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    lseek(fd,0L,SEEK_SET);	/* rewind the file */

postscript:
		    memset(argv,0,sizeof(argv)); /* build argvector for GS */
		    argc = 1;
		    argv[argc++] = (FaxInfo.vr? "-sDEVICE=dfaxhigh" : "-sDEVICE=dfaxlow");
		    argv[argc++] = "-sOutputFile=-";
		    argv[argc++] = "-q";
		    argv[argc++] = "-dNOPAUSE";
		    argv[argc++] = "-dSAFER";
		    argv[argc++] = "-";

		    /* start GhostScript process */

		    if (push(&gpid,&fd,GhostScript,argv) != SUCCESS) {
			sprintf(MsgBuf,"Cannot start %s",GhostScript);
			logmesg(MsgBuf);

			if (logfd >= 0) {
			    sprintf(ChatBuf,"\t%s\r\n",MsgBuf);
			    write(logfd,ChatBuf,strlen(ChatBuf));
			}

			close(fd);
			goto fail;
		    }

		    /* when Buffer is defined, start the buffering program */

		    if (Buffer != NULL) {
			memset(argv,0,sizeof(argv)); /* empty argvector */

			if (push(&bpid,&fd,Buffer,argv) != SUCCESS) {
			    sprintf(MsgBuf,"Cannot start %s",Buffer);
			    logmesg(MsgBuf);
			}
		    }

		    /* Let the above programs run for a while */

		    usleep(500000L);

		    /* cheat a bit, by forcing the page send loop below to
		     * start negotiating before doing the first read from
		     * the pipe.  that gains us some time for the GS process
		     * to format the first page while the modem negotiates
		     * the parameters with the remote FAX...  the 64-byte
		     * DigiFAX header will be skipped later.
		     */

		    skip = 64;
		    num = 0;
		    vr = FaxInfo.vr;
		    df = 0;
		} else
		    goto othertype;

		break;

	    case '#':
		if (GhostScript != NULL && buf[1] == '!' &&
		    !access(filename,X_OK))
		{
		    char *argv[2];

		    /* looks like a shell script */
		    /* we will execute it, and assume the output is Postscript */
		    /* data.  this can be achieved by including "pstext" as */
		    /* part of the script, if necessary */

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\tShell script: %s\r\n",filename);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    close(fd);
		    fd = open("/dev/null",O_RDWR); /* /dev/null as input */
		    num = 0;
		    memset(argv,0,sizeof(argv)); /* build argvector to exec */

		    if (push(&cpid,&fd,filename,argv) != SUCCESS) {
			sprintf(MsgBuf,"Cannot start %s",filename);
			logmesg(MsgBuf);
			close(fd);
			goto fail;
		    } else {
			/* as we cannot rewind the pipe, we have to assume */
			/* that the output is always the same type */

			goto postscript;
		    }
		}
		/* when it did not qualify as a shell script, fall through */

	    default:
othertype:
		if (PsText != NULL && GhostScript != NULL &&
		    buf[0] > 0 && buf[1] > 0 && buf[2] > 0 && buf[3] > 0)
		{
		    char *argv[2];

		    /* it probably is a text file, setup for conversion to PS */

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\tText file: %s\r\n",filename);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    lseek(fd,0L,SEEK_SET);	/* rewind the file */
		    num = 0;
		    memset(argv,0,sizeof(argv)); /* build argvector for PSTEXT */

		    if (push(&cpid,&fd,PsText,argv) != SUCCESS) {
			sprintf(MsgBuf,"Cannot start %s",PsText);
			logmesg(MsgBuf);
			close(fd);
			goto fail;
		    } else
			goto postscript;	/* output will be PS */
		} else {
		    sprintf(MsgBuf,"Unrecognized file: %s",filename);
		    logmesg(MsgBuf);

		    if (logfd >= 0) {
			sprintf(ChatBuf,"\t%s\r\n",MsgBuf);
			write(logfd,ChatBuf,strlen(ChatBuf));
		    }

		    close(fd);
		    goto fail;
		}
		break;
	    }

	    retry = 3;

	    while (FaxInfo.hng < 0)
	    {
		if (!gpid && !cpid)
		    pagepos = lseek(fd,0L,SEEK_CUR) - num;

		if (!skip)
		{
		    if (num < 64)		/* see if we need to read */
		    {
			memcpy(buf,buf + ptr,num); /* yes, keep data & fill buf */
			ptr = 0;

			if ((num += read(fd,buf + num,sizeof(buf) - num)) < 64)
			    break;		/* nothing left to read */
		    }

		    /* see if it has a DigiFAX header (Ghostscript) */

		    if (!strcmp(buf + ptr + 1,"PC Research, Inc")) {
			vr = buf[ptr + 29];	/* yes, get vert. resolution */
			ptr += 64;		/* skip header */
			num -= 64;
		    }
		}

		/* check if vertical resolution and data format match */
		/* negotiated values, re-negotiate when required */

		if (vr != FaxInfo.vr || df != FaxInfo.df) {
		    sprintf(ChatBuf,"AT+FDIS=%d,,,,%d\r",vr,df);
		    send(ChatBuf);

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\t%s\n",ChatBuf);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"OK")) {
			close(fd);
			goto fail;
		    }
		}

		/* check if we have not been interrupted yet */

		if (Signal) {
		    close(fd);
		    goto fail;
		}

		/* tell it we have a FAX page */

		send("AT+FDT\r");

		if (logfd >= 0)
		    write(logfd,"\tAT+FDT\r\n",9);

		if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"CONNECT") || Signal) {
		    close(fd);
		    goto fail;
		}

		/* re-check transfer parameters (result of negotiation) */

		if (vr != FaxInfo.vr || df != FaxInfo.df) {
		    sprintf(MsgBuf,"Parameter mismatch: vr=%d/%d df=%d/%d",
			    vr,FaxInfo.vr,df,FaxInfo.df);
		    logmesg(MsgBuf);

		    if (logfd >= 0) {
			sprintf(MsgBuf,"\tvr=%d/%d df=%d/%d\r\n",
				    vr,FaxInfo.vr,df,FaxInfo.df);
			write(logfd,MsgBuf,strlen(MsgBuf));
		    }

		    send(AudioEnd);
		    fax2rsp(logfd);
		    close(fd);
		    goto fail;
		}

		/* everything OK, start sending the FAX page */

		debugtime(D_FAX);
		debug2(D_FAX, "faxsend() sending FAX data\n");

		if (logfd >= 0)
		    write(logfd,"\t[page data]\r\n",14);

		if (Fw_features & FW_RTSCTSFLOW)
		    settermios(IMMEDIATE);	/* >= 6.13: hw flow ctl */
		else
		    settermios(XONXOFF);	/* XON/XOFF flow ctl */

		rmatch = 0;

		for (;;) {
		    /* refill the buffer when required */

		    if (num <= 0) {
			num = read(fd,buf,sizeof(buf));

			if (num <= 0 || Signal)
			    break;

			ptr = 0;
		    }

		    /* skip the DigiFAX header we faked when running GS */

		    if (skip) {
			if (num < skip) {
			    skip -= num;
			    num = 0;
			} else {
			    ptr += skip;
			    num -= skip;
			    skip = 0;
			}

			continue;
		    }

		    /* send this block of FAX data, doubling DLE as needed */

		    p = q = buf + ptr;

		    for (;;)
		    {
			while (--num >= 0) {
			    if (*p == fax_rtc[rmatch]) /* match with RTC */
				rmatch++;
			    else
				if (*p == fax_rtc[rmatch = 0])
				    rmatch++;

			    *p = rev_bits[(unsigned char)*p]; /* swap the bits in the file data */

			    if (*p++ == DLE ||	/* find DLE as we have to double it */
				rmatch == RTC_LEN) /* also quit when RTC seen */
				break;
			}

			if (write(STDOUT,q,p - q) <= 0 || Signal) /* write data */
			    break;

			if (rmatch == RTC_LEN)	/* did we see RTC? */
			    goto quit;

			if (num < 0)		/* all sent? */
			    break;

			q = p - 1;		/* next time around, DLE again! */
		    }

		    switch (c = getch())	/* something returned? */
		    {
		    case EOF:			/* nothing */
		    case XON:			/* leaked flow control char */
		    case XOFF:
			break;			/* ignore it and go on */

		    default:			/* something else */
			debug3(D_RUN, "faxsend() terminated by unexpected char %02x\n", c);
			InPtr--;		/* push it back */
			ioctl(STDOUT,TCFLSH,TCOFLUSH);
			ioctl(STDOUT,TCXONC,TCOON);
			waitfor(NULL,SHORTTIME); /* and flush the input */
			goto quit;		/* stop sending */
		    }
		}
quit:
		debugtime(D_FAX);
		debug2(D_FAX, "faxsend() FAX data sent\n");

		if (logfd >= 0)
		    write(logfd,"\t[DLE-ETX]\r\n",12);

		/* end of page, send DLE-ETX */

		send(AudioEnd);

		/* the time until response varies widely... */

		if (waitfor("OK",LONGTIME) == FAIL &&
		    waitfor("OK",LONGTIME) == FAIL &&
		    waitfor("OK",LONGTIME) == FAIL &&
		    waitfor("OK",LONGTIME) == FAIL) {
		    close(fd);
		    goto fail;
		}

		if (logfd >= 0)
		    write(logfd,"\tOK\r\n",5);

		settermios(INITIAL);		/* enable RTS/CTS handshake */

		/* now we have to tell it if more follows.  inconvenient. */

		if (num < 64)			/* see if we need to read */
		{
		    int n;

		    memcpy(buf,buf + ptr,num);	/* yes, keep data & fill buf */
		    ptr = 0;
		    n = read(fd,buf + num,sizeof(buf) - num);
		    num += n;
		}

		if (num >= 64 || (nfiles > 0 && file != nfiles))
		    fet = 0;			/* more pages follow */
		else
		    fet = 2;			/* end of document */

		sprintf(ChatBuf,"AT+FET=%d\r",fet);
		send(ChatBuf);

		if (logfd >= 0) {
		    sprintf(MsgBuf,"\t%s\n",ChatBuf);
		    write(logfd,MsgBuf,strlen(MsgBuf));
		}

		/* get the page transfer status */

		if (fax2rsp(logfd) != SUCCESS || strcmp(MsgBuf,"OK") || Signal) {
		    close(fd);
		    goto fail;
		}

		/* when page bad, retry it.  but not too often. */

		if (!gpid && !cpid && (FaxInfo.pts == 2 || FaxInfo.pts == 4) && retry--) {
		    lseek(fd,pagepos,SEEK_SET);
		    num = 0;
		}
	    }

	    close(fd);
	    alarm(10);
	    if (cpid)
		waitpid(cpid,&fd,0);
	    if (bpid)
		waitpid(bpid,&fd,0);
	    if (gpid)
		waitpid(gpid,&fd,0);
	    cpid = bpid = gpid = 0;
	    alarm(0);
	} while (FaxInfo.hng < 0 && nfiles > 0 && file != nfiles);

	if (FaxInfo.hng == 0)
	    rv = SUCCESS;

fail:
	send("ATH\r");
	waitfor(NULL,LONGTIME);
	CurLS = Fclass = 0;

	alarm(10);
	if (cpid)
	    waitpid(cpid,&fd,0);
	if (bpid)
	    waitpid(bpid,&fd,0);
	if (gpid)
	    waitpid(gpid,&fd,0);
	alarm(0);

	signal(SIGINT, oldint);
	signal(SIGQUIT, oldquit);
	signal(SIGTERM, oldterm);
	signal(SIGALRM, oldalarm);
	Signal = 0;

	/* free the storage associated with file list */

	if (nfiles > 0) {
	    for (file = 0; file < nfiles; file++)
		free(namelist[file]);

	    free(namelist);
	}

	sprintf(MsgBuf,"FAX send \"%s\" %s",
		name,(rv == SUCCESS? "okay":"failed"));
	if (FaxInfo.hng > 0)
	    sprintf(MsgBuf + strlen(MsgBuf),", code: %d",FaxInfo.hng);
	logmesg(MsgBuf);

	return rv;
}


/*
**	fax2rsp() - parse CLASS 2 FAX messages and save interesting data
*/

int
fax2rsp(logfd)
int logfd;

{
	char *rsp;

	FaxInfo.dtc = 0;			/* assume no DTC */

	while (waitfor(NULL,LONGTIME) == SUCCESS ||
	       waitfor(NULL,LONGTIME) == SUCCESS ||
	       waitfor(NULL,LONGTIME) == SUCCESS)
	{
	    /* save the reply, remove trailing CR/LF garbage and spaces */

	    rip(Match);
	    strcpy(MsgBuf,Match);
	    debug3(D_FAX, "fax2rsp: \"%s\"\n", MsgBuf);

	    if (logfd >= 0) {
		sprintf(ChatBuf,"\t%s\r\n",MsgBuf);
		write(logfd,ChatBuf,strlen(ChatBuf));
	    }

	    if (strncmp(MsgBuf,"+F",2))		/* no more CLASS 2 messages? */
		return SUCCESS;

	    rsp = MsgBuf + 2;			/* point to response type */

	    if (!strncmp(rsp,"CFR",3)) {	/* confirmation to receive */
		FaxInfo.cfr = 1;
		continue;
	    }

	    if (!strncmp(rsp,"CSI:",4) ||	/* called station ID */
		!strncmp(rsp,"TSI:",4) ||	/* transmit station ID */
		!strncmp(rsp,"CIG:",4))		/* polling station ID */
	    {
		rsp += 4;
		while (*rsp == ' ')		/* skip leading spaces */
		    rsp++;

		strncpy(FaxInfo.id,rsp,sizeof(FaxInfo.id));
		continue;
	    }

	    if (!strncmp(rsp,"DIS:",4) ||	/* remote capability */
		!strncmp(rsp,"DCS:",4) ||	/* current session capability */
		!strncmp(rsp,"DTC:",4))		/* request for polling */
	    {
		if (rsp[1] == 'T')
		    FaxInfo.dtc = 1;		/* request for polling rcvd */

		sscanf(rsp + 4,"%d,%d,%d,%d,%d,%d,%d,%d",
			&FaxInfo.vr,&FaxInfo.br,&FaxInfo.wd,&FaxInfo.ln,
			&FaxInfo.df,&FaxInfo.ec,&FaxInfo.bf,&FaxInfo.st);
		continue;
	    }

	    if (!strncmp(rsp,"ET:",3))	{	/* post page message response */
		FaxInfo.et = atoi(rsp + 3);
		continue;
	    }

	    if (!strncmp(rsp,"HNG:",4)) {	/* hangup */
		FaxInfo.hng = atoi(rsp + 4);
		continue;
	    }

	    if (!strncmp(rsp,"NSF:",4) ||	/* nonstandard facilities */
		!strncmp(rsp,"NSS:",4) ||	/* nonstandard setup */
		!strncmp(rsp,"NSC:",4))		/* nonstandard commands */
	    {
		int n;

		FaxInfo.nstyp = rsp[2];		/* F, S or C */
		FaxInfo.nslen = 0;
		rsp += 4;

		/* get hex digits and save into ns array */

		while (*rsp != '\0' && FaxInfo.nslen < sizeof(FaxInfo.ns)) {
		    if (sscanf(rsp,"%x",&n) != 1)
			break;

		    FaxInfo.ns[FaxInfo.nslen++] = n;
		    while (isxdigit(*rsp))	/* skip number */
			rsp++;
		    while (*rsp == ' ')		/* skip space */
			rsp++;
		}

		continue;
	    }

	    if (!strncmp(rsp,"POLL",4)) {	/* has document to poll */
		FaxInfo.poll = 1;
		continue;
	    }

	    if (!strncmp(rsp,"PTS:",4)) {	/* page transfer status */
		sscanf(rsp + 4,"%d,%d,%d,%d,%d",
			&FaxInfo.pts,&FaxInfo.lc,&FaxInfo.blc,&FaxInfo.cblc,
			&FaxInfo.lbc);
		continue;
	    }

	    logmesg(MsgBuf);			/* nonmatched response, log it */
	}

	debug2(D_FAX, "fax2rsp: timeout\n");
	return FAIL;				/* timeout */
}

/*
**	faxident() - return FAX identification, if any
**
**	for now, just dumps the NSF string as printable chars, as I
**	have no description of the contents
*/

char *
faxident ()
{
	char *p;
	int i,c;

	if (FaxInfo.nstyp != 'F' || FaxInfo.nslen == 0)
	    return "";

	strcpy(ChatBuf,"NSF=");
	p = ChatBuf + 4;

	for (i = 0; i < FaxInfo.nslen; i++) {
	    c = FaxInfo.ns[i];

	    if (isascii(c) && isprint(c))
		*p++ = c;
	    else
		*p++ = '.';
	}

	*p = '\0';
	return ChatBuf;
}
#endif

/*************************************************************************
**	VOICE mode handling						**
*************************************************************************/

/*
**	voicemode() - main handling of VOICE calls
*/

int
voicemode(answer,message)
int answer;
char *message;

{
	int result = FAIL;
	int rectime = 0;
	int rv;
	const char *action;

	debug2(D_RUN, "VOICE mode\n");

	sprintf(ChatBuf,"AT%s%s+VNH=1%s\r",	/* select VOICE mode */
		LSOn,make_fclass(8),make_vls(PhoneLS));
	send(ChatBuf);				/* (no hangup, no mute) */
	if (wait_vcon(MEDMTIME) == FAIL)	/* sometimes OK, sometimes VCON */
	    return FAIL;

	CurLS = PhoneLS;			/* remember current LS */

	if (answer != ONLINE) {
	    send("ATA\r");			/* answer the call */
	    rv = wait_vcon(LONGTIME);		/* sometimes OK, sometimes VCON */
	    if (rv == FAIL || rv == NOCARRIER)
		return FAIL;
	}

	sound_open();				/* access sound card */

	UserLevel = 0;				/* initial user level */
	Number[0] = '\0';			/* no user-entered number yet */

	/* send optional answertone */
	/* (could try to use this to trick calling modems) */

	if (answer != ONLINE)
	    beep(AnswerTone);

	/* do a first check of the line (when configured) */
	/* this allows quick, silent detection of FAX and DATA */

	if (CheckLine > 0)
	    if ((rectime = dtmfrecord("0123456789ABCD*#abcdel",CheckLine)) > 0 &&
		MsgBuf[0] != '\0')
	    {
		/* check for 'b' (BUSY) or 'd' (DIALTONE, only >= 6.11 firmware) */
		/* also 'l' (remote hangup) (2864) */

		if (strpbrk(MsgBuf,"bdl") != NULL)
		{
		    sprintf(ChatBuf,"VOICE call, hangup (%s)",MsgBuf);
		    logmesg(ChatBuf);
		    goto done;
		}

		/* check for 'c' (FAX CNG) */

		if (strchr(MsgBuf,'c') != NULL)
		{
		    logmesg("VOICE call, T.30 FAX tone detected");
		    result = FAX;
		    goto done;
		}

		/* check for 'e' (DATA CNG, only >= 6.11 firmware) */

		if (strchr(MsgBuf,'e') != NULL)
		{
		    logmesg("VOICE call, DATA tone detected");
		    result = DATA;
		    goto done;
		}

		/* handle other input */

		sprintf(ChatBuf,"VOICE call, DTMF key (%s)",MsgBuf);
		logmesg(ChatBuf);

		if (isdigit(MsgBuf[0]))		/* number keyed */
		    result = voicekey(VoiceKey,MsgBuf[0]);
		else
		    result = voicekey(VoiceKey,0); /* other key, ask for more */

		RecLS = 0;			/* reset current LS */

		if (result == SUCCESS)
		    play("goodbye","#bdl");

		goto done;
	    }

	/* play the greeting */

	if (message != NULL)
	{
	    if (play(message,"*#abcdel") == FAIL)
	    {
		sprintf(ChatBuf,"VOICE call, error during play (%s)",MsgBuf);
		logmesg(ChatBuf);
		goto done;
	    }

	    /* check for 'b' (BUSY) or 'd' (DIALTONE, only >= 6.11 firmware) */
	    /* also 'l' (remote hangup) (2864) */

	    if (strpbrk(MsgBuf,"bdl") != NULL)
	    {
		sprintf(ChatBuf,"VOICE call, hangup during play (%s)",MsgBuf);
		logmesg(ChatBuf);
		goto done;
	    }

	    /* check for 'c' (FAX CNG) */

	    if (strchr(MsgBuf,'c') != NULL)
	    {
		logmesg("VOICE call, T.30 FAX tone detected during play");
		result = FAX;
		goto done;
	    }

	    /* check for 'e' (DATA CNG, only >= 6.11 firmware) */

	    if (strchr(MsgBuf,'e') != NULL)
	    {
		logmesg("VOICE call, DATA tone detected during play");
		result = DATA;
		goto done;
	    }
	}

	/* record incoming message */

	strcpy(Recordfile,RecFile);

	switch (rectime = record(Recordfile,TRUE,"0123456789ABCD*#abcdelq",Silence,RecMode,120))
	{
	case 0:
	    logmesg("VOICE call, silence");
	    result = SILENCE;		/* try DATA mode */
	    break;

	case FAIL:
	    logmesg("VOICE call, error in record");
#ifdef DEBUG
	    linestat();
#endif
	    result = FAIL;
	    break;

	default:
	    /* message too short to keep? */

	    if (rectime <= DropMsg && unlink(Recordfile) == 0)
		action = ", erased";
	    else
		action = " recorded";

	    /* check for FAX/DATA tones */

	    if (strchr(MsgBuf,'c') != NULL)
	    {
		logmesg("VOICE call, T.30 FAX tone detected during record");
		result = FAX;
		break;
	    }

	    if (strchr(MsgBuf,'e') != NULL)
	    {
		logmesg("VOICE call, DATA tone detected during record");
		result = DATA;
		break;
	    }

	    /* not FAX or DATA, assume a human caller */

	    sprintf(ChatBuf,"VOICE call, %dsec message%s (%s)",rectime,action,MsgBuf);
	    logmesg(ChatBuf);

	    result = SUCCESS;			/* be optimistic */

	    if (strpbrk(MsgBuf,"bdl") == NULL) {
		if (strchr(MsgBuf,'*') != NULL)
		    result = voicekey(VoiceKey,0); /* '*' keyed */
		else
		    if (isdigit(MsgBuf[0]))	/* number keyed */
			result = voicekey(VoiceKey,MsgBuf[0]);

		RecLS = 0;			/* reset current LS */

		if (result == SUCCESS)
		    play("goodbye","#bdl");
	    }
	    break;
	}

done:
	/* when terminating call, reset LS and hangup
	 * when switching to other mode, keep the line active
	 */
	if (result == SUCCESS || result == FAIL) {
	    if (set_vls(0))
		result = FAIL;

	    send("ATH\r");
	    CurLS = Fclass = 0;
	    if (waitfor("OK",MEDMTIME))
		result = FAIL;
	}

	/* check if we have a recorded message (i.e. it was not too
	 * short and wasn't deleted by the caller) and perform optional
	 * action if so.  (entry "20" in the key file)
	 *
	 * the NUMBER variable is set to *datetime*length so it can
	 * be passed to a PAGER (when that supports this data format...)
	 * new: the message in NUMBER can be user-configured using the
	 * PAGERRECMSG value, which gets 4 arguments:
	 * - the date and time in MMDDHHMM format as a string
	 * - the length of the message as an integer
	 * - the Caller-ID as a string, or empty string when not available
	 * - the Caller-NAME as a string, or empty string when not available.
	 */

	if (answer != ONLINE && rectime > 0 && !access(Recordfile,R_OK)) {
	    strftime(MsgBuf,sizeof(MsgBuf),"%m%d%H%M",localtime(&RecTime));
	    sprintf(Number,PagerRecMsg,MsgBuf,rectime,
	    		(CndNumber != NULL? CndNumber : ""),
	    		(CndName != NULL? CndName : ""));
	    debug3(D_RUN, "Number: %s\n", Number);
	    voicecommand(VoiceKey,20,dtmfdigit,dtmfstring);
	}

	sound_close();

	debug3(D_RUN, "VOICE mode returns %d\n",result);
	return result;
}

/*
**	messagemode() - play an announcement and hangup
*/

int
messagemode(answer,message)
int answer;
char *message;

{
	int result = FAIL;
	int rv;

	debug3(D_RUN, "MESSAGE mode (%s)\n",message);

	sprintf(ChatBuf,"AT%s%s+VNH=1%s\r",	/* select VOICE mode */
		LSOn,make_fclass(8),make_vls(PhoneLS));
	send(ChatBuf);				/* (no hangup, no mute) */
	if (wait_vcon(MEDMTIME) == FAIL)	/* sometimes OK, sometimes VCON */
	    return FAIL;

	CurLS = PhoneLS;			/* remember current LS */

	if (answer != ONLINE) {
	    send("ATA\r");			/* answer the call */
	    rv = wait_vcon(LONGTIME);		/* sometimes OK, sometimes VCON */
	    if (rv == FAIL || rv == NOCARRIER)
		return FAIL;
	}

	sound_open();				/* access sound card */

	/* play the message */

	if (message != NULL)
	{
	    if (play(message,"*#abdl") == FAIL)
	    {
		sprintf(ChatBuf,"MESSAGE call, error during play (%s)",MsgBuf);
		logmesg(ChatBuf);
		goto done;
	    }

	    /* check for 'b' (BUSY) or 'd' (DIALTONE, only >= 6.11 firmware) */
	    /* also 'l' (remote hangup) (2864) */

	    if (strpbrk(MsgBuf,"bdl") != NULL)
	    {
		sprintf(ChatBuf,"MESSAGE call, hangup during play (%s)",MsgBuf);
		logmesg(ChatBuf);
		goto done;
	    }

	    /* check for 'c' (FAX CNG) */

	    if (strchr(MsgBuf,'c') != NULL)
	    {
		logmesg("MESSAGE call, T.30 FAX tone detected during play");
		goto done;
	    }

	    /* check for 'e' (DATA CNG, only >= 6.11 firmware) */

	    if (strchr(MsgBuf,'e') != NULL)
	    {
		logmesg("MESSAGE call, DATA tone detected during play");
		goto done;
	    }

	    result = SUCCESS;
	}

done:
	/* reset LS and hangup
	 */
	if (set_vls(0))
	    result = FAIL;

	send("ATH\r");
	CurLS = Fclass = 0;
	if (waitfor("OK",MEDMTIME))
	    result = FAIL;

	sound_close();

	debug3(D_RUN, "MESSAGE mode returns %d\n",result);
	return result;
}

/*
**	localvoice() - handling of local phone in voice mode (2864)
*/

int
localvoice()

{
	int result;
	int rv;

	if (LocalKey == NULL)
	    return NOLOCAL;		/* no handling defined, just return */

	debug2(D_RUN, "local phone VOICE mode\n");

	sprintf(ChatBuf,"AT%s%s%s\r",	/* select VOICE mode */
		Mute,make_fclass(8),make_vls(1));
	send(ChatBuf);			/* (no hangup, no mute) */
	if (wait_vcon(MEDMTIME) == FAIL) /* sometimes OK, sometimes VCON */
	    return FAIL;

	CurLS = 1;			/* remember current LS */

	send("ATA\r");
	rv = wait_vcon(LONGTIME);		/* sometimes OK, sometimes VCON */
	if (rv == FAIL || rv == NOCARRIER)
	    return FAIL;

	UserLevel = 0;			/* initial user level */
	Number[0] = '\0';		/* no user-entered number yet */

	result = voicekey(LocalKey,0);	/* handle DTMF presses */

	switch (result)
	{
	case SUCCESS:			/* 'exit' command, connect line */
	    if (set_vls(0))
		result = FAIL;

	    break;			/* modem will re-init */

	default:
	    result = FAIL;		/* program exits and restarts */
	    break;
	}

	return result;
}

/*
**	voicekey() - handling of * or number entry in VOICE mode
*/

int
voicekey(filename,firstdigit)
char *filename;				/* command filename */
int firstdigit;				/* first digit to be handled */

{
	int code;
	int result;

	debug4(D_RUN, "voicekey(\"%s\",%d)\n",filename,firstdigit);

	for (;;)
	{
	    /* get a digit */

	    code = firstdigit? firstdigit : dtmfdigit();

	    /* check for numerics, '*' and '#' */

	    if (isdigit(code))
		code -= '0';
	    else
		switch (code)
		{
		case '*':
		    code = 11;
		    break;

		case '#':
		    code = 12;
		    break;

		case 'A':
		    code = 13;
		    break;

		case 'B':
		    code = 14;
		    break;

		case 'C':
		    code = 15;
		    break;

		case 'D':
		    code = 16;
		    break;

		default:			/* specials: 'b', 'd', 's', 'q' */
		    return FAIL;		/* and FAIL code */
		}

	    /* handle the key, using the passed commands file */

	    switch (result = voicecommand(filename,code,dtmfdigit,dtmfstring))
	    {
	    case SUCCESS:			/* went OK */
	    	break;				/* next key */

	    case EXIT:				/* exit from voice mode */
		return SUCCESS;

	    case COMMAND:			/* go to expert mode */
		do
		{
		    MsgBuf[0] = '\0';

		    switch (code = dtmfcode())	/* ask for *nn# code */
		    {
		    case FAIL:
			return FAIL;

		    default:
			switch (result = voicecommand(VoiceCmd,code,dtmfdigit,dtmfstring))
			{
			case FAIL:
			case DATAFAX:
			case DATA:
			case FAX:
			case DIALBACK:
			    return result;

			case UNKNOWN:
			    beep(Error);
			    break;
			}
			break;
		    }
		}
		while (result != EXIT);
		break;

	    case UNKNOWN:			/* unknown key */
	    case ERROR:				/* some problem */
		beep(Error);
		break;

	    default:				/* mode changes, etc */
		return result;
	    }

	    firstdigit = 0;
	}
}

/*
**	voicecommand() - handling of VOICE expert commands
*/

int
voicecommand(filename,code,inputdigit,inputstring)
char *filename;
int code;
int (*inputdigit)(void);
char *(*inputstring)(void);

{
	FILE *fp;
	char *p,*q;
	int result;
	char buf[MAXLINE];

	if (filename == NULL || filename[0] == '\0')
	    return FAIL;

	/* open the commands file, which contains lines like:
	 * # comment
	 * 10: actions for code 10
	 *     more actions for code 10
	 * 20: actions for code 20
	 * ... etc
	 */
	if ((fp = fopen(filename,"r")) == NULL) {
	    debug3(D_RUN, "Open commandfile (%s) fails\n", filename);
	    return UNKNOWN;
	}

	/* find the proper starting point and parse commands */

	while (fgets(buf,sizeof(buf),fp) != NULL)
	{
	    if (buf[0] == '#' || (p = strchr(buf,':')) == NULL)
		continue;

	    *p++ = '\0';

	    if (isdigit(buf[0]) && atoi(buf) == code)
	    {
		/* found the code, now run commands until next label */

		for (;;)
		{
		    while (isspace(*p))		/* skip leading blanks */
			p++;

		    if ((q = strchr(p,'\n')) != NULL) /* remove \n */
			*q = '\0';

		    if (*p != '\0' &&
			(result = docommand(p,inputdigit,inputstring)) != SUCCESS)
		    {
			fclose(fp);
			return result;
		    }

		    do {
			if (fgets(buf,sizeof(buf),fp) == NULL || isdigit(buf[0])) {
			    fclose(fp);
			    return SUCCESS;
			}
		    } while (buf[0] == '#');

		    p = buf;
		}
	    }
	}

	fclose(fp);
	return UNKNOWN;
}


/*
**	docommand() - run a single command line from VOICECMD file
*/

int
docommand(line,inputdigit,inputstring)
char *line;
int (*inputdigit)(void);
char *(*inputstring)(void);

{
	char *p,*argv[NARG];
	int argc,quote;

	debug3(D_RUN, "docommand(\"%s\")\n", line);

	for(argc = 0;argc < NARG;argc++)
		argv[argc] = NULL;

	for(argc = 0;argc < NARG && *line != '\0';){
	    quote = 0;
	    /* Skip leading white space */
	    while(isspace(*line))
		line++;
	    if(*line == '\0')
		break;
	    /* Check for quoted token */
	    if(*line == '"' || *line == '\'')
		quote = *line++;	/* Suppress quote, remember it */
	    argv[argc++] = line;	/* Beginning of token */
	    /* Find terminating delimiter */
	    if(quote){
		/* Find quote, it must be present */
		if((line = strchr(line,quote)) == NULL)
		    return FAIL;

		*line++ = '\0';
	    } else {
		/* Find space or tab. If not present,
		 * then we've already found the last
		 * token.
		 */
		for (p = line; *p; p++)
		    if (isspace(*p))
			break;
		if (*p != '\0')
		    *p++ = '\0';
		line = p;
	    }
	}

	if (argc == 0)
	    return UNKNOWN;			/* nothing, shouldn't happen */

	/* execute command.  kind of an ad-hoc approach, this is */

	/* auplay <filename>
	 * plays a .au audio file (also known as .snd file)
	 */
	if (argc > 1 && !strcmp(argv[0],"auplay"))
	    return auplay(argv[1]);

	/* beep <VTS-string>
	 * sends a +VTS= string to generate audible beep
	 * default is the ERROR beep defined in config file
	 */
	if (!strcmp(argv[0],"beep"))
	    return beep(argc > 1? argv[1] : Error);

	/* checklevel <level>
	 * verify current user has specified level (as a bit field)
	 */
	if (argc > 1 && !strcmp(argv[0],"checklevel")) {
	    if (!(UserLevel & atoi(argv[1])))
		return ERROR;

	    return SUCCESS;
	}

	/* checkfile [not] <filename>
	 * verify file existance
	 */
	if (argc > 1 && !strcmp(argv[0],"checkfile")) {
	    char **argp;
	    int not;

	    argp = &argv[0];

	    /* optional 'not' inverts test */

	    if ((not = !strcmp(argv[1],"not"))) {
		argp++;
		if (--argc < 2)
		    return ERROR;
	    }

	    if (not ^ !access(argp[1],R_OK))
	    {
		debug2(D_RUN, "checkfile: success\n");
		return SUCCESS;
	    }

	    debug2(D_RUN, "checkfile: fail\n");
	    return ERROR;
	}

	/* checktime [not] <from_hhmm> <until_hhmm> [<from_weekday> <until_weekday>]
	 * verify current time is between the specified bounds
	 */
	if (argc > 2 && !strcmp(argv[0],"checktime")) {
	    time_t now;
	    struct tm *tm;
	    char **argp;
	    int hhmm,not;

	    now = time(NULL);
	    tm = localtime(&now);
	    hhmm = 100 * tm->tm_hour + tm->tm_min;
	    debug4(D_RUN, "checktime: now %04d weekday %d\n", hhmm, tm->tm_wday);

	    argp = &argv[0];

	    /* optional 'not' inverts test */

	    if ((not = !strcmp(argv[1],"not"))) {
		argp++;
		if (--argc < 3)
		    return ERROR;
	    }

	    if (not ^
		((hhmm >= atoi(argp[1]) && hhmm <= atoi(argp[2])) &&
		 (argc <= 4 ||
		  (tm->tm_wday >= atoi(argp[3]) && tm->tm_wday <= atoi(argp[4]))
		 )
		)
	       )
	    {
		debug2(D_RUN, "checktime: success\n");
		return SUCCESS;
	    }

	    debug2(D_RUN, "checktime: fail\n");
	    return ERROR;
	}

	/* command
	 * returns code COMMAND to upper level, to go to command mode
	 */
	if (!strcmp(argv[0],"command"))
	    return COMMAND;

	/* dial [<number>]
	 * sends hookflash and calls number in Number (or specified number)
	 */
	if (!strcmp(argv[0],"dial")) {
	    char *number;
	    char *dialprefix;
	    int rv;

	    if (argc > 1 && argv[1][0] != 'N')
		number = argv[1];
	    else
		number = Number;

	    if (number == NULL || !number[0])
		return ERROR;

	    dialprefix = DialPrefix;	/* save DialPrefix and clear it */
	    DialPrefix = NULL;		/* to allow dialup to internal # */

	    rv = dialup("H0H1",8,5,number); /* flash & dial in voice mode */

	    DialPrefix = dialprefix;	/* restore the DialPrefix */
	    return rv;
	}

	/* dialback [<number>]
	 * calls number in Number (or specified number) and enters voice mode
	 */
	if (!strcmp(argv[0],"dialback")) {
	    char *number;

	    if (argc > 1 && argv[1][0] != 'N')
		number = argv[1];
	    else
		number = Number;

	    if (!number[0])
		return ERROR;

	    if (Synth != NULL)
		say("now dialling back");
	    else
		play("gong","#bdl");

	    /* hangup existing call and init the MODEM
	     */
	    if (initmodem(TRUE) != SUCCESS)
		return FAIL;

	    sleep(15);

	    if (dialup("",8,5,number) == SUCCESS) /* dial in voice mode */
		return DIALBACK;

	    return FAIL;
	}

	/* exit
	 * returns code EXIT to upper level, to exit from voice mailbox
	 */
	if (!strcmp(argv[0],"exit"))
	    return EXIT;

	/* fail
	 * returns code FAIL to upper level, to hangup and re-init
	 */
	if (!strcmp(argv[0],"fail"))
	    return FAIL;

	/* fax <filename> [<number>]
	 * send a FAX file.  when <filename> is N, use FaxPoll file with
	 * Number entered using "readnumber" (4 digits usually).
	 * when <number> specified, dial it.  when <number> is N, use the
	 * Number entered using "readnumber".
	 * because this function needs the line for outgoing dial and/or
	 * will hangup the existing connection, it will return EXIT so that
	 * the session is ended...
	 */

	if (!strcmp(argv[0],"fax")) {
	    char faxfile[MAXLINE];

	    if (argc == 2 && argv[1][0] == 'N') {
		if (FaxPoll == NULL)
		    return ERROR;

		sprintf(ChatBuf,"%04u",atoi(Number));
		sprintf(faxfile,FaxPoll,ChatBuf);

		if (!access(faxfile,R_OK))
		    argv[1] = faxfile;
		else
		    return ERROR;
	    }

	    if (argc >= 3 && argv[2][0] == 'N')
		argv[2] = Number;

	    if (Synth != NULL)
		say("now calling your FAX");
	    else
		play("gong","#bdl");

	    sprintf(MsgBuf,"send FAX %s to %s",argv[1],
			(argc == 2 || argv[2][0] == '\0')? "caller" : argv[2]);
	    logmesg(MsgBuf);

	    if (argc == 2 || argv[2][0] == '\0') {
		/* select FAX mode on this same connection */
#ifdef ZFAX
		return ERROR;			/* cannot do that with ZFAX */
#else
		int n;

		sprintf(ChatBuf,"AT%s%s+FAA=0D\r",
			(Is_ZyXEL? "S38.4=0":""),make_fclass(2));
		send(ChatBuf);

		for (n = 5; n > 0; n--)
		    if (!waitfor(NULL,LONGTIME))
		    {
			rip(Match);

			if (!strcmp(Match,"+FCON"))
			    break;

			/* got something else than expected */

			logmesg(Match);		/* log what we got (NO CARRIER?) */
			debug3(D_RUN, "faxback - connect failed (%s)\n", Match);
			return FAIL;
		    }

		if (!n)				/* nothing received */
		    return FAIL;
#endif
		argv[2] = NULL;
	    } else {
		/* hangup existing call and init the MODEM
		 */
		if (initmodem(TRUE) != SUCCESS)
		    return FAIL;

		sleep(5);
	    }

	    if (faxsend(argv[2],argv[1],-1) == SUCCESS)
		return EXIT;

	    return FAIL;
	}

	/* logmesg "string"
	 * logs the specified message in the logfile
	 */
	if (argc > 1 && !strcmp(argv[0],"logmesg")) {
	    logmesg(argv[1]);
	    return SUCCESS;
	}

	/* mode <mode>
	 * returns specified <mode> to next upper level, which is
	 * supposed to switch the call to that mode.
	 */
	if (argc > 1 && !strcmp(argv[0],"mode")) {
	    if (!strcmp(argv[1],"DATAFAX"))
		return DATAFAX;

	    if (!strcmp(argv[1],"DATA"))
		return DATA;

	    if (!strcmp(argv[1],"FAX"))
		return FAX;

	    if (!strcmp(argv[1],"VOICE"))
		return VOICE;

	    if (!strcmp(argv[1],"MESSAGE"))
		return MESSAGE;

	    return UNKNOWN;
	}

	/* music <filename>
	 * plays a MIDI or MOD file
	 */
	if (argc > 1 && !strcmp(argv[0],"music"))
	    return music(argv[1]);

	/* pager [<info>]
	 * calls pager, sending optional info.
	 * when <info> is N, the Number entered using "readnumber" is used.
	 * because this function needs the line for outgoing dial, it will
	 * return EXIT so that the session is ended...
	 */

	if (!strcmp(argv[0],"pager")) {
	    if (PagerType == 0 || PagerNum == NULL)
		return ERROR;

	    if (argc >= 2 && argv[1][0] == 'N')
		argv[1] = Number;

	    if (pager(argv[1]) == SUCCESS)
		return EXIT;

	    return FAIL;
	}

	/* play [<filename> [<endstring>]]
	 * plays the specified file, or the recorded file by default
	 */
	if (!strcmp(argv[0],"play")) {
	    if (argc < 2 || argv[1][0] == '\0')
		argv[1] = Recordfile;

	    if (argc < 3)
		argv[2] = "#bdl";		/* default end characters */

	    if (play(argv[1],NULL) != SUCCESS)	/* verify the file is valid */
		return ERROR;

	    if (play(argv[1],argv[2]) == SUCCESS &&
		strpbrk(MsgBuf,"bdl") != NULL)	/* busy, dialtone, hangup */
		return FAIL;

	    return SUCCESS;
	}

	/* playback [<direc>]
	 * enters message playback mode, optionally for certain directory
	 * default directory is RecDir, the incoming message directory.
	 */
	if (!strcmp(argv[0],"playback"))
	    return playback(argv[1],inputdigit);

	/* readnumber
	 * gets a (phone) number, reads it aloud, and saves in a global
	 */
	if (!strcmp(argv[0],"readnumber")) {
	    if ((p = (*inputstring)()) != NULL) {
		strcpy(Number,p);		/* save it for later use */
		p = Number;

		debug3(D_RUN, "Number: %s\n", Number);

		/* read it aloud. use voice synth when available, else */
		/* play individual digit recordings */

		if (Synth != NULL)
		{
		    char *q = ChatBuf;
		    int ndig = 0;

		    /* convert numeric string into groups of 2 digits */
		    /* except for digit 0, which always ends a group. */
		    /* e.g.: 030715610 becomes 0 30 71 56 10 */
		    /* (this maybe has to be customizable for countries */
		    /* where numbers are structured like #-###-###-#### */
		    /* and one would read them in groups like that) */

		    while (*p != '\0') {
			*q++ = *p;
			ndig++;

			if (*p == '0' || ndig == 2) {
			    *q++ = ' ';
			    ndig = 0;
			}

			p++;
		    }

		    if (q != ChatBuf && *--q != ' ')
			q++;
		    *q = '\0';

		    if (say(ChatBuf) != SUCCESS)
			return FAIL;
		} else {
		    while (*p != '\0') {
			sprintf(ChatBuf,"%c",*p);
			if (play(ChatBuf,"bd") != SUCCESS ||
			    MsgBuf[0] == 'b' || MsgBuf[0] == 'd')
			    return FAIL;

			p++;
		    }
		}
	    }
	    return SUCCESS;
	}

	/* record <filename> <endstring> <vsd-string> <vsm-value> <maxtime>
	 */
	if (!strcmp(argv[0],"record")) {
	    char recordfile[MAXLINE];
	    int enc,maxtime;

	    /* setup the defaults */
	    if (argc < 2 || argv[1][0] == '\0')
		argv[1] = Recordfile;

	    if (argc < 3 || argv[2][0] == '\0')
		argv[2] = "#bdlq";

	    if (argc < 4 || argv[3][0] == '\0')
		argv[3] = Silence;

	    if (argc < 5 || argv[4][0] == '\0')
		enc = RecMode;
	    else
		enc = atoi(argv[4]);

	    if (argc < 6)
		maxtime = 120;
	    else
		maxtime = atoi(argv[5]);

	    /* make recording with these values */
	    strcpy(recordfile,argv[1]);

	    if (record(recordfile,TRUE,argv[2],argv[3],enc,maxtime) == FAIL)
		return FAIL;

	    return SUCCESS;
	}

	/* recLS <LS-nr>
	 * sets the input device (LS) for recording
	 * mainly useful for local testing and recording of messages
	 */
	if (argc > 1 && !strcmp(argv[0],"recLS")) {
	    RecLS = atoi(argv[1]);
	    return SUCCESS;
	}

	/* remove <file>
	 * removes a file, or the current Recordfile when no file
	 * is specified.
	 */
	if (!strcmp(argv[0],"remove")) {
	    if (unlink(argc > 1? argv[1] : Recordfile) != 0)
		return ERROR;

	    return SUCCESS;
	}

	/* run [<envvar>=<value>...] <string> [<args>...]
	 * run program and send result to voice synthesizer
	 */
	if (argc > 1 && !strcmp(argv[0],"run"))
	    return run(TRUE,argv + 1);

	/* runx [<envvar>=<value>...] <string> [<args>...]
	 * run program that uses the modem (e.g. uucp)
	 */
	if (argc > 1 && !strcmp(argv[0],"runx"))
	    return run(FALSE,argv + 1);

	/* say <string>
	 * calls voice synthesizer to say <string>
	 */
	if (argc > 1 && !strcmp(argv[0],"say"))
	    return say(argv[1]);

	/* userlevel <level>
	 * sets the user's level (after entry of a password)
	 */
	if (argc > 1 && !strcmp(argv[0],"userlevel")) {
	    UserLevel = atoi(argv[1]);
	    sprintf(MsgBuf, "User level %d", UserLevel);
	    logmesg(MsgBuf);
	    return SUCCESS;
	}

	return UNKNOWN;
}


/*
**	pager () - call a pager and send optional info
**
**	because the access to the paging service could be dependent on
**	local conventions, some alteration of this code may be needed.
**	a switch() is provided to insert local varieties of the algorithm.
**	if you need a different algorithm, please add it under a different
**	case, with a value matching your country access code, and please mail
**	the diffs to me so I can include it in a later version...
**
**	currently implemented algorithms:
**
**	31	Netherlands
**		Call the specified number and wait until it answers (VOICE).
**		When numeric info is to be sent, wait for the end of the
**		voice announcement and the appearance of a DIAL tone, then
**		send the info using DTMF (even when local dialling is PULSE).
**		Info is terminated with a # sign.
**		Then, wait for the confirmation voice message and beep.
**		Actually, we cannot match that and wait for the BUSY signal
**		which follows immediately after the beep.
**		Note that the appearance of the second dialtone is dependent
**		on the kind of pager, and therefore it is your responsability
**		to configure the program (defaults file and keys/commands file)
**		in such a way that extra info is sent when the pager expects
**		it, and is not sent for a pager that does not support it.
**
**	311	Alternative for Netherlands (BUZZER service)
**		Same as 31, but does not wait for second DIAL tone.
**
**	312	Alternative for Netherlands
**		Call a certain number in DATA mode, and wait for the prompt
**		for the pager number to call ("=0?=").
**		Send the pager number followed by RETURN.  Wait for prompt
**		for numeric information ("=2?=").
**		Send numeric information, followed by RETURN.
**		Wait for confirmation ("=80=") and disconnect.
**
**	313	IXO/TAP/PET  (CallMax)
**		Call a PAGERCENTRAL number in DATA mode, and send the page
**		using the TAP protocol (also known as IXO and PET).
**
**	314	UCP (GSM-SMS)
**		Call a PAGERCENTRAL number in DATA mode, and send the page
**		using the UCP protocol.
**
**	999	Make a VOICE call and announce the message.
**
*/

int
pager(info)
const char *info;				/* (optional) info to send */
{
	int result = FAIL;
	int retry,sum,ch;
	char *p;
	char buf[MAXLINE],mode[MAXLINE];

	debug3(D_RUN, "pager(\"%s\")\n", info);

	/* hangup existing call and init the MODEM
	 */
	if (initmodem(TRUE) != SUCCESS)
	    return FAIL;

	if (!(Fw_features & FW_ISDN))
	    sleep(2);

	mode[0] = '\0';

	if (!(Fw_features & FW_LINUX))
	    strcpy(mode,"M0L0");		/* should be made config. */

	switch (PagerType)
	{
	case 31:				/* Netherlands */
	case 311:				/* Netherlands - BUZZER */
	    if (Voice)				/* needs Voice for this */
	    {
		/* dial in voice mode, proceed when no ringback */

		if (!(Fw_features & FW_LINUX))
		    strcat(mode,(Fw_features & FW_2864? "+VRN=10" : "+VRN=100"));

		if (dialup(mode,8,3,PagerNum) != SUCCESS)
		    break;

		/* when info is to be sent, wait for the second dialtone */

		if (info != NULL && info[0] != '\0') {
		    if (PagerType != 311 &&	/* BUZZER does not send dialtone */
			(Fw_features & FW_DIALTONE)) { /* 6.11 implements 'd' */
			if (dtmfrecord("bdq",25) < 5 || strchr(MsgBuf,'d') == NULL)
			    break;
		    } else
			sleep(5);

		    /* send info + # */

		    if (beep(info) || beep("#"))
			break;
		}

		/* now we have to wait until the call is accepted. we cannot
		 * match the confirmation voice message and beep, but we should
		 * at least wait until BUSY is returned so that the paging
		 * service has hung up the call.  (if we hang up too soon, the
		 * request will be canceled...)
		 */
		if (dtmfrecord("bdlq",25) < 3 ||
		    (strchr(MsgBuf,'b') == NULL && strchr(MsgBuf,'l') == NULL))
		    break;
	    } else {				/* no Voice, dumb method */
		/* dial the pager number, wait a bit, send the message */

		sprintf(buf,"%s,,,,,%s#;",PagerNum,info);

		if (dialup(mode,-2,5,buf) != SUCCESS)
		    break;

		/* it worked OK as far as can be detected, wait a while */
		/* (if we hang up too soon, the request will be canceled...) */

		sleep(5);
	    }

	    result = SUCCESS;
	    break;

	case 312:				/* Netherlands, alternative */
	    /* set 7 bit, EVEN parity (this is what the service expects/sends)
	     */
	    settermios(SEVEN_E);
	    send("AT\r");
	    if (waitfor("OK",MEDMTIME))
		break;

	    /* the number to be dialled depends on the first digit in the
	     * pager number.  what a mess...
	     */
	    sprintf(buf,"065%c5065%c5",PagerNum[0],PagerNum[0]);

	    /* dialup in data mode.
	     * it seems they only support 2400/NONE, which is good enough
	     * and provides faster training/negotiation...
	     */
	    if (Is_ZyXEL)
		strcat(mode,"&N14&K0");

	    if (dialup(mode,0,3,buf) != SUCCESS)
		break;

	    if (waitfor("*=0?=",LONGTIME))
		break;

	    sprintf(ChatBuf,"%s\r",PagerNum);	/* send number + CR */
	    send(ChatBuf);

	    if (info != NULL && info[0] != '\0') {
		if (waitfor("*=2?=",LONGTIME))
		    break;

		sprintf(ChatBuf,"%s\r",info);	/* send info + CR */
		send(ChatBuf);
	    }

	    if (waitfor("*=80=",LONGTIME))
		break;

	    /* flush the garbage it sends after the confirmation
	     */
	    waitfor("FLUSH",MEDMTIME);
	    result = SUCCESS;
	    break;

	case 313:				/* IXO/TAP/PET */
	    /* set 7 bit, EVEN parity (this is what the service expects/sends)
	     */
	    settermios(SEVEN_E);
	    send("AT\r");
	    if (waitfor("OK",MEDMTIME))
		break;

	    /* dialup in data mode */
#if 0
	    /* it seems they only support 1200/NONE, which is good enough
	     * and provides faster training/negotiation...
	     */
	    if (Is_ZyXEL)
		strcat(mode,"&N15&K0");
#endif

	    if (dialup(mode,0,3,PagerCentral) != SUCCESS)
		break;

	    /* first step:
	     * attempt to get an "ID=" response sending CR characters
	     * in 2-second intervals (3 tries)
	     */

	    retry = 3;
	    ch = 0;
	    usleep(500000);

	    while (ch != '=' && retry--) {
		InPtr = InLen = 0;
		debugtime(D_CHAT);
		send("\r");
		settermios(MEDMTIME);
		debug2(D_CHAT,"   RECEIVE: ");
		while ((ch = getch()) != EOF) {
		    debug3(D_CHAT, ((ch < ' ') ? "^%c" : "%c"),
				   ((ch < ' ') ? ch | 0100 : ch));
		    if (ch == '=')
			break;
		}
		debug2(D_CHAT,"\n");
		InPtr = InLen = 0;
	    }

	    if (retry < 0)
		break;

	    /* second step:
	     * send the ID ("PG1") and await an ACK
	     */

	    retry = 3;
	    debugtime(D_CHAT);

	    do {
		send("\033PG1\r");
		if (waitfor(NULL,-LONGTIME))
		    break;
	    } while (Match[0] == NAK && retry--);

	    retry = 3;

	    while (Match[0] != ACK && retry--)
		if (waitfor(NULL,-LONGTIME))
		    break;

	    if (retry < 0)
		break;

	    /* third step:
	     * skip the message sequence and wait for GO AHEAD
	     */

	    if (waitfor("\033[p",-LONGTIME))
		break;

	    /* fourth step:
	     * build a transaction and send it, wait for ACK.
	     * it is assumed the transaction fits within one block (256 chars)
	     * so we won't have to mess with splitting across blocks.
	     */

	    sprintf(MsgBuf,"%c%s\r%s\r%c",STX,PagerNum,info,ETX);

	    sum = 0;
	    p = MsgBuf;
	    while (*p != '\0')
		sum += *p++;
	    sprintf(p,"%c%c%c\r",'0' + ((sum >> 8) & 0x0f),
				 '0' + ((sum >> 4) & 0x0f),
				 '0' + (sum & 0x0f));

	    retry = 3;

	    do {
		debugtime(D_CHAT);
		debug2(D_CHAT, "sending page data\n");
		write(STDOUT,MsgBuf,strlen(MsgBuf));
		if (waitfor(NULL,-LONGTIME))
		    break;
	    } while (Match[0] == NAK && retry--);

	    retry = 3;

	    while (Match[0] != ACK && retry--)
		if (waitfor(NULL,-LONGTIME))
		    break;

	    if (retry < 0)
		break;

	    /* fifth step:
	     * indicate end-of-transactions and wait for ESC EOT
	     */

	    debugtime(D_CHAT);
	    send("\004\r");
	    waitfor("\033\004",-LONGTIME);

	    waitfor("FLUSH",MEDMTIME);		/* flush line noise */
	    result = SUCCESS;
	    break;

	case 314:				/* UCP */
	    /* dialup in data mode */
	    /* it seems they only support 2400/NONE, which is good enough
	     * and provides faster training/negotiation...
	     */
	    if (Is_ZyXEL)
		strcat(mode,"&N14&K0");

	    if (dialup(mode,0,3,PagerCentral) != SUCCESS)
		break;

	    usleep(500000);
	    send("\003\003");
	    usleep(2000000);

	    /* build a transaction and send it, wait for ACK.
	     * it is assumed the transaction fits within one block (256 chars)
	     * so we won't have to mess with splitting across blocks.
	     */

	    sprintf(MsgBuf,"%c01/%05d/O/01/%s/%s//3/",STX,
		22 + strlen(PagerNum) + strlen(PagerSource) + 2 * strlen(info),
		PagerNum,PagerSource);

	    p = MsgBuf + strlen(MsgBuf);
	    while (*info != '\0') {
		sprintf(p,"%02X",*info++);
		p += 2;
	    }
	    strcat(p,"/");

	    sum = 0;
	    p = MsgBuf + 1;
	    while (*p != '\0')
		sum += *p++;
	    sprintf(p,"%02X%c",(unsigned char)sum,ETX);

	    retry = 3;

	    do {
		waitfor("FLUSH",SHORTTIME);	/* flush line noise */
		debugtime(D_CHAT);
		send(MsgBuf);
		settermios(LONGTIME);
		p = ChatBuf;

		while ((ch = getch()) != EOF)
		    if ((p-ChatBuf) >= (sizeof(ChatBuf)-3) ||
			(*p++ = ch) == ETX)
			break;

		*p = '\0';
		debugtime(D_CHAT);
		debug4(D_CHAT,"received %.*s\n",strlen(ChatBuf)-2,ChatBuf+1);

		sum = sscanf(ChatBuf,"\00201/%d/R/01/%c/%[^/]/",&sum,mode,buf);
	    } while ((sum != 3 || mode[0] != 'A') && retry--);

	    if (retry < 0)
		break;

	    debug3(D_RUN,"message %s\n",buf);
	    waitfor("FLUSH",MEDMTIME);		/* flush line noise */
	    result = SUCCESS;
	    break;

	case 999:				/* Voice mode (message) */
	    if (!Voice)				/* needs Voice for this */
		break;

	    if (dialup(mode,8,5,PagerNum) != SUCCESS) /* dial in voice mode */
		break;

	    if (info != NULL && info[0] != '\0')
		say(info);

	    result = DIALBACK;
	    break;
	}

	if (result != DIALBACK)
	    initmodem(TRUE);			/* done, force a hangup */

	sprintf(MsgBuf,"PAGER call %s",(result == SUCCESS? "OK":"FAILED"));
	if (info != NULL && info[0] != '\0')
	    sprintf(MsgBuf + strlen(MsgBuf)," (%s)",info);

	logmesg(MsgBuf);
	return result;
}


/*
**	dir_select() - select all recorded files
*/

int
dir_select(dirent)
const struct dirent *dirent;
{
	if (dirent->d_name[0] == '.')		/* no hidden files/directories */
	    return 0;

	return 1;
}


/*
**	playback() - play the messages in the recording directory and
**		     act upon them.  option: play from another directory
*/

int
playback(direc,inputdigit)
char *direc;					/* directory, or NULL */
int (*inputdigit)(void);			/* selection input function */

{
	int nfiles,i,d;
	struct dirent **namelist = NULL;
	struct stat st;
	char filename[MAXLINE];
	char archive[MAXLINE];

	if (direc == NULL)			/* take 'direc' as dir */
	    direc = RecDir;			/* default to RecDir */

	/* this line may generate a warning for arg #3, depending on */
	/* the version of <dirent.h> */
	nfiles = scandir(direc,&namelist,dir_select,alphasort);

	debug3(D_RUN, "%d files selected to playback\n", nfiles);

	if (nfiles == 0)
	    return SUCCESS;

	d = 1;					/* forward */

	for (i = 0; i >= 0 && i < nfiles; i += d)
	{
	    if (namelist[i] == NULL)		/* deleted file? */
		continue;

	    sprintf(filename,"%s/%.*s",direc,(int)namelist[i]->d_reclen,
						  namelist[i]->d_name);

	    debug3(D_RUN, "playback file (%s)\n", filename);

play_again:
	    /* don't detect BUSY here as the message may be a recording of
	     * a busy tone, and the playback mode would be interrupted when
	     * we played that one (I found that the hard way!)
	     * (probably true for DIALTONE as well, can't test that here)
	     */
	    if (play(filename,"#") != SUCCESS) {
		for (i = 0; i < nfiles; i++)
		    free(namelist[i]);
		free(namelist);
		return FAIL;
	    }

prompt_again:
	    switch ((inputdigit == NULL)? 0 : (*inputdigit)())
	    {
	    case '0':				/* quit playback */
	    case 'b':
	    case 'd':
	    case 'h':
	    case 'l':
		for (i = 0; i < nfiles; i++)
		    free(namelist[i]);
		free(namelist);
		return SUCCESS;

	    case '1':				/* play instructions */
		if (play("playbackhelp",NULL) == SUCCESS)
		    play("playbackhelp","#bdl");
		else {
		    say("to leave playback mode, press 0.");
		    say("to get these instructions, press 1.");
		    say("to read the message date and time, press 2.");
		    say("to advance to the next message, press 3.");
		    say("to delete the current message, press 4.");
		    say("to archive the current message, press 5.");
		    say("to go to the first message, press 7.");
		    say("to replay the current message, press 8.");
		    say("to go to the last message, press 9.");
		}
		goto prompt_again;

	    case '2':				/* read message's date/time */
		if (stat(filename,&st) == 0) {
		    char *p;
		    struct tm *tm,tmnow;
		    time_t now;
		    char buf[MAXLINE],tmp[10];

		    if (Synth != NULL) {
			now = time(NULL);
			tmnow = *localtime(&now);
			tm = localtime(&st.st_mtime);

			if (tm->tm_year == tmnow.tm_year ||
			    (now - st.st_mtime) < (90 * 24 * 3600L)) {
			    if (tm->tm_yday == tmnow.tm_yday)
				p = "%H:%M today";
			    else
				p = "%A, %B %d, %H:%M";
			} else
			    p = "%A, %B %d %Y, %H:%M";

			strftime(buf,sizeof(buf),p,tm);

			if (say(buf) != SUCCESS)
			    return FAIL;
		    } else {
			strftime(buf,sizeof(buf),"%Y%m%d%H%M",localtime(&st.st_mtime));

			for (p = buf; *p != '\0'; p++) {
			    sprintf(tmp,"%c",*p);
			    if (play(tmp,"bd") != SUCCESS ||
				MsgBuf[0] == 'b' || MsgBuf[0] == 'd')
				return FAIL;
			}
		    }
		}
		goto prompt_again;

	    case '3':				/* next message */
		break;

	    case '4':				/* delete */
		if (unlink(filename) == 0) {
		    if (Synth != NULL)
			say("message deleted");
		    else
			beep(MsgDel);
		    free(namelist[i]);
		    namelist[i] = NULL;
		}
		break;

	    case '5':				/* archive message */
		sprintf(archive,"%s/%.*s",Archive,(int)namelist[i]->d_reclen,
						       namelist[i]->d_name);

		debug4(D_RUN, "rename (%s) to (%s)\n", filename, archive);

		if (strcmp(filename,archive) != 0 &&
		    rename(filename,archive) == 0)
		{
		    if (Synth != NULL)
			say("message archived");
		    else
			beep(MsgArch);
		    free(namelist[i]);
		    namelist[i] = NULL;
		}
		break;

	    case '7':				/* first message */
		say("first message");
		i = -1;
		d = 1;
		break;

	    case '8':				/* play again */
		goto play_again;

	    case '9':				/* last message */
		say("last message");
		i = nfiles;
		d = -1;
		break;
	    }
	}

	for (i = 0; i < nfiles; i++)
	    free(namelist[i]);
	free(namelist);
	return SUCCESS;
}


/*
**	dtmfdigit() - wait for a single DTMF tone and return it
*/

int
dtmfdigit()

{
	int result;
	char *p;

	/* send prompt tone */

	beep(Prompt1);

	/* record a single tone, 20 seconds max */

	if ((result = dtmfrecord("0123456789ABCD*#bdlq",20)) <= 0)
	    return result;

	/* make sure we always see a 'b', 'd', 'h' or 'l' when present */

	if ((p = strpbrk(MsgBuf,"bdhl")) != NULL)
	    return *p;

	/* return the single digit */

	return MsgBuf[0];
}


/*
**	dtmfcode() - wait for a *<digits># DTMF sequence and return value
*/

int
dtmfcode()

{
	char *p;

	/* read a string, remove * and # */

	if ((p = dtmfstring()) == NULL)
	    return FAIL;

	debug3(D_RUN, "dtmfcode() returns %d\n",atoi(p));

	/* return the value */

	return atoi(p);
}


/*
**	dtmfstring() - wait for DTMF sequence (#-terminated)
*/

char *
dtmfstring()

{
	char *p,*where;

	/* send prompt tone */

	beep(Prompt2);

	do
	{
	    /* record a sequence, 20 seconds max */

	    if (dtmfrecord("#bdlq",20) <= 0)
		return NULL;

	    /* make sure we always see a 'b', 'd', 'h' or 'l' when present */

	    if (strpbrk(MsgBuf,"bdhl") != NULL)
		return NULL;
	} while (!strcmp(MsgBuf,"#"));		/* ignore only-# entry */

	/* remove everything up to last * (erase character) */

	where = MsgBuf;

	while ((p = strchr(where,'*')) != NULL)
	    where = p + 1;

	/* delete terminating # */

	if ((p = strchr(where,'#')) != NULL)
	    *p = '\0';

	/* delete any spurious 'silence' indications */

	while (*p == 's')
	    p++;

	debug3(D_RUN, "dtmfstring() returns (%s)\n",where);

	/* return the value, it still resides in MsgBuf! */

	return where;
}

/*************************************************************************
**	Voice mode low level routines					**
*************************************************************************/

/*
**	dtmfrecord() - get DTMF info using most efficient method
**
**	entered DTMF string is placed in MsgBuf, recording time is returned
*/

int
dtmfrecord (end,maxtime)
const char *end;			/* set of codes that ends recording */
int maxtime;				/* max recording time in seconds */

{
	char vsd[20];

	/* set the silence detection parameter suitable for the recording */
	/* interval (up to 25.5 seconds) */

	sprintf(vsd,"%d,%d",(Silence != NULL? atoi(Silence) : 20),10 * maxtime);

	/* newer versions allow recording of DTMF without recording voice */
	/* this is done by setting S39.6=1 and then (re-)issueing the +VLS */
	/* command to select the telephone line (only on the U-1496). */
	/* it only works when the phone line is selected for recording... */

	if ((Fw_features & FW_DTMFREC) && (CurLS & 2)) {
	    char *msg;
	    register int ch;
	    sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();

	    debugtime(D_RUN);
	    debug4(D_RUN, "dtmfrecord(\"%s\",%d)\n", end, maxtime);

	    /* set silence detection for DTMF recording */

	    if (strcmp(CurVSD,vsd)) {
		sprintf(ChatBuf,"AT+VSD=%s\r",vsd);
		send(ChatBuf);
		if (waitfor("OK",MEDMTIME))
		    return FAIL;
		strcpy(CurVSD,vsd);
	    }

	    /* trap signals occurring during recording, so that we can */
	    /* shut up the MODEM before exiting */

	    Signal = 0;
	    oldint = signal(SIGINT, sighandler);
	    oldquit = signal(SIGQUIT, sighandler);
	    oldterm = signal(SIGTERM, sighandler);
	    oldalarm = signal(SIGALRM, sighandler);

	    if (!(Fw_features & FW_2864)) {
		send("ATS39.6=1+VLS=2\r");	/* enable DTMF recording from phone */
		if (wait_vcon(MEDMTIME) == FAIL) { /* sometimes OK, sometimes VCON */
		    maxtime = FAIL;
		    goto error;
		}
	    }

	    /* set a maximal recording time limit */

	    if (maxtime <= 0 || maxtime > 120)	/* sanity check */
		maxtime = 120;

	    alarm(maxtime);

	    settermios(LONGTIME);
	    *(msg = MsgBuf) = '\0';

	    /* read the DTMF characters */

	    for (;;)
	    {
		switch (ch = getch())
		{
		case EOF:
		    break;			/* nothing, try again */

		case DLE:
		    switch (ch = getch())
		    {
		    case EOF:
			break;

		    case ETX:			/* end of message! */
			goto done;

		    default:			/* something else: DTMF or so */
			if (msg < (MsgBuf + sizeof(MsgBuf) - 1)) {
			    *msg++ = ch;	/* place codes in MsgBuf */
			    *msg = '\0';
			}

			/* terminate recording on certain character codes */

			if (strchr(end,ch) != NULL)
			    goto done;

			/* always terminate recording when 's'ilence */
			/* (and 's' not in the end codes) */

			if (ch == 's')
			    goto done;

			break;
		    }
		    break;

		default:			/* unexpected character */
		    alarm(0);
		    maxtime = FAIL;		/* return FAIL */
		    goto error;
		}

		if (Signal) {			/* timeout or other signal? */
		    Signal = 0;
		    alarm(0);
		    maxtime = FAIL;		/* return FAIL */
		    goto error;
		}
	    }

done:
	    if ((maxtime -= alarm(0)) == 0)	/* stop clock & measure time */
		maxtime = 1;			/* at least one second */

error:
	    if (!(Fw_features & FW_2864)) {
		send("ATS39.6=0+VLS=2\r");	/* turnoff the DTMF recording */
		if (wait_vcon(MEDMTIME) == FAIL)
		    return FAIL;
	    }

	    signal(SIGINT, oldint);
	    signal(SIGQUIT, oldquit);
	    signal(SIGTERM, oldterm);
	    signal(SIGALRM, oldalarm);
	    Signal = 0;

	    debugtime(D_RUN);
	    debug4(D_RUN, "dtmfrecord() returns %d, MsgBuf (%s)\n", maxtime, MsgBuf);
	    return maxtime;
	}

	/* do a recording with NULL output */

	return record(NULL,FALSE,end,vsd,ENC_CELP,maxtime);
}

/*
**	record() - record a message from current speaker, or speaker defined
**		   by RecLS
**
**	return number of seconds recorded, 0 for SILENCE, or FAIL
**	expanded filename will be returned in original 'filename' arg!
*/

int
record(filename,beeps,end,vsd,enc,maxtime)
char *filename;				/* NULL to discard data (DTMF only) */
boolean beeps;				/* send prompting beeps? */
const char *end;			/* set of codes that ends recording */
const char *vsd;			/* silence detection params */
int enc;				/* CELP/ADPCM2/ADPCM3 etc */
int maxtime;				/* max recording time in seconds */
{
	register FILE *fp = NULL;
	register int ch;
	int done;
	int result = 1;
	int cur_ls = -1;
	int soundc;
	char *msg;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	ZAD_HEADER header;

#ifdef DEBUG
	debugtime(D_RUN);
	if (filename != NULL)
	    debug3(D_RUN, "record(\"%s\"", filename)
	else
	    debug2(D_RUN, "record(NULL");
	debug4(D_RUN, ",\"%s\",\"%s\"", end, vsd);
	debug4(D_RUN, ",%d,%d)\n", enc, maxtime);
#endif

	/* sound an attention beep first on Linux ISDN, it calls ausend */

	if ((Fw_features & FW_LINUX) && beeps && beep(RecAttn) != SUCCESS) {
	    return FAIL;
	}

	/* set silence detection for message recording */

	if (vsd != NULL && strcmp(CurVSD,vsd)) {
	    sprintf(ChatBuf,"AT+VSD=%s\r",vsd);
	    send(ChatBuf);
	    if (waitfor("OK",MEDMTIME))
		return FAIL;
	    strcpy(CurVSD,vsd);
	}

	/* set vsm scheme */
	/* adapt to "nearest value" when attached modem does not support */
	/* the desired scheme */

	if ((enc = set_vsm(enc)) == FAIL)
	    return FAIL;

	/* create the output file, when requested */

	if (filename != NULL) {
	    /* create filename for recording */

	    if (RecDir != NULL && filename[0] != '/') {
		char buf[MAXLINE];

		sprintf(buf,"%s/%s",RecDir,filename);
		if (strcmp(filename + strlen(filename) - 4,AudioExt))
		    strcat(buf,AudioExt);
		strcpy(filename,buf);
	    }

	    /* %-macro's will be expanded using strftime, to generate */
	    /* a time-dependent recording file */

	    if (strchr(filename,'%') != NULL) {
		char buf[MAXLINE];

		RecTime = time(NULL);
		strftime(buf,sizeof(buf),filename,localtime(&RecTime));
		strcpy(filename,buf);
	    }

	    if ((fp = fopen(filename,"w")) == NULL) {
		sprintf(MsgBuf, "cannot create audiofile (%s)", filename);
		logmesg(MsgBuf);
		return FAIL;
	    }

	    /* prepare an audio file header */

	    memset(&header,0,sizeof(header));
	    memcpy(header.zyxel,ZAD_MAGIC,sizeof(header.zyxel));
	    header.enc = enc;

	    fwrite(&header,sizeof(header),1,fp);

	    sound_mode(enc,Fsamp);		/* set encoding on soundcard */
	}

	/* trap signals occurring during recording, so that we can */
	/* shut up the MODEM before exiting */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	*(msg = MsgBuf) = '\0';

	/* when RecLS has been set, save the current LS and set it */

	if (RecLS != 0 && RecLS != CurLS) {
	    cur_ls = CurLS;

	    if (set_vls(RecLS)) {		/* select speaker */
		debug3(D_RUN, "record() cannot set LS to %d\n", RecLS);
		result = FAIL;
		goto done;
	    }

	    debugtime(D_RUN);
	    debug4(D_RUN, "record() VLS was %d, is %d\n", cur_ls, RecLS);
	}

	/* set a maximal recording time limit, so that the MODEM will */
	/* not fill the entire disk when the silence detection fails... */

	if (maxtime <= 0 || maxtime > 900)	/* sanity check */
	    maxtime = 120;

	alarm(maxtime);

	/* sound an attention beep now on ZyXEL MODEM, it uses +VTS= */

	if (!(Fw_features & FW_LINUX) && beeps && beep(RecAttn) != SUCCESS) {
	    result = FAIL;
	    goto done;
	}

	/* start recording */

	send("AT+VRX\r");
	debugtime(D_CHAT);
	if (waitfor("CONNECT",MEDMTIME)) {
	    send(AudioRecAbort);		/* make sure we abort it */
	    result = FAIL;
	    goto done;
	}

	/* don't switch to XON/XOFF mode here.  we are receiving */
	/* binary data and eating these characters is not a good idea... */
	/* let's simply assume we are fast enough to keep up */

	settermios(SHORT255);
	if (!(Fw_features & FW_RTSCTSFLOW))	/* >= 6.13: RTS/CTS flow */
	    write(STDOUT,"\021",1);		/* send XON to be sure */

	done = soundc = 0;

	/* read the voice data, it is terminated by DLE-ETX */

	for (;;)
	{
	    switch (ch = getch())
	    {
	    case EOF:
		debugtime(D_RUN);
		debug2(D_RUN, "record() timeout\n");
		send(AudioRecAbort);		/* abort recording */
#ifdef DEBUG
		linestat();
#endif
		goto done;

	    case DLE:
		switch (ch = getch())
		{
		case EOF:
		    debugtime(D_RUN);
		    debug2(D_RUN, "record() timeout\n");
		    send(AudioRecAbort);	/* abort recording */
#ifdef DEBUG
		    linestat();
#endif
		    goto done;

		case ETX:			/* end of message! */
		    goto done;

		case DLE:			/* doubled DLE */
		case DC2:			/* resync symbol */
		case FS:			/* resync symbol */
		    if (enc == ENC_ALAW && (Fw_features & FW_LINUX))
			ch = rev_bits[ch];

		    if (fp != NULL) {
			if (enc & ENC_RESYNC)	/* when we are doing resync */
			    putc(DLE,fp);	/* the DLE should be in the file */

			putc(ch,fp);

			if (SoundFp != NULL)
			    putc(ch,SoundFp);
		    }

		    break;

		case SUB:			/* escaped DLE-DLE (2864) */
		    if (fp != NULL) {
			if (enc & ENC_RESYNC) { /* when we are doing resync */
			    putc(DLE,fp);	/* the DLE should be in the file */
			    putc(SUB,fp);	/* plus the SUB code */
			} else {
			    putc(DLE,fp);	/* data is two DLEs */
			    putc(DLE,fp);
			}

			if (SoundFp != NULL) {
			    putc(DLE,SoundFp);
			    putc(DLE,SoundFp);
			}
		    }
		    break;

		case DC4:			/* line hung up (Linux ISDN) */
		    ch = 'l';			/* treat as DLE-l (line hangup) */
		default:			/* something else: DTMF or so */
		    if (msg < (MsgBuf + sizeof(MsgBuf) - 1)) {
			*msg++ = ch;		/* place codes in MsgBuf */
			*msg = '\0';
		    }

		    /* terminate recording on certain character codes */

		    if (!done && strchr(end,ch) != NULL)
		    {
			debugtime(D_RUN);
			debug3(D_RUN, "record() terminated by '%c'\n", ch);
			send(AudioRecAbort);	/* abort recording */
			settermios(SHORTTIME);
			done = 1;
			break;
		    }

		    /* always terminate recording when 's'ilence */
		    /* (and 's' not in the end codes) */

		    if (ch == 's')
		    {
			result = 0;		/* will delete file */
			debugtime(D_RUN);
			debug2(D_RUN, "record() terminated by silence\n");
			if (!done) {
			    send(AudioRecAbort); /* abort recording */
			    settermios(SHORTTIME);
			    done = 1;
			    break;
			}
		    }

		    debugtime(D_RUN);
		    debug3(D_RUN, "record() received DLE-%c\n", ch);
		    break;
		}
		break;

	    default:
		if (enc == ENC_ALAW && (Fw_features & FW_LINUX))
		    ch = rev_bits[ch];

		if (fp != NULL) {
		    putc(ch,fp);		/* normal voice data */

		    if (SoundFp != NULL) {
			putc(ch,SoundFp);

			if (++soundc == 1024) {	/* limit the lag */
			    fflush(SoundFp);
			    soundc = 0;
			}
		    }
		}
		break;
	    }

	    if (Signal && !done) {
		debugtime(D_RUN);
		debug3(D_RUN, "record() terminated by signal %x\n", Signal);
		Signal = 0;
		send(AudioRecAbort);		/* abort recording */
		settermios(SHORTTIME);
		done = 1;
	    }
	}

done:
	maxtime -= alarm(0);			/* stop clock & measure time */

	if (SoundFp != NULL)
	    fflush(SoundFp);

	debugtime(D_CHAT);
	if (wait_vcon(MEDMTIME) == NOCARRIER)	/* get the result */
	    strcat(msg,"l");			/* add hangup code */

	/* return recorded time when no failure occurred */

	if (result > 0 && maxtime != 0)
	    result = maxtime;

	/* when current LS was saved, restore it now */

	if (cur_ls >= 0 && set_vls(cur_ls))
	    result = FAIL;

	if (beeps)				/* signal end of recording */
	    beep(RecDone);
	else
	    if (Fw_features & FW_LINUX)
		usleep(500000);

	if (fp != NULL && fclose(fp) == EOF)
	    result = FAIL;			/* disk full or so */

	if (filename != NULL && result <= 0)
	    unlink(filename);			/* failure or silence, remove the file */

	signal(SIGINT, oldint);
	signal(SIGQUIT, oldquit);
	signal(SIGTERM, oldterm);
	signal(SIGALRM, oldalarm);
	Signal = 0;

	debugtime(D_RUN);
	debug4(D_RUN, "record() returns %d, MsgBuf (%s)\n", result, MsgBuf);
	return result;
}


/*
**	beep() - send a tone string
*/

int
beep(s)
const char *s;

{
	int res;
	char buf[MAXLINE];		/* local buffer for command */
	char msgbuf[MAXLINE];		/* save buffer for MsgBuf */

	if (s != NULL) {
	    if (Fw_features & FW_LINUX) {
		debug3(D_RUN, "beep(\"%s\")\n", s);

		strcpy(msgbuf,MsgBuf);
		strcpy(Beep = buf,s);
		Bsamples = 0;
		res = ausend(0,readbeep,15);
		strcpy(MsgBuf,msgbuf);

		return res;
	    } else {
		sprintf(buf,"AT%s%s+VTS=%s\r",make_fclass(8),
			    (CurLS == 0? make_vls(16) : ""),s);
		send(buf);
		return wait_vcon(LONGTIME);
	    }
	}

	return SUCCESS;			/* NULL string, always ok */
}

/*
**	readbeep() - return beep data as 16-bit audio samples
*/

int
readbeep(fd,buf,size)
int fd;
char *buf;
size_t size;

{
	short *p;
	int n;
	int f1,f2,t;

	p = (short *)buf;
	n = size / 2;

	while (n > 0)
	{
	    if (Bsamples <= 0) {
		if (Beep == NULL || *Beep == '\0') /* end of spec? */
		    break;
		
		if (BeepPause) {
		    BeepPause = 0;
		    Beep++;
		    Bsamples = Fsamp / 8;
		    Beep1 = Beep2 = 0;
		} else {
		    BeepPause = 1;
		    Bsamples = Fsamp / 12;

		    switch (*Beep)
		    {
		    case '1': f1 = 697; f2 = 1209; break;
		    case '2': f1 = 697; f2 = 1336; break;
		    case '3': f1 = 697; f2 = 1477; break;
		    case '4': f1 = 770; f2 = 1209; break;
		    case '5': f1 = 770; f2 = 1336; break;
		    case '6': f1 = 770; f2 = 1477; break;
		    case '7': f1 = 852; f2 = 1209; break;
		    case '8': f1 = 852; f2 = 1336; break;
		    case '9': f1 = 852; f2 = 1477; break;
		    case '*': f1 = 941; f2 = 1209; break;
		    case '0': f1 = 941; f2 = 1336; break;
		    case '#': f1 = 941; f2 = 1477; break;
		    case 'A': f1 = 697; f2 = 1633; break;
		    case 'B': f1 = 770; f2 = 1633; break;
		    case 'C': f1 = 852; f2 = 1633; break;
		    case 'D': f1 = 941; f2 = 1633; break;
		    case 'E': f1 = 500; f2 = 0; break;
		    case 'H': f1 = 1866; f2 = 0; break;
		    case 'L': f1 = 933; f2 = 0; break;

		    case '[':			/* dualtone spec */
			f1 = f2 = t = 0;
			sscanf(Beep,"[%d,%d,%d]",&f1,&f2,&t);
			if ((Beep = strchr(Beep,']')) != NULL)
			    Beep++;

			BeepPause = 0;
			Bsamples = (Fsamp * t) / 10;
			break;

		    default:
		    	f1 = f2 = 0;
			break;
		    }

		    Beep1 = (6.283185307 * f1) / Fsamp;
		    Beep2 = (6.283185307 * f2) / Fsamp;
		}
	    }

	    while (n-- > 0 && Bsamples--) {
		*p++ = (sin(Beep1 * Samples) + sin(Beep2 * Samples)) * 3276;
		Samples++;
	    }
	}

	/* return number of generated bytes */

	return (int)((char*)p - buf);
}


/*
**	play() - play the specified file to current speaker
**
*/

int
play(filename,end)
const char *filename;
const char *end;			/* set of codes that ends playing */
{
	FILE *fp;
	char *p,*msg;
	int num,c,s,dle,done,state,enc;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	ZAD_HEADER header;
	char buf[BUFSIZ+32];

#ifdef DEBUG
	debugtime(D_RUN);
	if (end != NULL)
	    debug4(D_RUN, "play(\"%s\",\"%s\")\n", filename, end)
	else
	    debug3(D_RUN, "playcheck(\"%s\")\n", filename);
#endif

	/* when it isn't a full pathname, make it into one (and add extension) */

	if (PlayDir != NULL && filename[0] != '/') {
	    sprintf(buf,"%s/%s",PlayDir,filename);
	    if (strcmp(filename + strlen(filename) - 4,AudioExt))
		strcat(buf,AudioExt);
	    filename = buf;
	}

	/* open the file */

	if ((fp = fopen(filename,"r")) == NULL) {
	    if (end != NULL) {
		sprintf(MsgBuf, "cannot open audiofile (%s)", filename);
		logmesg(MsgBuf);
		MsgBuf[0] = '\0';
	    }
	    return ERROR;
	}

	/* read and check the file header */

	if (fread(&header,sizeof(header),1,fp) != 1 ||
	    memcmp(header.zyxel,ZAD_MAGIC,sizeof(header.zyxel)))
	{
	    fclose(fp);
	    sprintf(MsgBuf, "not a valid audiofile (%s)", filename);
	    logmesg(MsgBuf);
	    MsgBuf[0] = '\0';
	    return FAIL;
	}

	/* work around bug that causes play() to hang when CND is ON */

	if ((Fw_features & FW_PLAYBUG) && !PlayCND) {
	    send("ATS40.2=0\r");
	    if (waitfor("OK",MEDMTIME))
		return FAIL;
	    PlayCND = TRUE;
	}

	/* set encoding scheme (found in header, or BEST for Linux) */

	if ((Fw_features & FW_LINUX) && header.enc < ENC_ALAW)
	    enc = ENC_BEST;
	else
	    enc = header.enc;

	/* set the modem to this encoding */

	if ((c = set_vsm(enc)) != enc)
	    if ((enc == ENC_ALAW || enc == ENC_ULAW) && c == ENC_ADPCM4)
		enc = c;
	
	/* when it doesn't support it, value is modified and we FAIL */

	if (c != enc) {
	    debug3(D_RUN, "playfile encoding 0x%02x not supported by modem\n", header.enc);
	    fclose(fp);
	    MsgBuf[0] = '\0';
	    return FAIL;
	}

	/* when everything OK up to here and end=NULL, return */
	/* this is used to check audio file existence/correctness */

	if (end == NULL) {
	    debug2(D_RUN, "playcheck() OK\n");
	    fclose(fp);
	    MsgBuf[0] = '\0';
	    return SUCCESS;
	}

	if (enc != header.enc) {
	    initadpcm();			/* init the adpcm decoder */
	    sound_mode(enc,9600);
	} else
	    sound_mode(enc,Fsamp);		/* set the soundcard */

	/* start playing */

	debugtime(D_CHAT);
	send("AT+VTX\r");
	if (waitfor("CONNECT",MEDMTIME)) {
	    fclose(fp);
	    MsgBuf[0] = '\0';
	    return FAIL;
	}

	/* trap signals occurring during playing, so that we can */
	/* reset the MODEM before exiting */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	if (Fw_features & FW_RTSCTSFLOW)
	    state = IMMEDIATE;			/* >= 6.13: hw flow ctl */
	else
	    state = XONXOFF;			/* XON/XOFF flow ctl */

	settermios(state);

	debugtime(D_CHAT);
	debug2(D_CHAT, "play() sending audio data\n");

	dle = done = num = 0;
	*(msg = MsgBuf) = '\0';

	for (;;)
	{
	    alarm(BUFSIZ/200);			/* watchdog in case XON is missed */

	    /* read a block of audio data from the file */

	    if (!done)
	    {
		if (header.enc & ENC_RESYNC)	/* resync info in data? */
		{
		    if ((num = fread(buf,1,BUFSIZ,fp)) > 0)
			write(STDOUT,buf,num);
		    else
			done = 1;
		}
		else				/* no resync info */
		{
		    p = buf;
#define	putbyte(byte)	{ if (byte == DLE) *p++ = DLE; *p++ = byte; }
#define	putres(res)	{ s = linear2alaw(res); if (SoundFp != NULL) putc(s,SoundFp); if (enc == ENC_ALAW && (Fw_features & FW_LINUX)) s = rev_bits[s]; if (++num % 6) putbyte(s); }

		    while ((p - buf) < BUFSIZ)	/* fill buffer with data */
		    {
			if ((c = getc(fp)) == EOF) {
			    done = 1;
			    break;
			}

			if (enc == header.enc) { /* conversion needed? */
			    if (SoundFp != NULL)
				putc(c,SoundFp);

			    if (enc == ENC_ALAW && (Fw_features & FW_LINUX))
				c = rev_bits[c];

			    putbyte(c);
			    continue;
			}

			switch (header.enc)
			{
			case ENC_ADPCM2:
			    putres(fromadpcm(ENC_ADPCM2,c >> 6));	/* XX------ */
			    putres(fromadpcm(ENC_ADPCM2,(c >> 4) & 3));	/* --XX---- */
			    putres(fromadpcm(ENC_ADPCM2,(c >> 2) & 3));	/* ----XX-- */
			    putres(fromadpcm(ENC_ADPCM2,c & 3));	/* ------XX */
			    break;

			case ENC_ADPCM3:
			    putres(fromadpcm(ENC_ADPCM3,c >> 5));	/* XXX----- */
			    putres(fromadpcm(ENC_ADPCM3,(c >> 2) & 7));	/* ---XXX-- */

			    if ((s = getc(fp)) == EOF)
				s = 0;

			    c = (c << 8) | s;
			    putres(fromadpcm(ENC_ADPCM3,(c >> 7) & 7));	/* X-----XX */
			    putres(fromadpcm(ENC_ADPCM3,(c >> 4) & 7));	/* -XXX---- */
			    putres(fromadpcm(ENC_ADPCM3,(c >> 1) & 7));	/* ----XXX- */

			    if ((s = getc(fp)) == EOF)
				s = 0;

			    c = (c << 8) | s;
			    putres(fromadpcm(ENC_ADPCM3,(c >> 6) & 7));	/* XX-----X */
			    putres(fromadpcm(ENC_ADPCM3,(c >> 3) & 7));	/* --XXX--- */
			    putres(fromadpcm(ENC_ADPCM3,c & 7));	/* -----XXX */
			    break;

			case ENC_ADPCM4:
			    putres(fromadpcm(ENC_ADPCM4,c >> 4));	/* XXXX---- */
			    putres(fromadpcm(ENC_ADPCM4,c & 0x0f));	/* ----XXXX */
			    break;

			case ENC_ALAW:
			    s = toadpcm(ENC_ADPCM4,alaw2linear(c)) << 4; /* XXXX---- */

			    c = getc(fp);
			    s |= toadpcm(ENC_ADPCM4,alaw2linear(c));	/* ----XXXX */

			    putbyte(s);
			    break;

			case ENC_ULAW:
			    s = toadpcm(ENC_ADPCM4,ulaw2linear(c)) << 4; /* XXXX---- */

			    c = getc(fp);
			    s |= toadpcm(ENC_ADPCM4,ulaw2linear(c));	/* ----XXXX */

			    putbyte(s);
			    break;

			default:
			    putbyte(c);
			    break;
			}
		    }

		    if (p != buf)
			write(STDOUT,buf,p - buf);
#undef	putres
#undef	putbyte
		}

		if (done)
		{
		    alarm(5);			/* watchdog */
		    settermios(state);		/* wait until all sent */
		    if (Fw_features & FW_LINUX)
			usleep(500000);
		    debugtime(D_CHAT);
		    send(AudioEnd);		/* DLE-ETX ends the data */
		    settermios(LONGTIME);

		    if (SoundFp != NULL)
			fflush(SoundFp);
		}
	    }

	    /* read the returned characters, can contain DTMF codes etc */

	    while ((c = getch()) != EOF || done)
	    {
		if (dle)			/* previous was a DLE? */
		{
		    switch (c)			/* what is next character? */
		    {
		    case ETX:			/* end of message! */
			break;			/* ignore when sending */

		    case DLE:			/* doubled DLE */
		    case SUB:			/* DLE-SUB (2864) */
			break;

		    case DC4:			/* line hung up (Linux ISDN) */
			c = 'l';		/* treat as DLE-l (line hangup) */
		    default:			/* something else: DTMF or so */
			*msg++ = c;		/* place codes in MsgBuf */
			*msg = '\0';

			/* terminate playing on certain passed chars */

			if (!done &&
			    (strchr(end,c) != NULL ||
			     msg > (MsgBuf + sizeof(MsgBuf) - 3)))
			{
			    debugtime(D_RUN);
			    debug3(D_RUN, "play() terminated by '%c'\n", c);
			    ioctl(STDOUT,TCFLSH,TCOFLUSH);
			    ioctl(STDOUT,TCXONC,TCOON);
			    usleep(20000);
			    alarm(5);		/* watchdog */
			    send(AudioAbort);
			    settermios(LONGTIME);

			    if (SoundFp != NULL)
				fflush(SoundFp);

			    done = 1;
			} else {
			    if (Fw_features & FW_DTMFPAUSE)
				send(AudioResume); /* send DLE-r to resume */

			    debugtime(D_RUN);
			    debug3(D_RUN, "play() received DLE-%c\n", c);
			}
			break;
		    }

		    dle = 0;
		} else {
		    switch (c)			/* what is next character? */
		    {
		    case DLE:			/* is it a DLE? */
			dle = 1;
			break;

		    case XON:			/* a flow control char leaked thru? */
		    case XOFF:
			break;			/* ignore it */

		    default:			/* something else, probably ^M */
			settermios(INITIAL);

			if (done) {
			    /* this will be the VCON/OK at end of play */

			    alarm(0);
			    if (c != EOF) {
				InPtr--;	/* push it back */
				debugtime(D_CHAT);
				wait_vcon(MEDMTIME); /* let it finish */
			    }
			} else {
			    /* it came in while we were sending */
			    /* probably DATA/VOICE pressed or so */

			    debugtime(D_RUN);
			    debug3(D_RUN, "play() terminated by unexpected char %02x\n", c);
			    InPtr--;		/* push it back */
			    ioctl(STDOUT,TCFLSH,TCOFLUSH);
			    ioctl(STDOUT,TCXONC,TCOON);
			    usleep(20000);
			    send(AudioAbort);
			    waitfor(NULL,LONGTIME); /* let it riddle */
			}

			alarm(0);
			fclose(fp);

			signal(SIGINT, oldint);
			signal(SIGQUIT, oldquit);
			signal(SIGTERM, oldterm);
			signal(SIGALRM, oldalarm);
			Signal = 0;

			if (SoundFp != NULL)
			    fflush(SoundFp);

			if (checksync("play") != SUCCESS)
			    return FAIL;

			debugtime(D_RUN);
			debug3(D_RUN, "play() returns MsgBuf (%s)\n", MsgBuf);
			return SUCCESS;
		    }
		}
	    }

	    /* see if we have been requested to stop the playing */

	    if (Signal && !done) {
		debugtime(D_RUN);
		debug3(D_RUN, "play() terminated by signal %x\n", Signal);
		Signal = 0;
		ioctl(STDOUT,TCFLSH,TCOFLUSH);
		ioctl(STDOUT,TCXONC,TCOON);
		usleep(20000);
		send(AudioAbort);
		settermios(LONGTIME);
		done = 1;
	    }
	}
}

/*
**	auplay() - play audio file
**
*/

int
auplay(filename)
char *filename;
{
	int fd,ret;
	int fsamp;
	char buf[4];
	extern int read();

	/* open the file */

	if ((fd = open(filename,O_RDONLY)) < 0) {
	    sprintf(MsgBuf, "cannot open aufile (%s) errno=%d", filename, errno);
	    logmesg(MsgBuf);
	    return ERROR;
	}

	/* when it starts with .snd, assume 8k/s, else 9k6.  crude. */

	if (read(fd,buf,4) == 4 && !memcmp(buf,".snd",4)) {
	    fsamp = 8000;
	    lseek(fd,32L,SEEK_SET);
	} else {
	    fsamp = 9600;
	    lseek(fd,0L,SEEK_SET);
	}

	if (Fsamp == 9600 && fsamp == 8000)
	    ret = ausend(fd,readulaw8k,0);
	else
	    if (Fsamp == 8000 && fsamp == 9600)
		ret = ausend(fd,readulaw9k6,0);
	    else
		ret = ausend(fd,readulaw,0);

	close(fd);
	return ret;
}

/*
**	music() - play MIDI or MOD file
*/

int
music(filename)
char *filename;
{
	int fd,ret,argc;
	pid_t pid;
	char *prog,*flags;
	char *p,*q;
	unsigned char buf[4];
	char *argv[NARG];
	char flagbuf[MAXLINE];
	extern int read();

	debugtime(D_RUN);
	debug3(D_RUN, "music(\"%s\")\n", filename);

	/* open the file */

	if ((fd = open(filename,O_RDONLY)) < 0) {
	    sprintf(MsgBuf, "cannot open aufile (%s) errno=%d", filename, errno);
	    logmesg(MsgBuf);
	    return ERROR;
	}

	/* read magic number */

	if (read(fd,buf,4) != 4) {
	    close(fd);
	    return FAIL;
	}

	lseek(fd,0L,SEEK_SET);

	/* check the file type (MIDI or MOD) */

	if (!memcmp(buf,"MThd",4)) {
	    prog = MidPlay;
	    flags = MidPlayFlags;
	} else if ((buf[0] == 0xff && (buf[1] & 0xf0) == 0xf0) ||
		   !memcmp(buf,"ID3",3)) {
	    prog = Mp3Play;
	    flags = Mp3PlayFlags;
	} else {
	    prog = ModPlay;
	    flags = ModPlayFlags;
	}

	/* make sure there is a program to play these files */

	if (prog == NULL) {
	    close(fd);
	    return FAIL;
	}

	/* prepare args */

	memset(argv,0,sizeof(argv));
	argc = 0;

	if ((p = strrchr(prog,'/')) != NULL)
	    p++;
	else
	    p = prog;

	argv[argc++] = p;

	if (flags != NULL) {
	    strcpy(q = flagbuf,flags);

	    while ((p = strchr(q,' ')) != NULL) {
		*p++ = '\0';
		argv[argc++] = q;
		q = p;
	    }

	    argv[argc++] = q;
	}

	if (push(&pid,&fd,prog,argv) != SUCCESS) {
	    close(fd);
	    return FAIL;
	}

	ret = ausend(fd,(int(*)())read,600);

	close(fd);
	alarm(10);
	waitpid(pid,&fd,0);
	alarm(0);

	return ret;
}

/*
**	say() - call voice synthesizer program to read string
**
*/

int
say(string)
const char *string;
{
	int ifd,ofd,ret,len;
	pid_t pid;

	if (Synth == NULL)
	    return FAIL;

	debugtime(D_RUN);
	debug3(D_RUN, "say(\"%s\")\n", string);

	/* start the synthesizer process, return pid and pipe descriptors */

	if (startsynth(&pid,&ifd,&ofd) != SUCCESS)
	    return FAIL;

	/* send the string to the synthesizer process so it can */
	/* start off preparing the output */

	len = strlen(string);
	ret = write(ofd,string,len);
	close(ofd);

	if (ret != len)
	    return FAIL;

	/* transcode the audio data and send it to the modem */

	ret = ausend(ifd,auread,120);

	/* wait for the synthesizer process to finish */

	close(ifd);
	alarm(10);
	waitpid(pid,&ofd,0);
	alarm(0);

	return ret;
}

/*
**	run() - run an external program
**
**	read stdout on synthesizer when flag is TRUE
**	re-init the modem and make it available to external program when FALSE
**
*/

int
run(synth,argv)
int synth;
char *argv[];
{
	int ifd,ofd,ret,i;
	pid_t spid,rpid;
	char *prog;

	debugtime(D_RUN);
	debug4(D_RUN, "run(%d,\"%s\")\n",synth,argv[0]);

	if (synth) {
	    /* start the synthesizer process, return pid and pipe descriptors */

	    if (Synth != NULL && startsynth(&spid,&ifd,&ofd) != SUCCESS)
		return FAIL;
        } else {
	    /* reset the modem */
	    if (initmodem(TRUE) != SUCCESS)
		return FAIL;

	    /* now delete the lockfile to allow access to line */
	    rmlocks();
	}

	/* start the program to run */

	switch (rpid = fork())
	{
	case -1:				/* fork() failure */
	    if (synth && Synth != NULL) {
		close(ofd);
		close(ifd);
		alarm(10);
		waitpid(spid,&ofd,0);
		alarm(0);
	    }

	    logmesg("cannot fork 'run' process");
	    return FAIL;

	case 0:					/* child process */
	    close(STDIN);
	    open("/dev/null",O_RDWR);		/* STDIN to /dev/null */

	    if (synth && Synth != NULL)
		dup2(ofd,STDOUT);		/* STDOUT to synthesizer */
	    else {
		close(STDOUT);
		open("/dev/null",O_RDWR);	/* STDOUT to /dev/null */
	    }

	    close(STDERR);
	    open("/dev/null",O_RDWR);		/* STDERR to /dev/null */

	    for (i = 3; i < 20; i++)
		close(i);			/* close remaining files */
	    while (close(i) == 0)
		i++;

	    while (isalpha(argv[0][0]) && strchr(argv[0],'=') != NULL) {
		putenv(argv[0]);		/* put VAR=value in env */
		argv++;
	    }

	    prog = argv[0];

	    if ((argv[0] = strrchr(prog,'/')) != NULL) /* get basename */
		argv[0]++;
	    else
		argv[0] = prog;

	    for (i = 1; argv[i] != NULL; i++)	/* substitute magic args */
		if (argv[i][0] == '@' && strlen(argv[i]) == 2)
		    switch (argv[i][1])
		    {
		    case 'C':			/* @C -> last received CONNECT */
			argv[i] = Connect? Connect:"";
			break;

		    case 'I':			/* @I -> caller ID */
			argv[i] = CndNumber? CndNumber:"";
			break;

		    case 'J':			/* @J -> caller name */
			argv[i] = CndName? CndName:"";
			break;

		    case 'L':			/* @L -> line (Device) name */
			argv[i] = Device;
			break;

		    case 'M':			/* @M -> modem version */
			argv[i] = Modem;
			break;

		    case 'N':			/* @N -> readnumber value */
			argv[i] = Number;
			break;

		    case 'R':			/* @R -> current recording file name */
			argv[i] = Recordfile;
			break;

		    case 'S':			/* @S -> system nodename */
			argv[i] = SysName;
			break;
		    }

#ifdef DEBUG
	    if (Dfp != NULL) {
		fclose(Dfp);
		Dfp = NULL;
	    }
#endif	/* DEBUG */

	    for (i = 0; i < NSIG; i++)		/* set all signals to SIG_DFL */
		signal(i, SIG_DFL);

	    nice(9);				/* decrease priority */
	    execv(prog,argv);
	    logmesg("cannot execute 'run' process");
	    exit(1);

	default:				/* parent process */
	    if (synth && Synth != NULL)
		close(ofd);
	    break;
	}

	if (synth && Synth != NULL) {
	    /* transcode the audio data and send it to the modem */

	    ret = ausend(ifd,auread,120);

	    /* wait for the synthesizer process to finish */

	    close(ifd);
	    alarm(10);
	    waitpid(spid,&ofd,0);
	    alarm(0);
	} else
	    ret = SUCCESS;

	/* wait for 'run' process to finish */

	alarm(60);
	waitpid(rpid,&ofd,0);
	alarm(0);

	if (!synth) {
	    /* reestablish the lock file */
	    while (makelock(Lock) != SUCCESS)
		sleep(15);

	    /* reset the modem */
	    if (initmodem(TRUE) != SUCCESS)
		return FAIL;
	}

	return ret;
}

/*
**	startsynth() - start the voice synthesizer as a child process
**
**	return process id and filedescriptors of input and output pipes
**	via args
*/

int
startsynth(pid,ifd,ofd)
pid_t *pid;
int *ifd,*ofd;
{
	int ret1,ret2,i,argc;
	int pipe0[2],pipe1[2];
	char *p,*q;
	char *argv[NARG];

	/* setup the pipes we need */

	ret2 = -1;
	if ((ret1 = pipe(pipe0)) || (ret2 = pipe(pipe1))) {
	    if(ret1 == 0) {
		close(pipe0[0]);
		close(pipe0[1]);
		if(ret2 == 0) {
		    close(pipe1[0]);
		    close(pipe1[1]);
		}
	    }
	    logmesg("cannot setup pipes for synth process");
	    return FAIL;
	}

	/* start the voice synthesizer process */

	switch (*pid = fork())
	{
	case -1:				/* fork() failure */
	    close(pipe0[0]);
	    close(pipe0[1]);
	    close(pipe1[0]);
	    close(pipe1[1]);
	    logmesg("cannot fork synth process");
	    return FAIL;

	case 0:					/* child process */
	    dup2(pipe0[0],STDIN);
	    dup2(pipe1[1],STDOUT);
	    close(STDERR);
	    open("/dev/null",O_RDWR);		/* STDERR to /dev/null */

	    for (i = 3; i < 20; i++)
		close(i);			/* close remaining files */
	    while (close(i) == 0)
		i++;

	    if ((p = strrchr(Synth,'/')) != NULL) /* get Synth basename */
		p++;
	    else
		p = Synth;

	    argc = 0;				/* build argvector */
	    memset(argv,0,sizeof(argv));

	    argv[argc++] = p;

	    if (SynthFlags != NULL) {		/* add options */
		strcpy(q = ChatBuf,SynthFlags);

		while ((p = strchr(q,' ')) != NULL) {
		    *p++ = '\0';
		    argv[argc++] = q;
		    q = p;
		}

		argv[argc++] = q;
	    }

#ifdef DEBUG
	    if (Dfp != NULL) {
		fclose(Dfp);
		Dfp = NULL;
	    }
#endif	/* DEBUG */

	    for (i = 0; i < NSIG; i++)		/* set all signals to SIG_DFL */
		signal(i, SIG_DFL);

	    nice(5);				/* decrease priority */
	    setgid(UuGid);			/* decrease risk */
	    setuid(UuUid);
	    execv(Synth,argv);
	    logmesg("cannot execute synth process");
	    exit(1);

	default:				/* parent process */
	    close(pipe0[0]);
	    close(pipe1[1]);
	    break;
	}

	*ofd = pipe0[1];			/* return descriptors */
	*ifd = pipe1[0];

	return SUCCESS;
}

/*
**	ausend() - send audio 16-bit data from a file or pipe to the modem
**
**	the data is transcoded from the values read from the pipe into
**	ADPCM as required, and sent to the modem at proper pace.
**
**	the data from the read should be available in units of 16 bits,
**	signed (2-complement) audio data, LSB first, full 16-bit range.
**	a 4K read should never yield an odd number of bytes!
*/

int
ausend(fd,rdf,maxtime)
int fd;
int (*rdf)(int fd,char *buf,size_t size);
int maxtime;
{
	char *msg,*optr;
	register int c;
	int dle,done,enc,inptr,inlen;
	size_t num;
	sigh_t (*oldint)(),(*oldquit)(),(*oldterm)(),(*oldalarm)();
	char ibuf[4096],obuf[BUFSIZ+2];
#define END		65536
#define getpipe()	(inptr >= inlen && (inlen = (*rdf)(fd,ibuf,sizeof(ibuf))) <= (inptr = 0)? END : ((inptr += 2), (int)((unsigned char)ibuf[inptr-2] | (short)((unsigned char)ibuf[inptr-1] << 8))))
#define putres(res)	{ if (res == DLE) *optr++ = DLE; *optr++ = res; }
#define putsound(c)	{ if (SoundFp != NULL) { putc(c,SoundFp); putc((c>>8),SoundFp); } }

	/* work around bug that causes ausend() to hang when CND is ON */

	if ((Fw_features & FW_PLAYBUG) && !PlayCND) {
	    send("ATS40.2=0\r");
	    if (waitfor("OK",MEDMTIME))
		return FAIL;
	    PlayCND = TRUE;
	}

	/* set voice data encoding scheme */
	/* try to get best, fallback to ADPCM4, ADPCM3 when no A-Law/U-Law */

	if ((enc = set_vsm(ENC_BEST)) == FAIL) {
	    MsgBuf[0] = '\0';
	    return FAIL;
	}

	initadpcm();				/* init the adpcm coder */
	sound_mode(ENC_S16,Fsamp);

	/* start playing mode */

	send("AT+VTX\r");
	if (waitfor("CONNECT",MEDMTIME)) {
	    MsgBuf[0] = '\0';
	    return FAIL;
	}

	/* trap signals occurring during sending, so that we can */
	/* reset the MODEM before exiting */

	Signal = 0;
	oldint = signal(SIGINT, sighandler);
	oldquit = signal(SIGQUIT, sighandler);
	oldterm = signal(SIGTERM, sighandler);
	oldalarm = signal(SIGALRM, sighandler);

	if (Fw_features & FW_RTSCTSFLOW)
	    settermios(IMMEDIATE);		/* >= 6.13: hw flow ctl */
	else
	    settermios(XONXOFF);		/* XON/XOFF flow ctl */

	debugtime(D_CHAT);
	debug2(D_CHAT, "ausend() sending audio data\n");

	dle = done = inptr = inlen = 0;		/* initialize buffering */
	*(msg = MsgBuf) = '\0';
	optr = obuf;
	Samples = 0;

	alarm(maxtime);				/* guard the synth process */

	for (;;)
	{
	    /* read audio data from the pipe and send it to modem as ADPCM */
	    /* select packing algorithm (2, 3, 4 or 8 bits per sample) */

	    switch (enc)
	    {
	    case ENC_ADPCM2:
		while (!done && optr < (obuf + sizeof(obuf) - 2))
		{
		    register int res;

		    if ((inlen - inptr) < 8 && (num = optr - obuf) != 0)
			break;

		    if ((c = getpipe()) == END) /* get a sample (buffered) */
			break;			/* quit when no more data */

		    putsound(c);
		    res = toadpcm(ENC_ADPCM2,c) << 6;	/* XX------ */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM2,c) << 4;	/* --XX---- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM2,c) << 2;	/* ----XX-- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM2,c);	/* ------XX */

		    putres(res);
		}
		break;

	    case ENC_ADPCM3:
		while (!done && optr < (obuf + sizeof(obuf) - 6))
		{
		    register int res;

		    if ((inlen - inptr) < 16 && (num = optr - obuf) != 0)
			break;

		    if ((c = getpipe()) == END) /* get a sample (buffered) */
			break;			/* quit when no more data */

		    putsound(c);
		    res = toadpcm(ENC_ADPCM3,c) << 5;	/* XXX----- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM3,c) << 2;	/* ---XXX-- */

		    c = getpipe();
		    putsound(c);
		    res |= (c = toadpcm(ENC_ADPCM3,c)) >> 1; /* ------XX */
		    putres(res);
		    res = (c & 0x01) << 7;		/* X------- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM3,c) << 4;	/* -XXX---- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM3,c) << 1;	/* ----XXX- */

		    c = getpipe();
		    putsound(c);
		    res |= (c = toadpcm(ENC_ADPCM3,c)) >> 2; /* -------X */
		    putres(res);
		    res = (c & 0x03) << 6;		/* XX------ */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM3,c) << 3;	/* --XXX--- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM3,c);	/* -----XXX */

		    putres(res);
		}
		break;

	    case ENC_ADPCM4:
		while (!done && optr < (obuf + sizeof(obuf) - 2))
		{
		    register int res;

		    if ((inlen - inptr) < 4 && (num = optr - obuf) != 0)
			break;

		    if ((c = getpipe()) == END) /* get a sample (buffered) */
			break;			/* quit when no more data */

		    putsound(c);
		    res = toadpcm(ENC_ADPCM4,c) << 4;	/* XXXX---- */

		    c = getpipe();
		    putsound(c);
		    res |= toadpcm(ENC_ADPCM4,c);	/* ----XXXX */

		    putres(res);
		}
		break;

	    case ENC_ALAW:
		while (!done && optr < (obuf + sizeof(obuf) - 2))
		{
		    if ((inlen - inptr) < 2 && (num = optr - obuf) != 0)
			break;

		    if ((c = getpipe()) == END) /* get a sample (buffered) */
			break;			/* quit when no more data */

		    putsound(c);
		    c = linear2alaw(c);		/* convert */

		    if (Fw_features & FW_LINUX)	/* Linux ISDN reverses bits */
			c = rev_bits[c];

		    putres(c);
		}
		break;

	    case ENC_ULAW:
		while (!done && optr < (obuf + sizeof(obuf) - 2))
		{
		    if ((inlen - inptr) < 2 && (num = optr - obuf) != 0)
			break;

		    if ((c = getpipe()) == END) /* get a sample (buffered) */
			break;			/* quit when no more data */

		    putsound(c);
		    c = linear2ulaw(c);		/* convert */
		    putres(c);
		}
		break;
	    }

	    /* flush-out the buffer, send AudioEnd when nothing buffered */
	    /* (this indicates the routines above could not produce more */
	    /* data, and we're done.  set flag to avoid further reading) */

	    if (!done) {
		if (SoundFp != NULL)
		    fflush(SoundFp);

		if ((num = optr - obuf) != 0) {
		    write(STDOUT,obuf,num);
		    optr = obuf;
		} else {
		    if (Fw_features & FW_LINUX)
			usleep(500000);

		    debugtime(D_CHAT);
		    send(AudioEnd);	/* DLE-ETX ends the data */
		    settermios(LONGTIME);
		    done = 1;
		}
	    }

	    /* read the returned characters, can contain DTMF codes etc */

	    while ((c = getch()) != EOF || done)
	    {
		if (dle)		/* previous was a DLE? */
		{
		    switch (c)		/* what is next character? */
		    {
		    case ETX:		/* end of message! */
			break;		/* ignore when sending */

		    case DLE:		/* doubled DLE */
		    case SUB:		/* DLE-SUB (2864) */
			break;

		    case DC4:		/* line hung up (Linux ISDN) */
			c = 'l';	/* treat as DLE-l (line hangup) */
		    default:		/* something else: DTMF or so */
			*msg++ = c;	/* place codes in MsgBuf */
			*msg = '\0';

			/* terminate sending on certain chars */

			if (!done &&
			    (strchr("#bdl",c) != NULL ||
			     msg > (MsgBuf + sizeof(MsgBuf) - 3)))
			{
			    debugtime(D_RUN);
			    debug3(D_RUN, "ausend() terminated by '%c'\n", c);
			    ioctl(STDOUT,TCFLSH,TCOFLUSH);
			    ioctl(STDOUT,TCXONC,TCOON);
			    usleep(20000);
			    send(AudioAbort);
			    settermios(LONGTIME);

			    if (SoundFp != NULL)
				fflush(SoundFp);

			    done = 1;
			}
			break;
		    }

		    dle = 0;
		} else {
		    switch (c)		/* what is next character? */
		    {
		    case DLE:		/* is it a DLE? */
			dle = 1;
			break;

		    case XON:		/* a flow control char leaked thru? */
		    case XOFF:
			break;		/* ignore it */

		    default:		/* something else, probably ^M */
			settermios(INITIAL);

			if (done) {
			    /* this will be the VCON/OK at end */

			    if (c != EOF) {
				InPtr--;	/* push it back */
				debugtime(D_CHAT);
				wait_vcon(MEDMTIME); /* let it finish */
			    }
			} else {
			    /* it came in while we were sending */
			    /* probably DATA/VOICE pressed or so */

			    debugtime(D_RUN);
			    debug3(D_RUN, "ausend() terminated by unexpected char %02x\n", c);
			    InPtr--;		/* push it back */
			    ioctl(STDOUT,TCFLSH,TCOFLUSH);
			    ioctl(STDOUT,TCXONC,TCOON);
			    usleep(20000);
			    send(AudioAbort);
			    waitfor(NULL,LONGTIME); /* let it riddle */
			}

			alarm(0);
			signal(SIGINT, oldint);
			signal(SIGQUIT, oldquit);
			signal(SIGTERM, oldterm);
			signal(SIGALRM, oldalarm);
			Signal = 0;

			if (SoundFp != NULL)
			    fflush(SoundFp);

			if (checksync("ausend") != SUCCESS)
			    return FAIL;

			debugtime(D_RUN);
			debug3(D_RUN, "ausend() returns MsgBuf (%s)\n", MsgBuf);
			return SUCCESS;
		    }
		}
	    }

	    /* see if we have been requested to stop the sending */

	    if (Signal && !done) {
		debugtime(D_RUN);
		debug3(D_RUN, "ausend() terminated by signal %x\n", Signal);
		Signal = 0;
		ioctl(STDOUT,TCFLSH,TCOFLUSH);
		ioctl(STDOUT,TCXONC,TCOON);
		usleep(20000);
		send(AudioAbort);
		settermios(LONGTIME);
		done = 1;
	    }
	}
#undef END
#undef getpipe
#undef putres
}

/*
**	initadpcm() - initialize ADPCM coder/decoder
*/

void
initadpcm()
{
	Delta = 5;
	EstMax = 0;
	Leakage = (Fw_features & FW_NEWLEAK)? 4093 : 3973;
}

/*
**	toadpcm() - ADPCM coder
**
**	this routine takes a 16-bit signed value and returns
**	the n-bit ADPCM code
**
**	unfortunately there are no real specs.	ZyXEL releases a program
**	to convert soundblaster files to ADPCM, and updates this every few
**	ROM releases to adjust some coefficients.  the routine below is
**	patterned after this program (vcnvt.c)
*/

int
toadpcm(enc,edata)
int enc;					/* encoding, ENC_ADPCMx */
register short int edata;			/* value to encode */
{
	int tmpcompbit,databit,signbit;

	edata /= 4;				/* only 14 bits used */
	databit = signbit = 0;

	/* Check for the waveform data and quantize this data */
	if (edata -= EstMax) {
	    tmpcompbit = 1 << enc;
	    /* ----------------------------------------------------- */
	    /* If the data is negative, set flag and change the sign */
	    /* ----------------------------------------------------- */
	    if (edata < 0) {
		edata = -edata;
		signbit = tmpcompbit;
	    }
	    /* --------------------------------------------------- */
	    /* Quantize the waveform data, delta value is adaptive */
	    /* --------------------------------------------------- */
	    while (((edata -= Delta) > 0) && --tmpcompbit)
		databit += 1;
	}

	if (enc == ENC_ADPCM4 && databit == 0)
	    signbit = 8;

	/* the leakage */
	EstMax = (((long)EstMax * Leakage) + 2048) / 4096;

	if ((Delta & 1) && !signbit)
	    EstMax++;

	/* Calculate the Delta and EstMax */
	if (signbit)
	    EstMax -= (Delta >> 1) + (long)Delta * databit;
	else
	    EstMax += (Delta >> 1) + (long)Delta * databit;

	Delta = (((long)Delta * MaxTbl[enc][databit]) + 8192) >> 14;

	return signbit + databit;
}

/*
**      fromadpcm() - ADPCM decoder
**
**	this routine takes an ADPCM sample (2, 3 or 4 bits) and returns
**	the 16-bit signed audio value
**
**	inverse of toadpcm.
*/

short int
fromadpcm(enc,vdata)
int enc;					/* encoding, ENC_ADPCMx */
register char vdata;				/* bits to decode */
{
	int signbit;

	if (enc == ENC_ADPCM4 && !vdata)
	    Delta = 4;

	/* the leakage */
	EstMax = (((long)EstMax * Leakage) + 2048) / 4096;

	signbit = vdata & (1 << (enc & (ENC_ADPCM2|ENC_ADPCM3)));
	vdata &= ~signbit;
	if ((Delta & 1) && !signbit)
	    EstMax++;

	/* Calculate the Delta and EstMax */
	if (signbit)
	    EstMax -= (Delta >> 1) + (long)Delta * vdata;
	else
	    EstMax += (Delta >> 1) + (long)Delta * vdata;

	Delta = (((long)Delta * MaxTbl[enc][(int)vdata]) + 8192) >> 14;

	if (EstMax > 8191)
	    return 32767;

	if (EstMax < -8192)
	    return -32768;

	return 4 * EstMax;
}


/*
**	auread() - read (16 bit) audio data
**
**	this is basically just read(), but when no data is available
**	for an extended period of time, some zeroes are returned...
**
**	the audio player, once started, needs Fsamp samples per second.
**	the amount of data returned from this routine is computed from
**	that (or actually, the timeout on select is varied accordingly).
*/

int
auread(fd,buf,size)
int fd;
char *buf;
size_t size;
{
	int num;
	long usec;
	fd_set select_set;
	struct timeval timeval;

	size &= ~1;				/* return 16-bit words only */

	FD_ZERO(&select_set);			/* all bits off */
	FD_SET(fd,&select_set);			/* set bit for spec'd fd */

	if (Samples == 0)			/* first time? */
	    usec = 100000;			/* wait 100ms for data */
	else
	{
	    gettimeofday(&timeval,NULL);	/* find elapsed time */
	    usec = 1000000 * (timeval.tv_sec - MatchTime.tv_sec) +
		    timeval.tv_usec - MatchTime.tv_usec;

	    usec = (10000 * Samples) / (Fsamp / 100) - usec;
	    usec -= 100000;			/* 100ms safety margin */

	    if (usec < 10000)			/* minimum of 10ms */
		usec = 10000;
	}

	timeval.tv_sec = usec / 1000000;	/* define timeout */
	timeval.tv_usec = usec % 1000000;

	if (select(fd + 1,&select_set,NULL,NULL,&timeval) > 0) {
	    if ((num = read(fd,buf,size)) > 0)
		Samples += num/2;

	    return num;
	}

	if (size > 4096)
	    size = 4096;			/* max bytes to return */

	memset(buf,0,size);
	Samples += size/2;
	return size;
}

/*
**	readulaw() - read ulaw data from file and return as 16-bit
*/

int
readulaw(fd,buf,size)
int fd;
char *buf;
size_t size;
{
	short *p;
	char *q;
	size_t num,n;

	/* read 1/2 of requested data at the end of the buffer */

	n = size / 2;
	num = read(fd, q = buf + size - n, n);
	p = (short *)buf;

	/* convert bytes to shorts */

	while (num-- > 0)
	    *p++ = ulaw2linear(*q++);

	/* return number of converted bytes */

	return (int)((char*)p - buf);
}

/*
**	readulaw8k() - like readulaw() but do dirty conversion from 8k to 9k6
*/

int
readulaw8k(fd,buf,size)
int fd;
char *buf;
size_t size;
{
	short *p;
	char *q;
	size_t num,n;

	/* read 5/12 of requested data at the end of the buffer */

	n = 5 * (size / 12);
	num = read(fd, q = buf + size - n, n);
	p = (short *)buf;

	/* convert bytes, inserting one extra each 5th byte */

	for (n = num; n >= 5; n -= 5) {
	    *p++ = ulaw2linear(*q);
	    *p++ = ulaw2linear((q[0] + q[1]) / 2);
	    q++;
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	}

	while (n-- > 0)
	    *p++ = ulaw2linear(*q++);

	/* return number of converted bytes */

	return (int)((char *)p - buf);
}

/*
**	readulaw9k6() - like readulaw() but do dirty conversion from 9k6 to 8k
*/

int
readulaw9k6(fd,buf,size)
int fd;
char *buf;
size_t size;
{
	short *p;
	char *q;
	size_t num,n;

	/* read 6/10 of requested data at the end of the buffer */

	n = 6 * (size / 10);
	num = read(fd, q = buf + size - n, n);
	p = (short *)buf;

	/* convert bytes, deleting each 6th byte */

	for (n = num; n >= 6; n -= 6) {
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    *p++ = ulaw2linear(*q++);
	    q++;
	}

	while (n-- > 0)
	    *p++ = ulaw2linear(*q++);

	/* return number of converted bytes */

	return (int)((char *)p - buf);
}

/*
 * some conversion routines from SOX
 */

static short seg_end[8] = {0xFF, 0x1FF, 0x3FF, 0x7FF,
			    0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};
#define SIGN_BIT	(0x80)		/* Sign bit for a A-law byte. */
#define QUANT_MASK	(0xf)		/* Quantization field mask. */
#define SEG_SHIFT	(4)		/* Left shift for segment number. */
#define SEG_MASK	(0x70)		/* Segment field mask. */

static int
search(val, table, size)
	int	     val;
	short	   *table;
	int	     size;
{
	int	     i;

	for (i = 0; i < size; i++) {
		if (val <= *table++)
			return (i);
	}
	return (size);
}

/*
 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
 *
 * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
 *
 *      Linear Input Code       Compressed Code
 *     ------------------------	---------------
 *	0000000wxyza		000wxyz
 *	0000001wxyza		001wxyz
 *	000001wxyzab		010wxyz
 *	00001wxyzabc		011wxyz
 *	0001wxyzabcd		100wxyz
 *	001wxyzabcde		101wxyz
 *	01wxyzabcdef		110wxyz
 *	1wxyzabcdefg		111wxyz
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */
unsigned char
linear2alaw(pcm_val)
	int		pcm_val;	/* 2's complement (16-bit range) */
{
	int		mask;
	int		seg;
	unsigned char   aval;

	if (pcm_val >= 0) {
		mask = 0xD5;		/* sign (7th) bit = 1 */
	} else {
		mask = 0x55;		/* sign bit = 0 */
		pcm_val = -pcm_val - 8;
	}

	/* Convert the scaled magnitude to segment number. */
	seg = search(pcm_val, seg_end, 8);

	/* Combine the sign, segment, and quantization bits. */

	if (seg >= 8)		/* out of range, return maximum value. */
		return (0x7F ^ mask);
	else {
		aval = seg << SEG_SHIFT;
		if (seg < 2)
			aval |= (pcm_val >> 4) & QUANT_MASK;
		else
			aval |= (pcm_val >> (seg + 3)) & QUANT_MASK;
		return (aval ^ mask);
	}
}

/*
 * alaw2linear() - Convert an A-law value to 16-bit linear PCM
 *
 */
int
alaw2linear(a_val)
	unsigned char   a_val;
{
	int		t;
	int		seg;

	a_val ^= 0x55;

	t = (a_val & QUANT_MASK) << 4;
	seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
	switch (seg) {
	case 0:
		t += 8;
		break;
	case 1:
		t += 0x108;
		break;
	default:
		t += 0x108;
		t <<= seg - 1;
	}
	return ((a_val & SIGN_BIT) ? t : -t);
}

/*
 * linear2ulaw() - Convert a linear PCM value to u-law
 *
 * In order to simplify the encoding process, the original linear magnitude
 * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
 * (33 - 8191). The result can be seen in the following encoding table:
 *
 *      Biased Linear Input Code	Compressed Code
 *      ------------------------	---------------
 *	00000001wxyza			000wxyz
 *	0000001wxyzab			001wxyz
 *	000001wxyzabc			010wxyz
 *	00001wxyzabcd			011wxyz
 *	0001wxyzabcde			100wxyz
 *	001wxyzabcdef			101wxyz
 *	01wxyzabcdefg			110wxyz
 *	1wxyzabcdefgh			111wxyz
 *
 * Each biased linear code has a leading 1 which identifies the segment
 * number. The value of the segment number is equal to 7 minus the number
 * of leading 0's. The quantization interval is directly available as the
 * four bits wxyz.  * The trailing bits (a - h) are ignored.
 *
 * Ordinarily the complement of the resulting code word is used for
 * transmission, and so the code word is complemented before it is returned.
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */
#define BIAS		(0x84)	  /* Bias for linear code. */
unsigned char
linear2ulaw(pcm_val)
	int		pcm_val;	/* 2's complement (16-bit range) */
{
	int		mask;
	int		seg;
	unsigned char   uval;

	/* Get the sign and the magnitude of the value. */
	if (pcm_val < 0) {
		pcm_val = BIAS - pcm_val;
		mask = 0x7F;
	} else {
		pcm_val += BIAS;
		mask = 0xFF;
	}

	/* Convert the scaled magnitude to segment number. */
	seg = search(pcm_val, seg_end, 8);

	/*
	 * Combine the sign, segment, quantization bits;
	 * and complement the code word.
	 */
	if (seg >= 8)		/* out of range, return maximum value. */
		return (0x7F ^ mask);
	else {
		uval = (seg << 4) | ((pcm_val >> (seg + 3)) & 0xF);
		return (uval ^ mask);
	}
}

/*
 * ulaw2linear() - Convert a u-law value to 16-bit linear PCM
 *
 * First, a biased linear code is derived from the code word. An unbiased
 * output can then be obtained by subtracting 33 from the biased code.
 *
 * Note that this function expects to be passed the complement of the
 * original code word. This is in keeping with ISDN conventions.
 */
int
ulaw2linear(u_val)
	unsigned char	u_val;
{
	int		t;

	/* Complement to obtain normal u-law value. */
	u_val = ~u_val;

	/*
	 * Extract and bias the quantization bits. Then
	 * shift up by the segment number and subtract out the bias.
	 */
	t = ((u_val & QUANT_MASK) << 3) + BIAS;
	t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;

	return ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));
}


/*
**	set_voice() - set modem to VOICE mode and init params
**
*/

int
set_voice(ls)
int ls;						/* desired device */
{
	char buf[80];

	sprintf(buf,"AT%s+VIP",make_fclass(8));	/* select VOICE mode & reset */
	CurVSD[0] = '\0';			/* VSD setting gets reset */
	CurLS = CurVSM = 0;			/* VLS/VSM setting as well */
	CurSync = -1;

	if (Fw_features & FW_RTSCTSFLOW)	/* re-select flow control */
	    strcat(buf,"+FLO=2");

	sprintf(buf + strlen(buf),"+VSD=%s",Silence);

	if (Is_ZyXEL) {
	    if (Fw_features & FW_2864)
		strcat(buf,"+VRN=6+VTD=10");
	    else {
		strcat(buf,"+VRN=57+VBT=10");

		if (Fw_features & FW_VBTI)
		    strcat(buf,",10");
	    }

	    if ((Fw_features & FW_DTMFDET) && DtmfDet != NULL) {
		if (Fw_features & (FW_2864|FW_LINUX))
		    sprintf(buf + strlen(buf),"+VDD=%s",DtmfDet);
		else {
		    char *comma;

		    if ((comma = strchr(DtmfDet,',')) != NULL)
			*comma = '\0';

		    if (DtmfDet[0])
			sprintf(buf + strlen(buf),"+VDH=%s",DtmfDet);

		    if (comma != NULL && comma[1])
			sprintf(buf + strlen(buf),"+VDD=%s",comma + 1);
		}
	    }
	}

	strcat(buf,make_vls(ls));		/* select speaker */
	strcat(buf,"\r");
	send(buf);

	if (wait_vcon(MEDMTIME) == FAIL)	/* sometimes OK, sometimes VCON */
	    return FAIL;

	CurLS = ls;
	strcpy(CurVSD,Silence);
	return SUCCESS;
}

/*
**	set_vsm() - set audio encoding scheme
**
**	Input is the encoding scheme as used in ZFAX audio files.
**	This is not the same as the AT+VSM value sent to the modem!
*/

int
set_vsm(enc)
int enc;					/* desired encoding */
{
	int sync = 0;
	int vsm;
	char buf[20];

	/* check the encoding value.  see if the modem supports it. */

	switch (enc)
	{
	case ENC_CELP:
	    if (!(Fw_features & FW_PLUS))	/* not a PLUS, cannot record CELP */
		enc = ENC_ADPCM2;		/* do ADPCM2 instead... */

	    vsm = enc + 1;			/* VSM value is one higher */
	    break;

	case ENC_ADPCM2|ENC_RESYNC:
	case ENC_ADPCM3|ENC_RESYNC:
	    if (Fw_features & FW_RESYNC)	/* resync available? */
		sync = 1;
	    else
		enc -= ENC_RESYNC;

	    /* fall-through */

	case ENC_ADPCM2:
	case ENC_ADPCM3:
	    vsm = (enc & ~ENC_RESYNC) + 1;
	    break;				/* supported by all modems */

	case ENC_ADPCM4|ENC_RESYNC:
	    if (Fw_features & FW_RESYNC)	/* resync available? */
		sync = 1;
	    else
		enc -= ENC_RESYNC;

	    /* fall-through */

	case ENC_ADPCM4:
adpcm4:	    if (!(Fw_features & FW_NEWADPCM) || /* ADPCM4 available? */
		Speed < 57600)			/* and 57600 bps */
		enc -= (ENC_ADPCM4-ENC_ADPCM3); /* else do ADPCM3 instead... */

	    vsm = (enc & ~ENC_RESYNC) + 1;
	    break;

	case ENC_ADPCM3N|ENC_RESYNC:		/* new ADPCM3 + resync */
	    if (Fw_features & FW_RESYNC)	/* resync available? */
		sync = 1;
	    else
		enc -= ENC_RESYNC;

	    /* fall-through */

	case ENC_ADPCM3N:			/* new ADPCM3 */
	    if (!(Fw_features & FW_NEWADPCM)) { /* ADPCM3N available? */
		enc -= (ENC_ADPCM3N-ENC_ADPCM3); /* do ADPCM3 instead... */
		vsm = (enc & ~ENC_RESYNC) + 1;
		break;
	    }

	    vsm = 30;
	    break;

	case ENC_ALAW:				/* A-Law (Linux ISDN) */
	case ENC_ULAW:				/* U-Law (Linux ISDN) */
	    if (!(Fw_features & FW_XLAW)) {
		enc = ENC_ADPCM4;		/* use ADPCM4 if not avail */
		goto adpcm4;
	    }

	    vsm = (enc - ENC_ALAW) + 5;
	    break;

	default:
	    return FAIL;			/* unsupported value */
	}

	if (sync == CurSync && vsm == CurVSM)	/* encoding already OK? */
	    return enc;

	strcpy(buf,"AT");
	if (Fw_features & FW_RESYNC)		/* resync available? */
	    sprintf(buf + 2,"S39.7=%d",sync);
	sprintf(buf + strlen(buf),"+VSM=%d\r",vsm);
	send(buf);

	if (waitfor("OK",MEDMTIME))
	    return FAIL;

	CurVSM = vsm;				/* save the set encoding */
	CurSync = sync;

	return enc;				/* return modified encoding */
}

/*
**	set_vls() - set audio recording/playback device
**
*/

int
set_vls(ls)
int ls;						/* desired device */
{
	char buf[16];

	sprintf(buf,"AT%s\r",make_vls(ls));
	send(buf);

	if (wait_vcon(MEDMTIME) == FAIL)	/* sometimes OK, sometimes VCON */
	    return FAIL;

	CurLS = ls;
	return SUCCESS;
}

/*
**	make_vls() - return +VLS cmd to set audio recording/playback device
**
*/

char *
make_vls(ls)
int ls;
{
	static char buf[10];

	sprintf(buf,"+VLS=%d",ls);
	return buf;
}

/*
**	make_fclass() - return +FCLASS cmd to set the FCLASS
**
**	returns the proper +FCLASS=n command when modem not yet in the
**	requested FCLASS, or empty string if it already is.
**
*/

char *
make_fclass(fclass)
int fclass;
{
	static char buf[16];

	if (fclass >= 0 && fclass != Fclass) {
	    sprintf(buf,"+FCLASS=%d",fclass);
	    Fclass = fclass;
	} else
	    buf[0] = '\0';

	return buf;
}

/*
**	waitvcon() - wait for VCON or OK response (CR/LF terminated)
**
**	The U-1496 usually sends VCON in voice mode, but sometimes OK
**	The 2864 series always sends OK
**
**	This routine returns FAIL when the response wasn't VCON or OK.
**	Don't test for OK or zero, the returnvalue for VCON is different.
*/

int
wait_vcon(state)
int state;
{
	int retry = 3;

	/* some retries are made, because the modem could send some */
	/* garbage like ^P^C^M^J before the expected response, which */
	/* would be returned as a response by waitfor(NULL,state). */
	/* then, FAIL would be returned and the real response (OK) */
	/* would remain in the input buffer and would be misinterpreted */
	/* as a response to the next command.... */

	while (retry--) {
	    if (waitfor(NULL,state))		/* sometimes OK, sometimes VCON */
		return FAIL;

	    rip(Match);

	    if (!strcmp(Match,"OK"))
		return SUCCESS;

	    if (!strcmp(Match,"VCON"))
		return VCON;

	    if (!strcmp(Match,"NO CARRIER") || !strcmp(Match,"NO ANSWER"))
		return NOCARRIER;

	    if (!strcmp(Match,"ERROR"))
		break;
	}

	checksync("wait_vcon");
	return FAIL;				/* unexpected response */
}

/*
**	checksync() - make sure we are still in sync with the modem
*/

int
checksync(where)
char *where;					/* caller (for logging) */

{
	send("AT\r");				/* quick check */

	if (waitfor("OK",MEDMTIME)) {
	    long arg;

	    debugtime(D_RUN);
	    debug3(D_RUN, "%s() resyncing\n", where);

	    arg = TIOCM_RTS;			/* RTS OFF */
	    ioctl(STDIN,TIOCMBIC,&arg);
	    usleep(20000);
	    arg = TIOCM_RTS;			/* RTS ON */
	    ioctl(STDIN,TIOCMBIS,&arg);

	    send("\\003\\020\\003\\021\r\\pAT\r"); /* ETX DLE ETX XON \r pause AT \r */

	    if (waitfor("OK",MEDMTIME)) {
		debugtime(D_RUN);
		debug4(D_RUN, "%s() lost MODEM (%s)\n", where, MsgBuf);
#ifdef DEBUG
		linestat();
#endif
		return FAIL;
	    }

	    waitfor("FLUSH",SHORTTIME);
	}

	return SUCCESS;
}


/*************************************************************************
**	Misc utility functions						**
*************************************************************************/

/*
**	dialup() - dial a number and wait for connection
**
**	will return RING when a call comes in while trying to dial
*/

int
dialup(mode,fclass,timeout,number)
const char *mode;
int fclass;					/* which +FCLASS to use */
int timeout;					/* #tries to wait for connect */
const char *number;

{
	const char *tomatch;
	char *savematch;
#ifdef TIOCMGET
	int mstat;
#endif

	sprintf(MsgBuf, "Dialing FCLASS=%d (%s)", fclass, number);
	logmesg(MsgBuf);
	debugtime(D_RUN);
	debug3(D_RUN, "dialup() - %s\n", MsgBuf);

#ifdef TIOCMGET
	if ((mstat = modemstat()) != -1 && (mstat & TIOCM_RNG))
	    return RING;
#endif

	while (waitfor(NULL,SHORTTIME) == SUCCESS)
	    if (!strncmp(Match,"RING",4))
		return RING;

	if (Signal)
	    return FAIL;

	/* put MODEM in the correct mode and FCLASS
	 */
	sprintf(ChatBuf,"AT%s%s",mode,make_fclass(fclass));

	if ((Fw_features & FW_LINUX) && fclass == 8)
	    strcat(ChatBuf,"S18=1");		/* force audio dialout */

	strcat(ChatBuf,"\r");
	send(ChatBuf);
	CurLS = 0;

	if (waitfor(NULL,MEDMTIME) || Signal)
	    return FAIL;

	rip(Match);

	if (!strncmp(Match,"RING",4))
	    return RING;

	if (strcmp(Match,"OK") && !(Fw_features & FW_FAXBUG))
	    return FAIL;

#ifdef TIOCMGET
	if (mstat != -1 && (mstat = modemstat()) != -1 && (mstat & TIOCM_RNG))
	    return RING;
#endif

	/* now dial the number...
	 */
	sprintf(ChatBuf,"AT%s%sD%s%s\r",
			(fclass > 0? "" : Mute),
			(fclass == 8? make_vls(PhoneLS) : ""),
			(DialPrefix != NULL? DialPrefix : ""),
			number);
	send(ChatBuf);

	switch (fclass)
	{
#ifndef ZFAX
	case 2:					/* CLASS2 FAX, expect +FCON */
	    tomatch = "+FCON";
	    break;
#endif
	case 8:					/* VOICE class, expect VCON */
	    if (Fw_features & FW_2864)
		tomatch = "OK";			/* ... except on 2864 */
	    else
		tomatch = "VCON";
	    break;

	case -2:				/* don't care, dialling only */
	    tomatch = "OK";			/* ATDxxx; results in OK */
	    break;

	default:				/* -1, 0, 6 */
	    tomatch = "CONNECT";
	    break;
	}

	/* now wait for connection, may take some time
	 */
	timeout += DialTime;			/* optional extra timeout */

	while (timeout-- && !Signal)
	    if (!waitfor(NULL,LONGTIME)) {
		rip(Match);

		if (!strncmp(Match,tomatch,strlen(tomatch))) /* prefix match only */
		    break;

		/* got something else than expected */

		logmesg(Match);			/* log what we got (NO CARRIER?) */
		debugtime(D_RUN);
		debug3(D_RUN, "dialup() (%s) - FAIL\n", Match);

		savematch = strdup(Match);	/* a (small) memory leak */
		send("\r\\pATH\r");
		waitfor("OK",MEDMTIME);
		Match = savematch;
		CurLS = Fclass = 0;

		if (!strncmp(Match,"RING",4))
		    return RING;

#ifdef TIOCMGET
		if (mstat != -1 && (mstat = modemstat()) != -1 && (mstat & TIOCM_RNG))
		    return RING;
#endif

		return FAIL;
	    }

	if (timeout < 0 || Signal) {		/* nothing received */
	    send("\r\\pATH\r");
	    waitfor("OK",MEDMTIME);
	    Match = NULL;
	    CurLS = Fclass = 0;
	    return FAIL;
	}

	if (fclass == 8)
	    CurLS = 2;

	if (fclass != 2)			/* +FCON not so interesting */
	    logmesg(Match);			/* log the CONNECT we got! */

	free(Connect);
	Connect = strdup(Match);		/* and save it */

	debugtime(D_RUN);
	debug3(D_RUN, "dialup() (%s) - SUCCESS\n", Match);
	return SUCCESS;
}


/*
**	push() - start a process and 'push' it onto an fd
**
**	STDIN of the childprocess will be the fd passed as a parameter, and
**	STDOUT of the childprocess will replace that fd. thus, the childprocess
**	is effectively 'pushed' onto the fd, providing a filter function on
**	the file data.
**
**	The PID of the process is returned via parameter 1, as the caller
**	should waitpid() on it.
**
*/

int
push(pid,fd,program,argv)
pid_t *pid;					/* pid returned here */
int *fd;					/* is input file, becomes pipe */
char *program;					/* program name (full pathname) */
char *argv[];					/* arguments (ends in NULL) */

{
	int i;
	int pipe1[2];
	pid_t fpid;
	char *p;

	/* get program basename */

	if ((p = strrchr(program,'/')) != NULL)
	    argv[0] = p + 1;
	else
	    argv[0] = program;

	/* setup the pipe we need */

	if (pipe(pipe1) != 0) {
	    logmesg("cannot setup pipe");
	    return FAIL;
	}

	switch (fpid = fork())
	{
	case -1:				/* fork() failure */
	    close(pipe1[0]);
	    close(pipe1[1]);
	    debug2(D_RUN, "push(): cannot fork\n");
	    logmesg("cannot fork");
	    return FAIL;

	case 0:					/* child process */
	    dup2(*fd,STDIN);			/* input file */
	    dup2(pipe1[1],STDOUT);
	    close(STDERR);
	    open("/dev/null",O_RDWR);		/* STDERR to /dev/null */

	    for (i = 3; i < 20; i++)
		close(i);			/* close remaining files */
	    while (close(i) == 0)
		i++;

#ifdef DEBUG
	    if (Dfp != NULL) {
		fclose(Dfp);
		Dfp = NULL;
	    }
#endif	/* DEBUG */

	    for (i = 0; i < NSIG; i++)		/* set all signals to SIG_DFL */
		signal(i, SIG_DFL);

	    nice(5);				/* decrease priority */
	    setgid(UuGid);			/* decrease risk */
	    setuid(UuUid);
	    execv(program,argv);
	    sprintf(MsgBuf,"cannot execute %s",program);
	    debug3(D_RUN, "push(): %s\n",MsgBuf);
	    logmesg(MsgBuf);
	    exit(1);

	default:				/* parent process */
	    close(pipe1[1]);
	    close(*fd);
	    *fd = pipe1[0];
	    *pid = fpid;
	    break;
	}

	return SUCCESS;
}

/*
**	Fputs() - does fputs() with '\' and '@' expansion
**
**	Returns EOF if an error occurs.
*/

int
Fputs(s,fd)
register const char *s;
int fd;
{
	char c, n, buf[32];
	time_t tim;

	while ((c = *s++) != '\0')
	{
	    if ((c == '@') && (n = *s++))
	    {
		switch (n)
		{
		case 'B':	/* speed (baud rate) */
			sprintf(buf, "%d", Speed);
			if (Fputs(buf,fd) == EOF)
			    return EOF;
			break;
		case 'C':	/* last received CONNECT */
			if (Connect && Fputs(Connect,fd) == EOF)
			    return EOF;
			break;
		case 'D':	/* date */
			tim = time(NULL);
			strftime(buf,sizeof(buf),"%d %B %Y",localtime(&tim));
			if (Fputs(buf,fd) == EOF)
			    return EOF;
			break;
		case 'I':	/* calling number */
			if (CndNumber && Fputs(CndNumber,fd) == EOF)
			    return EOF;
			break;
		case 'J':	/* calling name */
			if (CndName && Fputs(CndName,fd) == EOF)
			    return EOF;
			break;
		case 'K':	/* calling number including fixed text */
			if (CndNumber &&
			    (Fputs("Calling number: ",fd) == EOF ||
			     Fputs(CndNumber,fd) == EOF))
			    return EOF;
			break;
		case 'L':	/* line */
			if (Device && Fputs(Device,fd) == EOF)
			    return EOF;
			break;
		case 'M':	/* modem make & version */
			if (Fputs(Modem,fd) == EOF)
			    return EOF;
			break;
		case 'N':	/* user-entered Number */
			if (Fputs(Number,fd) == EOF)
			    return EOF;
			break;
		case 'S':	/* system node name */
			if (SysName && Fputs(SysName,fd) == EOF)
			    return EOF;
			break;
		case 'T':	/* time */
			tim = time(NULL);
			strftime(buf,sizeof(buf),"%X",localtime(&tim));
			if (Fputs(buf,fd) == EOF)
			    return EOF;
			break;
		case 'U':	/* number of active users */
			sprintf(buf, "%d", Nusers);
			if (Fputs(buf,fd) == EOF)
			    return EOF;
			break;
		case 'V':	/* version */
			if (Version && Fputs(Version,fd) == EOF)
			    return EOF;
			break;
		case '@':	/* in case '@@' was used */
			if (write(fd, &n, 1) != 1)
			    return EOF;
			break;
		}
	    } else {
		switch (c)
		{
		case '\\':
		    s = unquote(s, &c);
		case '\n':
		    /* we're in raw mode: send CR before every LF
		     */
		    if (c == '\n') {
			if (write(fd, "\r\n", 2) != 2)
			    return EOF;
		    } else {
			if (write(fd, &c, 1) != 1)
			    return EOF;
		    }
		    break;

		default:
		  {
		    const char *p = s - 1;
		    size_t len;

		    /* try to collect some plain chars in single write */
		    while ((c = *s) != '\0' && c != '\n' &&
			    c != '@' && c != '\\')
			s++;
		    len = s - p;
		    if (write(fd, p, len) != len)
			return EOF;
		  }
		}
	    }
	}
	return SUCCESS;
}


/*
**	getuname() - retrieve the system's node name
**
**	Returns pointer to name or a zero-length string if not found.
*/

char *
getuname()
{
#ifdef DOUNAME
	struct utsname uts;
#endif	/* DOUNAME */
	static char name[80];

	name[0] = '\0';

#ifdef DOUNAME				/* dig it out of the kernel */
	if (uname(&uts) != FAIL)
		strcpy(name, uts.nodename);
#endif	/* DOUNAME */

#ifdef PHOSTNAME			/* get it from the shell */

	if (strlen(name) == 0) {
		FILE *cmd;
		if ((cmd = popen(PHOSTNAME, "r")) != NULL) {
			fgets(name, sizeof(name), cmd);
			pclose(cmd);
			name[strlen(name)-1] = '\0';
		}
	}

#endif	/* PHOSTNAME */

	return name;
}


/*
**	filespresent() - checks if files are present in a given dir
*/
int
filespresent(dirname)
const char *dirname;
{
	DIR *dirp;
	struct dirent *de;
	int found = 0;

	if ((dirp = opendir(dirname)) == NULL)
	    return 0;			/* no dir, so no files */

	while ((de = readdir(dirp)) != 0)
	    if (de->d_name[0] != '.')
	    {
		found++;		/* found a non-hidden file! */
		break;			/* might as well quit */
	    }

	closedir(dirp);
	return found;
}

/*
**	rip() - remove trailing CR and spaces from a string
**
**	returns length of remaining string
*/
int
rip(string)
char *string;
{
	char *p;

	if ((p = strchr(string,'\r')) != NULL)
	    *p = '\0';
	else
	    p = string + strlen(string);

	while (--p >= string && *p == ' ')
	    *p = '\0';

	return (int)(++p - string);
}

/*
**	settermios() - setup tty termios values (depending on state)
*/

void
settermios(state)
int state;
{
#ifdef __linux__
	int divisor;
#endif

	if (state != INITIALIZE && state == Tstate)
		return;

	switch (state) {
	case INITIALIZE:			/* initialize everything */
		memset(&Termios,0,sizeof(Termios));
		ioctl(STDIN, TCGETS, &Termios);
#ifdef __linux__
		memset(&Serial,0,sizeof(Serial));
		if (ioctl(STDIN, TIOCGSERIAL, &Serial) < 0)
		    Serial.type = 0;
		else
		    Serial.flags &= ~ASYNC_SPD_MASK;
#endif

		switch (Speed)
		{
		case 9600:
			Cbaud = B9600;
			break;
		case 19200:
			Cbaud = B19200;
			break;
		case 38400:
			Cbaud = B38400;
			break;
#ifdef B57600
		case 57600:
			Cbaud = B57600;
			break;
#else
# if defined(__linux__) && defined(ASYNC_SPD_HI)
		case 57600:
			if (Serial.type != 0) {
			    Serial.flags |= ASYNC_SPD_HI;
			    Cbaud = B38400;
			    break;
			}
# endif
#endif
#ifdef B76800
		case 76800:
			Cbaud = B76800;
			break;
#endif
#ifdef B115200
		case 115200:
			Cbaud = B115200;
			break;
#else
# if defined(__linux__) && defined(ASYNC_SPD_HI)
		case 115200:
			if (Serial.type != 0) {
			    Serial.flags |= ASYNC_SPD_VHI;
			    Cbaud = B38400;
			    break;
			}
# endif
#endif
#ifdef B230400
		case 230400:
			Cbaud = B230400;
			break;
#endif
#ifdef B460800
		case 460800:
			Cbaud = B460800;
			break;
#endif
		default:
#ifdef __linux__
			if (Serial.type != 0) {
			    divisor = Serial.baud_base / Speed;
			    if ((divisor * Speed) == Serial.baud_base) {
				Serial.flags |= ASYNC_SPD_CUST;
				Serial.custom_divisor = divisor;
				Cbaud = B38400;
				break;
			    }
			}
#endif
			sprintf(MsgBuf, "Invalid speed (%d)",Speed);
			logmesg(MsgBuf);
			Cbaud = B9600;
			break;
		}

#ifdef __linux__
		if (Serial.type != 0)
		    ioctl(STDIN,TIOCSSERIAL,&Serial);
#endif

		/* initial settings - reasonable to talk to the MODEM
		 */
		Termios.c_iflag = BRKINT | INPCK;
		Termios.c_oflag = 0;
		Termios.c_cflag = Cbaud | CS8 | CREAD | HUPCL | CLOCAL | CRTSCTS;
		Termios.c_lflag = 0;
#ifdef N_TTY
		Termios.c_line = N_TTY;
#endif

		/* set c_cc[] chars to reasonable values
		 */
		Termios.c_cc[VINTR] = CINTR;
		Termios.c_cc[VQUIT] = CQUIT;
		Termios.c_cc[VERASE] = CERASE;
		Termios.c_cc[VKILL] = CKILL;
		Termios.c_cc[VEOF] = CEOF;
		Termios.c_cc[VSTART] = CSTART;
		Termios.c_cc[VSTOP] = CSTOP;
		Termios.c_cc[VSUSP] = CSUSP;
		Termios.c_cc[VEOL] = CEOL;
		Termios.c_cc[VREPRINT] = CRPRNT;
		Termios.c_cc[VWERASE] = CWERASE;
		Termios.c_cc[VLNEXT] = CLNEXT;

		Termios.c_cc[VTIME] = 0;	/* block read for 1 char */
		Termios.c_cc[VMIN] = 1;

		ioctl(STDIN, TCSETSF, &Termios);
		return;

	case INITIAL:				/* back to initial state */
		Termios.c_iflag = BRKINT | INPCK;
		Termios.c_oflag = 0;
		Termios.c_cflag = Cbaud | CS8 | CREAD | HUPCL | CLOCAL | CRTSCTS;
		Termios.c_lflag = 0;

		Termios.c_cc[VTIME] = 0;
		Termios.c_cc[VMIN] = 1;
		break;

	case DROPDTR:				/* lower DTR line */
		Termios.c_cflag &= ~CBAUD;
#ifdef CBAUDEX
		Termios.c_cflag &= ~CBAUDEX;
#endif
		break;

	case SHORTTIME:				/* read() returns quickly */
		if (Termios.c_cc[VTIME] == 2 && Termios.c_cc[VMIN] == 0)
		    return;			/* already set! */

		Termios.c_cc[VTIME] = 2;	/* 200ms */
		Termios.c_cc[VMIN] = 0;
		break;

	case SHORT255:				/* read() returns quickly */
		Termios.c_cc[VTIME] = 2;	/* 200ms */
#ifndef BROKEN_TERMIOSV
		Termios.c_cc[VMIN] = 255;	/* prefer to get some chars */
#else
		Termios.c_cc[VMIN] = 127;	/* only for broken systems */
#endif
		break;

	case MEDMTIME:				/* read() returns medium */
		if (Termios.c_cc[VTIME] == (Is_ZyXEL? 10 : 25) &&
		    Termios.c_cc[VMIN] == 0)
		    return;			/* already set! */

		Termios.c_cc[VTIME] = Is_ZyXEL? 10 : 25; /* 1 or 2.5 second */
		Termios.c_cc[VMIN] = 0;
		break;

	case LONGTIME:				/* read() returns slowly */
		if (Termios.c_cc[VTIME] == 150 && Termios.c_cc[VMIN] == 0)
		    return;			/* already set! */

#ifndef BROKEN_TERMIOSV
		Termios.c_cc[VTIME] = 150;	/* 15 seconds */
#else
		Termios.c_cc[VTIME] = 127;	/* 12.7 seconds for broken systems */
#endif
		Termios.c_cc[VMIN] = 0;
		break;

	case XONXOFF:				/* XON/XOFF handshake */
		Termios.c_iflag |= IXON | IXOFF;
		Termios.c_cflag &= ~CRTSCTS;
	case IMMEDIATE:				/* immediate return on read */
		Termios.c_cc[VTIME] = 0;
		Termios.c_cc[VMIN] = 0;
		break;

	case WATCHDCD:				/* SIGHUP when DCD lost */
		Termios.c_cflag &= ~CLOCAL;
		break;

	case SEVEN_E:				/* 7-bit even parity */
		Termios.c_iflag |= ISTRIP;
		Termios.c_cflag &= ~CS8;
		Termios.c_cflag |= PARENB | CS7;
		break;

	case FINAL:				/* after login */
		Termios.c_iflag &= ~(BRKINT | INPCK);
		if (FinalXoff)
		    Termios.c_iflag |= IXON | IXANY;
		Termios.c_oflag |= OPOST | TAB3;
		Termios.c_cflag &= ~CLOCAL;
		Termios.c_lflag |= ISIG | ICANON | ECHO | ECHOE | ECHOK;
#if VEOF == VMIN
		Termios.c_cc[VEOF] = CEOF;
#endif
#if VEOL == VTIME
		Termios.c_cc[VEOL] = CEOL;
#endif
		break;
	}

	ioctl(STDIN, TCSETSW, &Termios);	/* wait until sent, then set */
	Tstate = state;
}


#ifdef TIOCMGET
/*
**	modemstat() - return status of MODEM handshake lines
*/

int
modemstat()

{
	unsigned int mstat;

	errno = 0;

	if (ioctl(0, TIOCMGET, &mstat) < 0) {	/* read MODEM line status */
	    mstat = -1;				/* failed, assume not supported */
	    debug3(D_CHAT, "modemstat(): TIOCMGET failed, errno=%d\n", errno);
	} else {
#ifdef DEBUG
	    char buf[80];

	    buf[0] = '\0';

	    if (mstat & TIOCM_DSR)
		strcat(buf," DSR");
	    if (mstat & TIOCM_RNG)
		strcat(buf," RNG");
	    if (mstat & TIOCM_CAR)
		strcat(buf," CAR");
	    if (mstat & TIOCM_CTS)
		strcat(buf," CTS");
	    if (mstat & TIOCM_RTS)
		strcat(buf," RTS");
	    if (mstat & TIOCM_DTR)
		strcat(buf," DTR");
	    if (mstat & TIOCM_SR)
		strcat(buf," SR");
	    if (mstat & TIOCM_ST)
		strcat(buf," ST");
	    if (mstat & TIOCM_LE)
		strcat(buf," LE");

	    debugtime(D_CHAT);
	    debug5(D_CHAT, "modemstat(): %04o 0x%03x%s\n", mstat, mstat, buf);
#endif
	}

	return mstat;
}
#endif

#ifdef DEBUG
/*
**	linestat() - print some line status for MODEM problem debugging
*/

void
linestat()

{
	long arg;

# ifdef TIOCMGET
	(void)modemstat();
# endif
# ifdef TIOCOUTQ
	if (!ioctl(1, TIOCOUTQ, &arg) && arg != 0) {
	    debug3(D_CHAT, "output queue: %d\n", arg);
	}
# endif
# ifdef TIOCINQ
	if (!ioctl(0, TIOCINQ, &arg) && arg != 0) {
	    debug3(D_CHAT, "input queue: %d\n", arg);
	}
# endif
}
#endif


/*
**	huphandler() - handles SIGHUP when enabled
*/

sigh_t
huphandler(sig)
int sig;
{
	debugtime(D_RUN);
	debug2(D_RUN, "SIGHUP caught\n");
	logmesg("CARRIER lost");
	send("ATZ\r");
	settermios(DROPDTR);
	exit(EXIT_FAILURE);
}

/*
**	inthandler() - handles SIGINT when not trapped elswhere
*/

sigh_t
inthandler(sig)
int sig;
{
	debugtime(D_RUN);
	debug2(D_RUN, "SIGINT caught\n");
	logmesg("BREAK received");
	send("ATZ\r");
	settermios(DROPDTR);
	exit(EXIT_FAILURE);
}

/*
**	sighandler() - handles misc signals when required
*/

sigh_t
sighandler(sig)
int sig;
{
	if (sig > 0 && sig < NSIG)
	    signal(sig, sighandler);
	else
	    sig = 0;

	Signal |= 1 << sig;		/* save the signal number as a flag */
}


/*************************************************************************
**	Lockfile handling for sharing of single port with UUCP		**
*************************************************************************/

/*
**	makelock() - attempt to create a lockfile
**
**	Returns FAIL if lock could not be made (line in use).
*/

int
makelock(name)
char *name;
{
	int fd, pid;
	char *temp, buf[MAXLINE+1];
#ifdef ASCIIPID
	char apid[16];
#endif	/* ASCIIPID */

	debug3(D_LOCK, "makelock(%s) called\n", name);

	/* first make a temp file
	 */
	strcpy(buf, Lock);
	if ((temp = strrchr(buf, '/')) != NULL)
		strcpy(temp + 1, "TM.XXXXXX");

	if ((fd = creat((temp = mktemp(buf)), 0444)) == FAIL) {
		sprintf(MsgBuf, "cannot create tempfile (%s)", temp);
		logmesg(MsgBuf);
		return FAIL;
	}
	debug3(D_LOCK, "temp = (%s)\n", temp);

	/* owner uucp, to prevent possible trouble when the program
	 * crashes without removing the lockfile
	 */
	chown(temp, UuUid, UuGid);

	/* put my pid in it
	 */
#ifdef ASCIIPID
	sprintf(apid, "%10d\n", (int)getpid());
	write(fd, apid, strlen(apid));
#else
	pid = (int)getpid();
	write(fd, (char *)&pid, sizeof(pid));
#endif	/* ASCIIPID */
	close(fd);

	/* link it to the lock file
	 */
#ifdef DEBUG
	errno = 0;
#endif
	while (link(temp, name) == FAIL) {
		debug3(D_LOCK, "link(temp,name) failed, errno=%d\n", errno);
		if (errno == EEXIST) {		/* lock file already there */
		    if ((pid = readlock(name)) != FAIL && pid != 0 &&
			(pid == (int)getpid() ||
			 (kill(pid, 0) == FAIL && errno == ESRCH))) {
			    /* the lockfile was created by this process, or */
			    /* pid that created lockfile is gone */
			    if (unlink(name) == 0)
				    continue;
		    }
		}
		debug2(D_LOCK, "lock NOT made\n");
		unlink(temp);
		return FAIL;
	}
	debug2(D_LOCK, "lock made\n");
	unlink(temp);
	return SUCCESS;
}

/*
**	checklock() - test for presense of valid lock file
**
**	Returns TRUE if lockfile found, FALSE if not.
*/

boolean
checklock(name)
char *name;
{
	int pid;
	struct stat st;

	debug3(D_LOCK, "checklock(%s) called\n", name);

	if ((stat(name, &st) == FAIL) && errno == ENOENT) {
		debug2(D_LOCK, "stat failed, no file\n");
		return FALSE;
	}

	if ((pid = readlock(name)) == FAIL) {
		debug2(D_LOCK, "couldn't read lockfile\n");
		return FALSE;
	}

	if (pid != 0 && kill(pid, 0) == FAIL && errno == ESRCH) {
		debug2(D_LOCK, "no active process has lock, will remove\n");
		unlink(name);
		return FALSE;
	}

	debug2(D_LOCK, "active process has lock, return TRUE\n");
	return TRUE;
}

/*
**	readlock() - read contents of lockfile
**
**	Returns pid read or FAIL on error.
*/

int
readlock(name)
char *name;
{
	int fd, n;
	int pid = 0;
#ifdef ASCIIPID
	char apid[16];
#endif	/* ASCIIPID */

	if ((fd = open(name, O_RDONLY)) == FAIL)
		return FAIL;

#ifdef ASCIIPID
	n = read(fd, apid, sizeof(apid) - 1);
	close(fd);

	if (n > 0) {			/* not empty file */
	    apid[n] = '\0';

# ifdef BOTHPID
	    if (n == sizeof(pid))
		memcpy(&pid,apid,sizeof(pid));	/* binary pid */
	    else
# endif
	    if (sscanf(apid,"%d",&pid) != 1) /* ascii pid */
		pid = 0;
	}
#else
	n = read(fd, (char *)&pid, sizeof(pid));
	close(fd);

	if (n == 0)
		pid = 0;		/* empty file */
#endif	/* ASCIIPID */

	debug3(D_LOCK, "read %d from the lockfile\n", pid);
	return pid;
}

/*
**	rmlocks() - remove lockfile(s), if they are mine
*/

void
rmlocks()
{
	if (Lock[0] != '\0' && readlock(Lock) == (int)getpid()) {
	    unlink(Lock);
	    debug3(D_LOCK, "lockfile %s removed\n", Lock);
	}
}



/*************************************************************************
**	Chatting functions using UUCP-like chat scripts			**
*************************************************************************/

/*
**	chat() - handle expect/send sequence to Device
**
**	Returns FAIL if an error occurs.
*/

int
chat(s)
char *s;
{
	register int state = EXPECT;
	boolean finished = FALSE, if_fail = FALSE;
	char c, *p;
	char word[MAXLINE+1];		/* buffer for next word */

	debug3(D_CHAT, "chat(%s) called\n", s);

	while (!finished) {
		p = word;
		while (((c = (*s++ & 0177)) != '\0') && c != ' ' && c != '-')
			/*
			 *  SMR - I don't understand this, because if c is \0
			 *  then it is 0, isn't it?  If so we end the loop and
			 *  terminate the word anyway.
			 *
			*p++ = (c) ? c : '\177';
			 */
			*p++ = c;

		finished = (c == '\0');
		if_fail = (c == '-');
		*p = '\0';

		switch (state) {
		case EXPECT:
			if (expect(word) == FAIL) {
				if (if_fail == FALSE)
					return FAIL;	/* no if-fail seq */
			} else {
				/* eat up rest of current sequence
				 */
				if (if_fail == TRUE) {
					while ((c = (*s++ & 0177)) != '\0' &&
						c != ' ')
						;
					if (c == '\0')
						finished = TRUE;
					if_fail = FALSE;
				}
			}
			state = SEND;
			break;
		case SEND:
			if (send(word) == FAIL)
				return FAIL;
			state = EXPECT;
			break;
		}
	}
	debug2(D_CHAT, "chat() successful\n");
	return (SUCCESS);
}


/*
**	unquote() - decode char(s) after a '\' is found.
**
**	Returns the pointer s; decoded char in *c.
*/

const char	valid_oct[] = "01234567";
const char	valid_dec[] = "0123456789";
const char	valid_hex[] = "0123456789aAbBcCdDeEfF";

const char *
unquote(s, c)
const char *s;
char *c;
{
	int value, base;
	char n;
	const char *valid;

	n = *s++;
	switch (n) {
	case 'b':
		*c = '\b';	break;
	case 'c':
		if ((n = *s++) == '\n')
			*c = '\0';
		else
			*c = n;
		break;
	case 'f':
		*c = '\f';	break;
	case 'n':
		*c = '\n';	break;
	case 'r':
		*c = '\r';	break;
	case 's':
		*c = ' ';	break;
	case 't':
		*c = '\t';	break;
	case '\n':
		*c = '\0';	break;	/* ignore NL which follows a '\' */
	case '\\':
		*c = '\\';	break;	/* '\\' will give a single '\' */
	default:
		if (isdigit(n)) {
			value = 0;
			if (n == '0') {
				if (*s == 'x') {
					valid = valid_hex;
					base = 16;
					s++;
				} else {
					valid = valid_oct;
					base = 8;
				}
			} else {
				valid = valid_dec;
				base = 10;
				s--;
			}
			while (strchr(valid,*s) != NULL) {
				value *= base;
				if (isdigit(*s))
				    value += (int) (*s - '0');
				else
				    value += (int) (tolower(*s) - 'a' + 10);
				s++;
			}
			*c = (char) value;
		} else {
			*c = n;
		}
		break;
	}
	return s;
}


/*
**	send() - send a string to stdout
*/

int
send(s)
register const char *s;
{
	register int retval = SUCCESS;
	char ch;
	int inbuf;
	long diff;
	struct timeval now;
	char buf[MAXBUF];

	debug2(D_CHAT, "   SEND: ");

	/* before we send anything, make sure at least 20ms has passed */
	/* since the last matched input.  this is required because the */
	/* modem is not ready immediately after sending the response, */
	/* but requires slightly more delay....	 bummer! */

	gettimeofday(&now,NULL);

	if ((diff = now.tv_sec - MatchTime.tv_sec) <= 1) {
	    diff *= 1000000;			/* make microseconds */
	    diff += now.tv_usec - MatchTime.tv_usec;

	    if ((diff = 20000 - diff) > 0) {	/* calculate diff to 20ms */
		usleep(diff);			/* sleep the remainder */
		debug3(D_CHAT, "[%ldms] ", (diff+500)/1000);
	    }

	    MatchTime.tv_sec = 0;		/* don't sleep again */
	}

	debug2(D_CHAT, "(");

	if (!strcmp(s, "\"\"")) {	/* ("") used as a place holder */
		debug2(D_CHAT, "[nothing])\n");
		return retval;
	}

	debug1(D_CHAT, s);

	inbuf = 0;

	while ((ch = *s++) != '\0')
	{
	    if (ch == '\\')
	    {
		switch (*s)
		{
		case 'p':		/* '\p' == pause */
			if (inbuf && write(STDOUT, buf, inbuf) != inbuf) {
				retval = FAIL;
				break;
			}
			inbuf = 0;
			usleep(200000);
			s++;
			continue;
		case 'd':		/* '\d' == delay */
			if (inbuf && write(STDOUT, buf, inbuf) != inbuf) {
				retval = FAIL;
				break;
			}
			inbuf = 0;
			usleep(1000000);
			s++;
			continue;
		case 'K':		/* '\K' == BREAK */
			if (inbuf && write(STDOUT, buf, inbuf) != inbuf) {
				retval = FAIL;
				break;
			}
			inbuf = 0;
			ioctl(STDOUT, TCSBRK, 0);
			s++;
			continue;
		default:
			s = unquote(s, &ch);
			break;
		}
	    }
	    buf[inbuf++] = ch;
	    if (inbuf == MAXBUF) {
		if (write(STDOUT, buf, inbuf) != inbuf) {
		    retval = FAIL;
		    break;
		}
		inbuf = 0;
	    }
	}
	if (inbuf && write(STDOUT, buf, inbuf) != inbuf)
		retval = FAIL;
	debug3(D_CHAT, ") -- %s\n", (retval == SUCCESS) ? "OK" : "FAILED");
	return retval;
}


/*
**	expect() - look for a specific string on stdin
*/

static jmp_buf env;	/* here so expalarm() sees it */

int
expect(s)
register const char *s;
{
	register int i,ch;
	volatile int expfail;
	volatile int retval;
	sigh_t (*oldalarm)();
	char c, *p, word[MAXLINE+1];

	if (!strcmp(s, "\"\"")) {	/* ("") used as a place holder */
		debug2(D_CHAT, "   EXPECT: ([nothing])\n");
		gettimeofday(&MatchTime,NULL);	/* save current time */
		return SUCCESS;
	}

	expfail = EXPFAIL;		/* default fail timeout */

	/* look for escape chars in expected word
	 */
	for (p = word; (c = (*s++ & 0177)) != '\0';) {
		if (c == '\\') {
			if (*s == 'T') {	/* change expfail timeout */
				if (isdigit(*++s)) {
					s = unquote(s, &c);
					/* allow 3 - 255 second timeout */
					if ((expfail = ((unsigned char) c)) < 3)
						expfail = 3;
				}
				continue;
			} else
				s = unquote(s, &c);
		}
		*p++ = (c) ? c : '\177';
	}
	*p = '\0';

	settermios(INITIAL);
	oldalarm = signal(SIGALRM, expalarm);
	alarm((unsigned) expfail);

	if (setjmp(env)) {	/* expalarm returns non-zero here */
		debug3(D_CHAT, "[timed out after %d seconds]\n", expfail);
		signal(SIGALRM, oldalarm);
		gettimeofday(&MatchTime,NULL);	/* save current time */
		return FAIL;
	}

	debug3(D_CHAT, "   EXPECT: <%d> (", expfail);
	debug1(D_CHAT, word);
	debug2(D_CHAT, "), GOT: ");

	retval = FAIL;
	p = ChatBuf;

	while ((ch = getch()) != EOF) {
		debug3(D_CHAT, ((ch < ' ') ? "^%c" : "%c"),
			       ((ch < ' ') ? ch | 0100 : ch));
		*p++ = (char) ((int) ch & 0177);
		*p = '\0';
		if ((int)strlen(ChatBuf) >= (int)strlen(word)) {
			for (i=0; ChatBuf[i]; i++)
				if (expmatch(&ChatBuf[i], word)) {
					retval = SUCCESS;
					break;
				}
		}
		if (retval == SUCCESS)
			break;
	}
	alarm((unsigned) 0);
	signal(SIGALRM, oldalarm);
	debug3(D_CHAT, " -- %s\n", (retval == SUCCESS) ? "got it" : "Failed");
	gettimeofday(&MatchTime,NULL);		/* save current time */
	return retval;
}


/*
**	waitfor() - wait for a certain response (CR/LF terminated)
**
**	"state" is the settermios() state set before waiting, usually to
**	set the timeout.  when a negative value is passed, waitfor()
**	waits for CR, rather than LF.
*/

int
waitfor(word, state)
const char *word;
int state;
{
	int retval = FAIL;
	int ch;
	int bufpos;
#ifdef DEBUG
	struct timeval t1;
	time_t elapsed = 0;
#endif

	settermios(state < 0? -state : state);	/* set requested handling */

	debug3(D_CHAT, "   WAITFOR: <%d> (", Termios.c_cc[VTIME]);
	debug1(D_CHAT, word);
	debug2(D_CHAT, "), GOT: ");

#ifdef DEBUG
	if (Debug & D_CHAT)
	    gettimeofday(&t1,NULL);		/* for timing */
#endif

	Match = ChatBuf;			/* position for match */
	bufpos = 0;

	/* discard any leading CR and LF characters (from previous line) */

	while ((ch = getch()) != EOF && (ch == '\r' || ch == '\n' || ch == '\0'))
	    debug3(D_CHAT, "^%c", ch | 0100);

	if (ch == EOF)
	    goto done;

	ChatBuf[bufpos++] = ch;			/* character from above */
	debug3(D_CHAT, ((ch < ' ') ? "^%c" : "%c"),
		       ((ch < ' ') ? ch | 0100 : ch));

	/* keep filling the chatbuf until the read times out */

	while (bufpos < (sizeof(ChatBuf) - 1) && (ch = getch()) != EOF)
	{
	    debug3(D_CHAT, ((ch < ' ') ? "^%c" : "%c"),
			   ((ch < ' ') ? ch | 0100 : ch));

	    if (ch == '\0' || (ChatBuf[bufpos++] = ch) != (state < 0? '\r' : '\n'))
		continue;			/* wait for CR or LF */

	    /* we have received a CR or LF, check for matches in buffer */

	    ChatBuf[bufpos] = '\0';		/* 0-terminate */

	    if (word == NULL || expmatch(Match, word)) {
		retval = SUCCESS;
		break;
	    }

	    Match = &ChatBuf[bufpos];		/* start looking here */
	}

done:
	ChatBuf[bufpos] = '\0';			/* 0-terminate */
	gettimeofday(&MatchTime,NULL);		/* save current time */

#ifdef DEBUG
	if (Debug & D_CHAT) {
	    elapsed = (MatchTime.tv_usec - t1.tv_usec) / 1000 +
			(MatchTime.tv_sec - t1.tv_sec) * 1000;
	}
	debug4(D_CHAT, " -- %s (%d ms)\n", (retval == SUCCESS) ? "got it" : "Failed", elapsed);
#endif
	return retval;
}


/*
**	expmatch() - compares expected string with the one gotten
*/

boolean
expmatch(got, exp)
register const char *got;
register const char *exp;
{
	if (*exp == '*') {			/* wildcard? */
	    exp++;

	    while (*got)			/* match at all positions */
		if (expmatch(got++,exp))
		    return TRUE;

	    return FALSE;
	} else {
	    while (*exp) {
		if (*got++ != *exp && *exp != '?')
		    return FALSE;		/* no match */
		exp++;
	    }

	    return TRUE;
	}
}


/*
**	expalarm() - called when expect()'s SIGALRM goes off
*/

sigh_t
expalarm(sig)
int sig;
{
	longjmp(env, sig);
}


/*************************************************************************
**	Defaults file functions						**
*************************************************************************/

/*
**	defbuild() - create in-core list of defaults
**
**	Returns (DEF**)NULL if no defaults file found or an error occurs.
*/

DEF **
defbuild(filename)
char *filename;
{
	register int i;
	register DEF *dp;
	register DEF *next;
	FILE *fp;
	char *fname, defname[MAXLINE+1], buf[MAXLINE+1];
	static DEF *deflist[MAXDEF+1];		/* in-core list */
	struct stat st;

	debug3(D_DEF, "defbuild(%s) called\n",
			((filename == NULL) ? "NULL" : filename));

	/* look to see if there's a DEFAULTS/MyName.Device file
	 */
	sprintf(buf, "%s", DEFAULTS);
	strcat(buf, ".%s");
	sprintf(defname, buf, MyName, Device);
	debug3(D_DEF, "looking for %s\n", defname);
	if ((stat(defname, &st) == FAIL) && errno == ENOENT) {	/* Nope */
		debug2(D_DEF, "stat failed, no file\n");
		sprintf(defname, DEFAULTS, MyName);
	}

	fname = (filename != NULL) ? filename : defname;

	/* if fname doesn't begin with a '/', assume it's a
	 * filename to be made "DEFAULTS/fname"
	 */
	if (*fname != '/') {
		sprintf(defname, DEFAULTS, fname);
		fname = defname;
	}

	debug3(D_DEF, "fname = (%s)\n", fname);

	if ((fp = defopen(fname)) == NULL) {
		debug2(D_DEF, "defopen() failed\n");
		return NULL;		/* couldn't open file */
	}

	for (i=0; i < MAXDEF; i++) {
		if ((dp = defread(fp)) == NULL)
			break;
		if ((next = (DEF *) malloc((unsigned) sizeof(DEF))) == NULL) {
			logmesg("malloc() failed: defaults list truncated");
			break;
		}
		next->name = dp->name;
		next->value = dp->value;
		deflist[i] = next;
		debug5(D_DEF, "deflist[%d]: name=(%s), value=(%s)\n",
				i, deflist[i]->name, deflist[i]->value);
	}
	deflist[i] = NULL;	/* terminate list */
	defclose(fp);
	debug2(D_DEF, "defbuild() successful\n");
	return deflist;
}


/*
**	defvalue() - locate the value in "deflist" that matches "name"
**
**	Returns (char*)NULL if no match is made.
*/

char *
defvalue(deflist, name)
register DEF **deflist;
register const char *name;
{
	debug3(D_DEF, "defvalue(%s) called\n", name);

	if (deflist != NULL)
		for (; *deflist != NULL; deflist++)
			if (!strcmp(name, (*deflist)->name)) {
				debug3(D_DEF, "defvalue returns (%s)\n",
						(*deflist)->value);
				return (*deflist)->value;  /* normal exit */
			}

	debug2(D_DEF, "defvalue returns NULL\n");
	return NULL;
}


/*
**	defopen() - open the defaults file
**
**	Returns (FILE*)NULL if file not found or an error occurs.
*/

FILE *
defopen(filename)
register char *filename;
{
	if (filename != NULL)
		return fopen(filename, "r");

	return NULL;
}


/*
**	defread() - read a line from the defaults file
**
**	Returns (DEF*)NULL if an error occurs.
*/

DEF *
defread(fp)
register FILE *fp;
{
	register char *p;
	char buf[MAXLINE+1];	/* buffer large enough for 1 line */
	static DEF def;

	do {
		if (fgets(buf, sizeof(buf), fp) == NULL)
			return NULL;	/* no more lines */

	} while ((buf[0] == '#') || (buf[0] == '\n'));
	  /* SMR - ignore comment lines */

	buf[strlen(buf)-1] = '\0';		/* rm trailing \n */

	/* lines should be in the form "NAME=value"
	 */
	if ((p = strchr(buf, '=')) == NULL) {
		sprintf(MsgBuf, "bad defaults line: %s", buf);
		logmesg(MsgBuf);
		return NULL;
	}
	*p++ = '\0';		/* split into two fields, name and value */
	def.name = strdup(buf);
	def.value = strdup(p);

	return &def;
}


/*
**	defclose() - closes the defaults file
**
**	Returns EOF if an error occurs.
*/

int
defclose(fp)
register FILE *fp;
{
	return fclose(fp);
}


/*************************************************************************
**	Error and Event logging						**
*************************************************************************/

/*
**	logmesg() - display/log a message
*/

void
logmesg(msg)
register const char *msg;
{
	register FILE *fp;
	time_t tim;
	char logfile[80];

	sprintf(logfile, LOGFILE, MyName);	/* make logfile name */

	if ((fp = fopen(logfile, "a")) != NULL ||
	    (fp = fopen(logfile, "w")) != NULL) /* REJ: Linux can't append to /dev/console */
	{
		tim = time(NULL);
		fprintf(fp, "%.24s %s: %s\n", ctime(&tim), Device, msg);
		fclose(fp);
	}

#ifdef TRYMAIL
	else {
		char buf[MAXLINE];

		sprintf(buf, "%s %s", Mailer, NOTIFY);
		if ((fp = popen(buf, "w")) != NULL) {
			fprintf(fp, "From: %s %s\n", MyName, Device);
			fprintf(fp, "To: %s\n", NOTIFY);
			fprintf(fp, "Subject: %s problem\n\n", MyName);
			fprintf(fp, "%s: %s\n", Device, msg);
			pclose(fp);
		}
	}
#endif	/* TRYMAIL */
}


/*************************************************************************
**	Debugging functions						**
*************************************************************************/

#ifdef DEBUG

/*
**	opendebug() - open the debug outputfile when Debug configured
**
**	first checks if file not yet opened, as it is tried both after
**	parsing the options and after reading the defaults file
**
*/

int
opendebug ()

{
	time_t tim;
	int fd;
	char buf[MAXLINE+1];

	if (Debug && Dfp == NULL) {
	    sprintf(buf, "/tmp/%s:%s", MyName, Device);
	    if ((Dfp = fopen(buf, "a")) == NULL) {
		logmesg("cannot open debug file");
		return FAIL;
	    } else {
		if (fileno(Dfp) < 3) {
		    if ((fd = dup2(fileno(Dfp),3)) > 2) {
			fclose(Dfp);
			Dfp = fdopen(fd, "a");
		    }
		}
		tim = time(NULL);
		fprintf(Dfp, "%s [%.32s] Started: %s", MyName, Id + 4, ctime(&tim));
		fflush(Dfp);
	    }
	}

	return SUCCESS;
}

/*
**	debug() - an fprintf to the debug file
**
**	Only does the output if the requested level is "set."
*/

/*VARARGS*/
void
debug(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (Dfp != NULL) {
	    vfprintf(Dfp, fmt, args);
	    fflush(Dfp);
	}
	va_end(args);
}

/*
**	debugt() - put timestamp in debug file
**
**	Does not flush the output, it is assumed a debug call follows.
*/

void
debugt()
{
	struct timeval tv;
	time_t tim;
	char buf[32];

	if (Dfp != NULL) {
	    gettimeofday(&tv,NULL);
	    tim = tv.tv_sec;
	    strftime(buf,sizeof(buf),"%X",localtime(&tim));
	    fprintf(Dfp,"%s.%03d: ",buf,(int)(tv.tv_usec / 1000));
	}
}

/*
**	dprint() - like debug(), but shows control chars
*/

void
dprint(word)
const char *word;
{
	const char *p, *fmt;
	char ch;

	if (word == NULL)
	    fprintf(Dfp,"(null)");
	else {
	    p = word;
	    while ((ch = *p++) != '\0') {
		if (ch < ' ') {
			fmt = "^%c";
			ch = ch | 0100;
		} else {
			fmt = "%c";
		}
		fprintf(Dfp, fmt, ch);
	    }
	}
	fflush(Dfp);
}

#endif	/* DEBUG */

/*************************************************************************
**	Soundcard functions						**
*************************************************************************/

/*
**	sound_open() - open the sound device and set params
**
**	will attempt to connect to the specified sound device and sets
**	the parameters common to all usage of the soundcard.
*/

int
sound_open()

{
	sound_close();

#ifdef WITH_SOUND
	if (SoundDev != NULL) {
	    int n;

	    if ((SoundFp = fopen(SoundDev,"w")) == NULL)
	    	return FAIL;

	    n = APF_NETWORK;			/* underruns occur */
	    ioctl(fileno(SoundFp),SNDCTL_DSP_PROFILE,&n);

	    if (ioctl(fileno(SoundFp),SNDCTL_DSP_GETBLKSIZE,&n) < 0 ||
		n < 1 || n > 65536)
	    {
	    	sound_close();
		return FAIL;
	    }

	    debug4(D_RUN, "sound: dev %s blksize %d\n",SoundDev,n);
	    setvbuf(SoundFp,NULL,_IOFBF,n);	/* buffer BLKSIZE bytes */

	    n = 0;				/* select MONO */
	    if (ioctl(fileno(SoundFp),SNDCTL_DSP_STEREO,&n) < 0) {
	    	sound_close();
		return FAIL;
	    }

	    sound_mode(ENC_BEST,Fsamp);		/* initial enc/rate setting */
	}
#endif
	return SUCCESS;
}

/*
**	sound_mode() - set encoding and samplerate
**
**	gets an ENC_xxx parameter and sample rate in samples/s
*/

int
sound_mode(enc,fsamp)
int enc;
int fsamp;

{
#ifdef WITH_SOUND
	if (SoundFp != NULL) {
	    int fmt;				/* sample format */

	    switch (enc)
	    {
	    case ENC_S16:
		fmt = AFMT_S16_LE;		/* Little-endian signed 16bit */
		break;

	    case ENC_ALAW:
		fmt = AFMT_A_LAW;
		break;

	    case ENC_ULAW:
		fmt = AFMT_MU_LAW;
		break;

	    default:
		fmt = AFMT_S8;
		break;
	    }

	    if (fmt != SoundFmt) {
		fflush(SoundFp);
		ioctl(fileno(SoundFp),SNDCTL_DSP_SYNC,NULL);

		debug3(D_RUN, "sound: setfmt %04x\n",fmt);
		if (ioctl(fileno(SoundFp),SNDCTL_DSP_SETFMT,&fmt) < 0) {
		    sound_close();
		    return FAIL;
		}

		SoundFmt = fmt;
	    }

	    if (fsamp != SoundRt) {
		fflush(SoundFp);
		ioctl(fileno(SoundFp),SNDCTL_DSP_SYNC,NULL);

		debug3(D_RUN, "sound: speed %d\n",fsamp);
		if (ioctl(fileno(SoundFp),SNDCTL_DSP_SPEED,&fsamp) < 0) {
		    sound_close();
		    return FAIL;
		}

		SoundRt = fsamp;
	    }
	}
#endif
	return SUCCESS;
}


/*
**	sound_close() - close the sound device, free buffer
*/

void
sound_close()

{
#ifdef WITH_SOUND
	if (SoundFp != NULL) {
	    fflush(SoundFp);
	    ioctl(fileno(SoundFp),SNDCTL_DSP_SYNC,NULL);
	    fclose(SoundFp);
	    SoundFp = NULL;
	    SoundFmt = SoundRt = 0;
	}
#endif
}
